﻿unit Queue.main;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Classes, System.Variants,
  FMX.Types, FMX.Controls, FMX.Forms, FMX.Graphics, FMX.Dialogs,

  Cinema.Interfaces.FuncExport,
  Cinema.Scenario.Interfaces,
  Cinema.Interfaces, FMX.Memo.Types, FMX.Controls.Presentation, FMX.ScrollBox, FMX.Memo, System.ImageList, FMX.ImgList, FMX.StdCtrls, FMX.Layouts, FMX.Edit
  ;

type
  TfrmQueue = class(TForm, ICinemaPlugin)
    mmAcceptedMessages: TMemo;
    mmNewMessage: TMemo;
    Splitter1: TSplitter;
    loNewMessage: TLayout;
    sbSendNewMsg: TSpeedButton;
    ImageList: TImageList;
    Layout1: TLayout;
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure sbSendNewMsgClick(Sender: TObject);
  private
    const
      csQueueGuid: TGuid = '{CFD59BF6-CC8E-4CE6-9A0F-6E27E95830F0}';
      csDequeueGuid: TGuid = '{C820AD9B-5A82-4954-B4E4-CE106DB16A54}';
    type
      TScenarioExternalEventListener = class(IExternalEventListener)
      private
        FOwner: TfrmQueue;
      public
        constructor Create(AOwner: TfrmQueue);
        procedure OnChangeNodeStateEvent(AEvent: IExternalEventListener.IStateEvent); override; stdcall;
        procedure OnAcceptNewMessageEvent(AEvent: IExternalEventListener.IMessageEvent); override; stdcall;
      end;
  private
    { Private declarations }
    FChart: ICinemaScenario;
    FChartEventListener: TScenarioExternalEventListener;
    FQueuePlugin: IScenarioPlugin;
    FDataStreamProcessor: Pointer;
    FDequeueDataStreamProcessor: Pointer;
    FDeQueuePlugin: IScenarioPlugin;
    procedure OnAcceptNewMessageEvent(AEvent: string);
    function CreateDataStreamProcessor(AInputDataStream: TDataStream; ANumberInput: Integer): Pointer; stdcall;
    procedure ProcessDataStream(AProcessor: Pointer; AData: PByte; ASize: Int64); stdcall;
    procedure DestroyDataStreamProcessor(AProcessor: Pointer); stdcall;
    /// <summary> обновить графику плагина (таблицы и прочее что в будущем ещё может понадобится вывести) </summary>
    procedure RedrawView; stdcall;
  public
    { Public declarations }
  end;

var
  frmQueue: TfrmQueue;

implementation

uses
  Cinema.Attributes.Constants,
  Cinema.Configuration;

{$R *.fmx}

function TfrmQueue.CreateDataStreamProcessor(AInputDataStream: TDataStream; ANumberInput: Integer): Pointer;
begin
  Result := nil;
end;

procedure TfrmQueue.DestroyDataStreamProcessor(AProcessor: Pointer);
begin

end;

procedure TfrmQueue.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if Assigned(FDataStreamProcessor) then
  begin
    FQueuePlugin.CinemaPlugin.DestroyDataStreamProcessor(FDataStreamProcessor);
    FDeQueuePlugin.CinemaPlugin.DestroyDataStreamProcessor(FDequeueDataStreamProcessor);
  end;
  FQueuePlugin := nil;
  FChart.TryTerminate;
  FChart := nil;
  TThread.RemoveQueuedEvents(TThread.Current);
  FreeAndNil(FChartEventListener);
end;

procedure TfrmQueue.FormShow(Sender: TObject);
var
  //cinema: ICinema;
  pluginObject: TInterfacedObject;
  //destIp: AnsiString;
begin
  FChartEventListener := TScenarioExternalEventListener.Create(Self);
  FChart := LoadScenarioExt(PWideChar(TCinemaConfiguration.RtlScenariosPath + 'QueueTest.json'), FChartEventListener);
  FChart.TryStart;
  pluginObject := FChart.GetChildByGuid(csQueueGuid);
  if Supports(pluginObject, IScenarioPlugin, FQueuePlugin) then
  begin
    FDataStreamProcessor := FQueuePlugin.CinemaPlugin.CreateDataStreamProcessor(nil, 1);
    //destIp := AnsiString(edtIp.Text);
    //if destIp <> '' then
    //  cinema.AddDataStreamAttribute(FDataStreamProcessor, csAttrClientAddress, @destIp[1], length(destIp));
    pluginObject := FChart.GetChildByGuid(csDequeueGuid);
    if Supports(pluginObject, IScenarioPlugin, FDeQueuePlugin) then
    begin
      FDequeueDataStreamProcessor := FDeQueuePlugin.CinemaPlugin.CreateDataStreamProcessor(Self as ICinemaPlugin, 2);
    end;
  end;
end;

procedure TfrmQueue.OnAcceptNewMessageEvent(AEvent: string);
begin
  mmAcceptedMessages.Lines.Add(AEvent);
end;

procedure TfrmQueue.ProcessDataStream(AProcessor: Pointer; AData: PByte; ASize: Int64);
var
  eventMsg: string;
begin
  inherited;
  eventMsg := PWideChar(AData);
  TThread.ForceQueue(TThread.Current, procedure begin OnAcceptNewMessageEvent(eventMsg); end);
end;

procedure TfrmQueue.RedrawView;
begin

end;

procedure TfrmQueue.sbSendNewMsgClick(Sender: TObject);
var
  msg: string;
begin
  if Assigned(FDataStreamProcessor) then
  begin
    msg := mmNewMessage.Text;
    if msg <> '' then
      FQueuePlugin.CinemaPlugin.ProcessDataStream(FDataStreamProcessor, @msg[1], msg.Length * 2);
    mmNewMessage.Lines.Clear;
  end;
end;

{ TfrmQueue.TScenarioExternalEventListener }

constructor TfrmQueue.TScenarioExternalEventListener.Create(AOwner: TfrmQueue);
begin
  FOwner := AOwner;
end;

procedure TfrmQueue.TScenarioExternalEventListener.OnAcceptNewMessageEvent(AEvent: IExternalEventListener.IMessageEvent);
var
  eventMsg: string;
begin
  inherited;
  eventMsg := AEvent.Msg;
  TThread.ForceQueue(TThread.Current, procedure begin FOwner.OnAcceptNewMessageEvent(eventMsg); end);
end;

procedure TfrmQueue.TScenarioExternalEventListener.OnChangeNodeStateEvent(AEvent: IExternalEventListener.IStateEvent);
begin
  inherited;

end;

end.
