﻿unit Cinema.Attributes.Dictionary;

// <summary> Словарь атрибутов канала данных </summary>

{$ifdef FPC}
	{$mode delphi}
{$endif}

interface

uses
  {%H-}Rtti,
  TypInfo,
  Generics.Collections,
  Cinema.Types,
  Cinema.Log
  ;

type

  TAttribute = class;

  TGroupAttributes = class
  private
    FName: string;
    FDescription: string;
    FId: int32;
    FParentGroup: TGroupAttributes;
    FAttributes: TList<TAttribute>;
  public
    constructor Create(AId: int32; const AName, ADesc: string; AParentGroup: TGroupAttributes);
    destructor Destroy; override;
    procedure AddAttribute(AAttribute: TAttribute);
    property Id: int32 read FId;
    property Name: string read FName;
    property Description: string read FDescription;
  end;

  TAttribute = class
  private
    FName: string;
    FId: int32;
    FType: TAttributeType;
    FGroup: TGroupAttributes;
    FDescription: string;

  public
    constructor Create(AId: int32; AType: TAttributeType; const AName, ADesc: string; AGroup: TGroupAttributes);
    property Group: TGroupAttributes read FGroup;
    property Id: int32 read FId;
    property &Type: TAttributeType read FType;
    property Name: string read FName;
    property Description: string read FDescription;
  end;

  TAttributesDictionary = class
  private
    FGroups: TList<TGroupAttributes>;
    FAttributes: TList<TAttribute>;
    FAttributesDic: TDictionary<string, TAttribute>;
    FJsonFileWithAttributes: string;
    function GetAttribute(ID: Integer): TAttribute; overload;
    function GetAttribute(const Name: string): TAttribute; overload;
    procedure AddGroup(AGroup: TGroupAttributes);
    procedure AddAttribute(AAttribute: TAttribute);
    procedure Clear;
    procedure Load;
    class function StrToNative(ATargetType: PTypeInfo; const AValue: string): TValue;
  public
    constructor Create(const AJsonFileWithAttributes: string);
    destructor Destroy; override;
    procedure AfterConstruction; override;
    class function StringToNative<T>(const AValue: string): T; overload;
    class function StringToValue(AAttribute: TAttribute; const AValue: string): TValue; overload;
    class function StringToValue(AAttributeType: TAttributeType; const AValue: string): TValue; overload;
    class function ValueToString(AAttribute: TAttribute; AValue: PByte; ASize: int32): string; overload;
    class function ValueToString(AAttributeType: TAttributeType; AValue: PByte; ASize: int32): string; overload;
    property Attribute[ID: Integer]: TAttribute read GetAttribute;
    property Attributes[const Name: String]: TAttribute read GetAttribute;
  end;

implementation

uses
  SysUtils,
  {$ifdef FPC}
  fpjson,
  jsonparser,
  StrUtils,
  {$else}
  System.IOUtils,
  System.JSON,
	{$endif}
  Cinema.Utils
  ;

type

  TStringToValueConverter = function (const AValue: string): TValue;
  TValueToStringConverter = function (AValue: PByte; ASize: int32): string;

function StringToBlob(const AValue: string): TValue;
//var
//  bytes: TArray<Byte>;
//  i: Integer;
begin
//  SetLength(bytes, Length(AValue) shr 1);
//  for i := 1 to Length(AValue) shr 1 do
//  begin
//    bytes[i-1] := HexToByte[byte(AData[i shl 1 - 1])] shl 4 + HexToByte[byte(AData[i shl 1])];
//  end;
//  TValue.Make(@bytes[0], Result);
  // todo
  Result := TValue.Empty;
end;

function StringToGuid1(const AValue: string): TValue;
var
  val: TGuid;
begin
  val := StringToGUID(AValue);
  TValue.Make(@val, TypeInfo(TGuid), Result);
end;

function StringToInt(const AValue: string): TValue;
var
  val: Integer;
begin
  val := StrToInt(AValue);
  TValue.Make(@val, TypeInfo(Integer), Result);
end;

function StringToInt64(const AValue: string): TValue;
var
  val: Int64;
begin
  val := StrToInt64(AValue);
  TValue.Make(@val, TypeInfo(Int64), Result);
end;

function StringToFloat(const AValue: string): TValue;
var
  val: Single;
begin
  val := StrToFloat(AValue);
  TValue.Make(@val, TypeInfo(single), Result);
end;

function StringToString(const AValue: string): TValue;
var
  val: WideString;
begin
  val := WideString(AValue);
  TValue.Make(@val, TypeInfo(WideString), Result);
end;

function StringToAnsiString(const AValue: string): TValue;
var
  val: AnsiString;
begin
  val := AnsiString(AValue);
  TValue.Make(@val, TypeInfo(AnsiString), Result);
end;

function StringToDateTime(const AValue: string): TValue;
var
  val: TDateTime;
begin
  // без промежуточной переменной FPC падает в Internal error 200304235
  val := StrToDateTime(AValue);
  TValue.Make(@val, TypeInfo(TDateTime), Result);
end;

function StringToDate(const AValue: string): TValue;
var
  val: TDate;
begin
  // без промежуточной переменной FPC падает в Internal error 200304235
  val := StrToDateTime(AValue);
  TValue.Make(@val, TypeInfo(TDate), Result);
end;

function StringToTime(const AValue: string): TValue;
var
  val: TTime;
begin
  // без промежуточной переменной FPC падает в Internal error 200304235
  val := StrToDateTime(AValue);
  TValue.Make(@val, TypeInfo(TTime), Result);
end;

function StringToExtended(const AValue: string): TValue;
var
  val: Extended;
begin
  // без промежуточной переменной FPC падает в Internal error 200304235
  val := StrToFloat(AValue);
  TValue.Make(@val, TypeInfo(Extended), Result);
end;

function StringToDouble(const AValue: string): TValue;
var
  val: double;
begin
  val := StrToFloat(AValue);
  // без промежуточной переменной FPC падает в Internal error 200304235
  TValue.Make(@val, TypeInfo(double), Result);
end;

function StringToBoolean(const AValue: string): TValue;
var
  val: boolean;
begin
  val := StrToBool(AValue);
  TValue.Make(@val, TypeInfo(boolean), Result);
end;

function StringToCurrency(const AValue: string): TValue;
var
  curr: Currency;
begin
  curr := StrToFloat(AValue);
  TValue.Make(@curr, TypeInfo(Currency), Result);
end;

function StringToByte(const AValue: string): TValue;
var
  val: Byte;
begin
  val := StrToUInt(AValue);
  TValue.Make(@val, TypeInfo(Byte), Result);
end;

function {%H-}StringToTimeStamp(const AValue: string): TValue;
//var
//  timeStamp: TSQLTimeStamp;
begin
  raise Exception.Create('TODO: fix me!');
  // todo
  //Result := TValue.Empty;
//  if TryStrToSQLTimeStamp(AData, timeStamp) then
//    AToParam.AsSQLTimeStamps[AIndexCell] := timeStamp;
end;

function {%H-}StringToTimeStampOffset(const AValue: string): TValue;
//var
//  timeStamp: TSQLTimeStampOffset;
begin
  raise Exception.Create('TODO: fix me!');
  // todo
  //Result := TValue.Empty;
//  if TryStrToSQLTimeStampOffset(AData, timeStamp) then
//    AToParam.AsSQLTimeStamps[AIndexCell] := timeStamp;
end;

function {%H-}TODOStringToValue(const AValue: string): TValue;
begin
  raise Exception.Create('TODO: fix me!');
end;

function {%H-}TODOToString(AValue: PByte; ASize: int32): string;
begin
  raise Exception.Create('TODO: fix me!');
end;

function BlobToString(AValue: PByte; ASize: int32): string;
//var
//  bytes: TArray<Byte>;
//  i: Integer;
begin
//  bytes := AField.AsBytes;
//  if length(bytes) > 0 then
//  begin
//    SetLength(Result, length(bytes)*2);
//    for i := 0 to length(bytes) - 1 do
//    begin
//      move(ByteToHex[bytes[i]][1], Result[i shl 1 + 1], 4);
//    end;
//  end else
//    Result := '';
  // TODO
  Result := '';
end;

function Int8ToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(Shortint) then
    Result := IntToStr(PShortint(AValue)^)
  else
    Result := '';
end;

function UInt8ToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(byte) then
    Result := IntToStr(PByte(AValue)^)
  else
    Result := '';
end;

function UInt16ToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(word) then
    Result := IntToStr(PWord(AValue)^)
  else
    Result := '';
end;

function Int16ToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(Smallint) then
    Result := IntToStr(PSmallint(AValue)^)
  else
    Result := '';
end;

function Int32ToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(Int32) then
    Result := IntToStr(PInteger(AValue)^)
  else
    Result := '';
end;

function UInt32ToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(UInt32) then
    Result := IntToStr(PCardinal(AValue)^)
  else
    Result := '';
end;

function Int64ToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(Int64) then
    Result := IntToStr(PInt64(AValue)^)
  else
    Result := '';
end;

function BoolToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(Boolean) then
    Result := BoolToStr(PBoolean(AValue)^, true)
  else
    Result := '';
end;

function AnsiStrToString(AValue: PByte; ASize: int32): string;
var
  str: AnsiString;
begin
  if ASize > 0 then
  begin
    SetLength(str, ASize);
    move(AValue^, str[1], ASize);
    Result := string(str);
  end else
    Result := '';
end;

function StrToString(AValue: PByte; ASize: int32): string;
begin
  if ASize > 0 then
  begin
    SetLength(Result, ASize shl 1);
    move(AValue^, Result[1], ASize);
  end else
    Result := '';
end;

function FloatToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(Single) then
    Result := FloatToStr(PSingle(AValue)^)
  else
    Result := '';
end;

function CurrencyToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(Currency) then
    Result := FloatToStr(PCurrency(AValue)^)
  else
    Result := '';
end;

function ExtendedToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(Extended) then
    Result := FloatToStr(PExtended(AValue)^)
  else
    Result := '';
end;

function DateToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(TDate) then
    Result := DateToStr(PCurrency(AValue)^)
  else
    Result := '';
end;

function TimeToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(TTime) then
    Result := TimeToStr(PCurrency(AValue)^)
  else
    Result := '';
end;

function DateTimeToString(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(TDateTime) then
    Result := DateTimeToStr(PCurrency(AValue)^)
  else
    Result := '';
end;

function GuidToStr(AValue: PByte; ASize: int32): string;
begin
  if ASize = SizeOf(TGuid) then
    Result := GUIDToString(PGuid(AValue)^)
  else
    Result := '';
end;

const
{$region FromString}
  CONVERTERS_FROM_STR: array[TAttributeType] of TStringToValueConverter =
  (
    StringToBlob,       // ftUnknown,
    StringToAnsiString, // ftString,
    StringToInt,        // ftSmallint,
    StringToInt,        // ftInteger,
    StringToInt,        // ftWord, // 0..4
    StringToBoolean,    // ftBoolean,
    StringToFloat,      // ftFloat,
    StringToCurrency,   // ftCurrency,
    TODOStringToValue,  // ftBCD,
    StringToDate,       // ftDate,
    StringToTime,       // ftTime,
    StringToDateTime,   // ftDateTime, // 5..11
    StringToBlob,       // ftBytes,
    StringToBlob,       // ftVarBytes,
    TODOStringToValue,  // ftAutoInc,
    StringToBlob,       // ftBlob,
    TODOStringToValue,  // ftMemo,
    TODOStringToValue,  // ftGraphic,
    TODOStringToValue,  // ftFmtMemo, // 12..18
    TODOStringToValue,  // ftParadoxOle,
    TODOStringToValue,  // ftDBaseOle,
    TODOStringToValue,  // ftTypedBinary,
    TODOStringToValue,  // ftCursor,
    StringToAnsiString, // ftFixedChar,
    StringToString,     // ftWideString, // 19..24
    StringToInt64,      // ftLargeint,
    TODOStringToValue,  // ftADT,
    TODOStringToValue,  // ftArray,
    TODOStringToValue,  // ftReference,
    TODOStringToValue,  // ftDataSet,
    StringToBlob,       // ftOraBlob,
    StringToBlob,       // ftOraClob, // 25..31
    TODOStringToValue,  // ftVariant,
    TODOStringToValue,  // ftInterface,
    TODOStringToValue,  // ftIDispatch,
    StringToGuid1,      // ftGuid,
    StringToTimeStamp,  // ftTimeStamp,
    TODOStringToValue,  // ftFMTBcd, // 32..37
    StringToString,     // ftFixedWideChar,
    TODOStringToValue,  // ftWideMemo,
    StringToTimeStamp,  // ftOraTimeStamp,
    TODOStringToValue,  // ftOraInterval, // 38..41
    StringToInt64,      // ftLongWord,
    StringToInt,        // ftShortint,
    StringToByte,       // ftByte,
    StringToExtended,   // ftExtended,
    TODOStringToValue,  // ftConnection,
    TODOStringToValue,  // ftParams,
    TODOStringToValue,  // ftStream, //42..48
    StringToTimeStamp,  // ftTimeStampOffset,
    TODOStringToValue,  // ftObject,
    StringToFloat       // ftSingle
  );

{$endregion}

  STRING_TO_NATIVE_CONVERTERS: array[TTypeKind] of TStringToValueConverter =
  (
      StringToBlob,       // tkUnknown,
      StringToInt,        // tkInteger,
      StringToAnsiString, // tkChar
      TODOStringToValue,  // tkEnumeration,
      StringToFloat,      // tkFloat,
      StringToAnsiString, // tkString,
      TODOStringToValue,  // tkSet,
      TODOStringToValue,  // tkClass,
      TODOStringToValue,  // tkMethod,
      StringToAnsiString, // tkWChar,
      TODOStringToValue,  // tkLString,
      StringToString,     // tkWString,
      TODOStringToValue,  // tkVariant,
      StringToBlob,       // tkArray,
      TODOStringToValue,  // tkRecord,
      TODOStringToValue,  // tkInterface,
      StringToInt,        // tkInt64,
      StringToBlob,       // tkDynArray,
      TODOStringToValue,  // tkUString,
      TODOStringToValue,  // tkClassRef,
      TODOStringToValue,  // tkPointer,
      TODOStringToValue,  // tkProcedure,
      TODOStringToValue   // tkMRecord
      {$ifdef FPC}
      , nil, nil, nil, nil, nil, nil, nil
      {$endif}
  );

{$region ToString}
  CONVERTERS_TO_STRING: array[TAttributeType] of TValueToStringConverter =
  (
    BlobToString, // ftUnknown,
    AnsiStrToString, // AnsiStrToString,
    Int16ToString, // ftSmallint,
    Int32ToString, // ftInteger,
    UInt16ToString, // ftWord, // 0..4
    BoolToString, // ftBoolean,
    FloatToString, // ftFloat,
    CurrencyToString, // ftCurrency,
    TODOToString, // ftBCD,          // TODO
    DateToString, // ftDate,
    TimeToString, // ftTime,
    DateTimeToString, // ftDateTime, // 5..11
    BlobToString, // ftBytes,
    BlobToString, // ftVarBytes,
    Int32ToString, // ftAutoInc,
    BlobToString, // ftBlob,
    BlobToString, // ftMemo,
    BlobToString, // ftGraphic,
    BlobToString, // ftFmtMemo, // 12..18
    BlobToString, // ftParadoxOle,
    BlobToString, // ftDBaseOle,
    BlobToString, // ftTypedBinary,
    BlobToString, // ftCursor,
    AnsiStrToString, // ftFixedChar,
    StrToString, // ftWideString, // 19..24
    Int64ToString, // ftLargeint,
    BlobToString, // ftADT,
    BlobToString, // ftArray,
    BlobToString, // ftReference,
    TODOToString, // ftDataSet,
    BlobToString, // ftOraBlob,
    BlobToString, // ftOraClob, // 25..31
    TODOToString, // ftVariant,
    TODOToString, // ftInterface,
    TODOToString, // ftIDispatch,
    GuidToStr, // ftGuid,
    TODOToString, // ftTimeStamp,
    TODOToString, // ftFMTBcd, // 32..37
    StrToString, // ftFixedWideChar,
    TODOToString, // ftWideMemo,
    TODOToString, // ftOraTimeStamp,
    TODOToString, // ftOraInterval, // 38..41
    UInt32ToString, // ftLongWord,
    Int8ToString, // ftShortint,
    UInt8ToString, // ftByte,
    ExtendedToString, // ftExtended,
    TODOToString, // ftConnection,
    TODOToString, // ftParams,
    TODOToString, // ftStream, //42..48
    TODOToString, // ftTimeStampOffset,
    TODOToString, // ftObject,
    FloatToString  // ftSingle
    );

{$endregion}

{ TAttributesDictionary }

class function TAttributesDictionary.StrToNative(ATargetType: PTypeInfo; const AValue: string): TValue;
begin
  Result := STRING_TO_NATIVE_CONVERTERS[ATargetType.Kind](AValue);
end;

procedure TAttributesDictionary.AddAttribute(AAttribute: TAttribute);
begin
  // ID атрибута является индексом в списке (для максимального быстродействия словаря),
  // поэтому устанавливаем соответствующий размер списка
  if FAttributes.Count - 1 < AAttribute.Id then
    FAttributes.Count := AAttribute.Id + 1;

  FAttributes.Items[AAttribute.Id] := AAttribute;

  FAttributesDic.Add(AAttribute.Name, AAttribute);
end;

procedure TAttributesDictionary.AddGroup(AGroup: TGroupAttributes);
begin
  FGroups.Add(AGroup);
end;

procedure TAttributesDictionary.AfterConstruction;
begin
  inherited;
  Load;
end;

procedure TAttributesDictionary.Clear;
var
  i: Integer;
begin
  for i := 0 to FAttributes.Count - 1 do
  begin
    FAttributes.Items[i].Free;
  end;
  FAttributes.Count := 0;

  for i := 0 to FGroups.Count - 1 do
  begin
    FGroups.Items[i].Free;
  end;
  FGroups.Count := 0;

  FAttributesDic.Clear;
end;

constructor TAttributesDictionary.Create(const AJsonFileWithAttributes: string);
begin
  FJsonFileWithAttributes := AJsonFileWithAttributes;
  FAttributes := TList<TAttribute>.Create;
  FGroups := TList<TGroupAttributes>.Create;
  FAttributesDic := TDictionary<string, TAttribute>.Create;
end;

destructor TAttributesDictionary.Destroy;
begin
  Clear;
  FAttributesDic.Free;
  FAttributes.Free;
  FGroups.Free;
  inherited;
end;

function TAttributesDictionary.GetAttribute(const Name: string): TAttribute;
begin
  FAttributesDic.TryGetValue(Name, Result);
end;

function TAttributesDictionary.GetAttribute(ID: Integer): TAttribute;
begin
  Result := FAttributes.Items[ID];
end;

procedure TAttributesDictionary.Load;
var
  i, j: int32;
  jsnValue: TJsonCinema;
  jsnDicGroups: TJsonCinema;
  jsnListAttr: TJsonCinema;
  jsnGroup: TJsonCinema;
  jsnAttr: TJsonCinema;
  group: TGroupAttributes;
  attr: TAttribute;
begin
  {$ifndef FPC}
  if TPath.IsRelativePath(FJsonFileWithAttributes) then
    jsnValue := Cinema.Utils.GetJson(TPath.GetFullPath(FJsonFileWithAttributes), true)
  else
  {$endif}
    jsnValue := Cinema.Utils.GetJson(FJsonFileWithAttributes, true);
  if not Assigned(jsnValue) then
    exit;
  try
    if jsnValue.TryGetValue<TJsonCinema>('dictionary', jsnDicGroups) then
    begin
      for i := 0 to TJSONArray(jsnDicGroups).Count - 1 do
      begin
        jsnGroup := TJSONArray(jsnDicGroups).Items[i];
        group := TGroupAttributes.Create(jsnGroup.GetValue<int32>('id'), jsnGroup.GetValue<string>('group'), jsnGroup.GetValue<string>('description'), nil);
        AddGroup(group);
        if jsnGroup.TryGetValue<TJsonCinema>('Attributes', jsnListAttr) then
        begin
          for j := 0 to TJSONArray(jsnListAttr).Count - 1 do
          begin
            jsnAttr := TJSONArray(jsnListAttr).Items[j];
            attr := TAttribute.Create(jsnAttr.GetValue<int32>('id'), TAttributeType(jsnAttr.GetValue<Integer>('type')), jsnAttr.GetValue<string>('name'), jsnAttr.GetValue<string>('description'), group);
            AddAttribute(attr);
          end;
        end;
      end;
    end;
  finally
    jsnValue.Free;
  end;

end;

class function TAttributesDictionary.StringToNative<T>(const AValue: string): T;
var
  value: TValue;
begin
  value := StrToNative(System.TypeInfo(T), AValue);
  value.ExtractRawData(@Result);
end;

class function TAttributesDictionary.StringToValue(AAttributeType: TAttributeType; const AValue: string): TValue;
begin
  Result := CONVERTERS_FROM_STR[AAttributeType](AValue);
end;

class function TAttributesDictionary.StringToValue(AAttribute: TAttribute; const AValue: string): TValue;
begin
  Result := CONVERTERS_FROM_STR[AAttribute.&Type](AValue);
end;

class function TAttributesDictionary.ValueToString(AAttribute: TAttribute; AValue: PByte; ASize: int32): string;
begin
  Result := CONVERTERS_TO_STRING[AAttribute.&Type](AValue, ASize);
end;

class function TAttributesDictionary.ValueToString(AAttributeType: TAttributeType; AValue: PByte; ASize: int32): string;
begin
  Result := CONVERTERS_TO_STRING[AAttributeType](AValue, ASize);
end;

{ TGroupAttributes }

procedure TGroupAttributes.AddAttribute(AAttribute: TAttribute);
begin
  FAttributes.Add(AAttribute);
end;

constructor TGroupAttributes.Create(AId: int32; const AName, ADesc: string; AParentGroup: TGroupAttributes);
begin
  FId := AId;
  FName := AName;
  FDescription := ADesc;
  FParentGroup := AParentGroup;
  FAttributes := TList<TAttribute>.Create;
end;

destructor TGroupAttributes.Destroy;
begin
  FAttributes.Free;
  inherited;
end;

{ TAttribute }

constructor TAttribute.Create(AId: int32; AType: TAttributeType; const AName, ADesc: string; AGroup: TGroupAttributes);
begin
  FId := AId;
  FType := AType;
  FName := AName;
  FDescription := ADesc;
  FGroup := AGroup;
end;

end.
