﻿unit Cinema.Monitor.FPC;

{$mode Delphi}

interface

implementation

uses
    SysUtils
  , syncobjs
  ;

type
  TCinemaMonitor = TCriticalSection;

var
  FpcMonitorManager: TMonitorManager;

function GetObjectMonitor(AObject: TObject): TCinemaMonitor; inline;
var
  cm: TCinemaMonitor;
begin
  Result := TCinemaMonitor(FpcMonitorManager.DoGetMonitorObjectData(AObject));
  if Result = nil then
  begin
    cm := TCinemaMonitor.Create;
    Result := FpcMonitorManager.DoSetMonitorObjectData(AObject, cm, nil);
    if Result = nil then
      Result := cm
    else
    if Result <> cm then
      cm.Free;
  end;
end;

procedure CinemaMonitorEnter(const AObject: TObject);
begin
  GetObjectMonitor(AObject).Enter;
end;

function CinemaMonitorTryEnter(const AObject: TObject): boolean;
begin
  Result := GetObjectMonitor(AObject).TryEnter;
end;

procedure CinemaMonitorExit(const AObject: TObject);
var
  cm: TCinemaMonitor;
begin
  cm := TCinemaMonitor(FpcMonitorManager.DoGetMonitorObjectData(AObject));
  if cm <> nil then
    cm.Leave;
end;

function {%H-}CinemaMonitorEnterTimeout(const AObject: TObject; ATimeout: Cardinal) : Boolean;
begin
  raise Exception.Create('TODO: CinemaMonitorEnterTimeout');
end;

procedure CinemaFreeMonitorData(AData: Pointer);
begin
  if Assigned(AData) then
    TCinemaMonitor(AData).Free;
end;

procedure InitMonitorManager;
begin
  FpcMonitorManager.DoEnter := @CinemaMonitorEnter;
  FpcMonitorManager.DoTryEnter := @CinemaMonitorTryEnter;
  FpcMonitorManager.DoExit := @CinemaMonitorExit;
  //FpcMonitorManager.DoPulse:=@CinemaMonitorPulse;
  //FpcMonitorManager.DoPulseAll:=@CinemaMonitorPulseAll;
  //FpcMonitorManager.DoEnterTimeout:=@CinemaMonitorEnterTimeout;
  //FpcMonitorManager.DoWait:=@CinemaMonitorWait;
  //FpcMonitorManager.DoSetDefaultSpinCount:=@SysNoMonitorSetDefaultSpinCount;
  //FpcMonitorManager.DoGetDefaultSpinCount:=@SysNoMonitorGetDefaultSpinCount;
  //FpcMonitorManager.DoWaitLock:=@CinemaMonitorWait;
  FpcMonitorManager.DoFreeMonitorData := @CinemaFreeMonitorData;
  SetMonitorManager(FpcMonitorManager);
end;

initialization
{$ifdef FPC}
  InitMonitorManager;
{$endif}
end.

