unit Cinema.Net.PCap;

{$ifdef FPC}
	{$mode delphi}
{$endif}

  { -*- Mode: c; tab-width: 8; indent-tabs-mode: 1; c-basic-offset: 8; -*-  }
  {
   * Copyright (c) 1993, 1994, 1995, 1996, 1997
   *	The Regents of the University of California.  All rights reserved.
   *
   * Redistribution and use in source and binary forms, with or without
   * modification, are permitted provided that the following conditions
   * are met:
   * 1. Redistributions of source code must retain the above copyright
   *    notice, this list of conditions and the following disclaimer.
   * 2. Redistributions in binary form must reproduce the above copyright
   *    notice, this list of conditions and the following disclaimer in the
   *    documentation and/or other materials provided with the distribution.
   * 3. All advertising materials mentioning features or use of this software
   *    must display the following acknowledgement:
   *	This product includes software developed by the Computer Systems
   *	Engineering Group at Lawrence Berkeley Laboratory.
   * 4. Neither the name of the University nor of the Laboratory may be used
   *    to endorse or promote products derived from this software without
   *    specific prior written permission.
   *
   * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
   * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
   * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
   * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
   * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
   * SUCH DAMAGE.
   *
   * $FreeBSD: src/contrib/libpcap/pcap.h,v 1.11 2005/07/11 03:43:25 sam Exp $
   * @(#) $Header: /tcpdump/master/libpcap/pcap.h,v 1.52 2004/12/18 08:52:11 guy Exp $ (LBL)
    }

interface

uses
  {$IFDEF MSWINDOWS}
    Windows,
    WinSock2,
    {$ifdef FPC}
    Sockets,
    {$endif}
  {$ENDIF}

  {$IFDEF POSIX}
    {$ifdef FPC}
    baseunix, Unix, Sockets, netdb,
    {$else}
    Posix.NetinetIn, Posix.ArpaInet, Posix.SysSocket, Posix.SysSelect,
    {$endif}
  {$ENDIF}
  Types
  ;

{$IFDEF FPC}
  {$PACKRECORDS C}
  {$ifndef NO_SMART_LINK}
    {$smartlink on}
  {$endif}
{$ENDIF}

const
  {$ifdef posix}
  PCAP_LIB_NAME = 'libpcap.so';
  {$endif}
  {$ifdef MSWINDOWS}
  PCAP_LIB_NAME = 'wpcap.dll';
  {$endif}
  PCAP_SRC_IF_STRING: AnsiString = 'rpcap://';
  PCAP_Optimize = 1;
  PCAP_NoOptimize = 0;
  PCAP_OK = 0;
  PCAP_ERROR = -1;
  PCAP_OPENFLAG_PROMISCUOUS = 1;
  PCAP_VERSION_MAJOR = 2;
  PCAP_VERSION_MINOR = 4;
  PCAP_ERRBUF_SIZE = 256;
  { interface is loopback  }
  PCAP_IF_LOOPBACK = $00000001;

  {
   * The first record in the file contains saved values for some
   * of the flags used in the printout phases of tcpdump.
   * Many fields here are 32 bit ints so compilers won't insert unwanted
   * padding; these files need to be interchangeable across architectures.
   *
   * Do not change the layout of this structure, in any way (this includes
   * changes that only affect the length of fields in this structure).
   *
   * Also, do not change the interpretation of any of the members of this
   * structure, in any way (this includes using values other than
   * LINKTYPE_ values, as defined in "savefile.c", in the "linktype"
   * field).
   *
   * Instead:
   *
   *	introduce a new structure for the new format, if the layout
   *	of the structure changed;
   *
   *	send mail to "tcpdump-workers@tcpdump.org", requesting a new
   *	magic number for your new capture file format, and, when
   *	you get the new magic number, put it in "savefile.c";
   *
   *	use that magic number for save files with the changed file
   *	header;
   *
   *	make the code in "savefile.c" capable of reading files with
   *	the old file header as well as files with the new file header
   *	(using the magic number to determine the header format).
   *
   * Then supply the changes to "patches@tcpdump.org", so that future
   * versions of libpcap and programs that use it (such as tcpdump) will
   * be able to read your new capture file format.
    }
  { gmt to local correction  }
  { accuracy of timestamps  }
  { max length saved portion of each pkt  }
  { data link type (LINKTYPE_*)  }
  type
    {$ifdef posix}
    timeval = record
      tv_sec: Longint;         // seconds
      tv_usec: Longint;        // and microseconds
    end;
    TTimeVal = timeval;
    PTimeVal = ^timeval;
    {$endif}

     PPLongint = ^PLongint;
     PPcap_File_Header = ^TPcap_File_Header;
     TPcap_File_Header = record
       magic : DWord;
       version_major : Word;
       version_minor : Word;
       thiszone : Longint;
       sigfigs : DWord;
       snaplen : DWord;
       linktype : DWord;
     end;

     PBPF_Insn = ^TBPF_Insn;
     TBPF_Insn = record
       code: Word;
       jt: Byte;
       jf: Byte;
       k: DWord;
     end;

     PBPF_Program = ^TBPF_Program;
     TBPF_Program = record
       bf_len: Int64;
       bf_insns: PBPF_Insn;
     end;

     PDirection = ^TDirection;
     TDirection = (D_INOUT, D_IN, D_OUT);
  {
   * Each packet in the dump file is prepended with this generic header.
   * This gets around the problem of different headers for different
   * packet interfaces.
    }
  { time stamp  }
  { length of portion present  }
  { length this packet (off wire)  }
     PPPcap_Pkthdr = ^PPcap_Pkthdr;
     PPcap_Pkthdr = ^TPcap_Pkthdr;
     TPcap_Pkthdr = record
       ts : TTimeVal;
       caplen : DWord;
       len : DWord;
     end;

  {
   * As returned by the pcap_stats()
    }
  { number of packets received  }
  { number of packets dropped  }
  { drops by interface XXX not yet supported  }

     PPcap_Stat = ^TPcap_Stat;
     TPcap_Stat = record
       ps_recv   : DWord;
       ps_drop   : DWord;
       ps_ifdrop : DWord;
       bs_capt   : DWord;
     end;

  {
   * Representation of an interface address.
    }
  { address  }
  { netmask for that address  }
  { broadcast address for that address  }
  { P2P destination address for that address  }

     PPcap_Addr = ^TPcap_Addr;
     TPcap_Addr = record
       next : PPcap_Addr;
       addr : PSockAddr;
       netmask : PSockAddr;
       broadaddr : PSockAddr;
       dstaddr : PSockAddr;
     end;

  {
   * Item in a list of interfaces.
    }
  { name to hand to "pcap_open_live()"  }
  { textual description of interface, or NULL  }
  { PCAP_IF_ interface flags  }
     PPPcap_If = ^PPcap_If;
     PPcap_If = ^TPcap_If;
     TPcap_If = record
       next : PPcap_If;
       name : PAnsiChar;
       description : PAnsiChar;
       addresses : PPcap_Addr;
       flags : DWord;
     end;
     
  { obfuscated C types }
   PPcap = ^TPcap;
   TPcap = record end;

   PPcapDumper = ^TPcapDumper;
   TPcapDumper = record end;

  TPcapHandler = procedure (para1: PAnsiChar; Header: PPcap_Pkthdr; Data: PAnsiChar); cdecl;

var
  pcap_lookupdev: function (ErrBuf: PAnsiChar): PAnsiChar; cdecl;

  pcap_lookupnet: function (Device: PAnsiChar; NetP: PLongWord; MaskP: PLongWord; ErrBuf: PAnsiChar): Longint; cdecl;

  pcap_open_live: function (Device : PAnsiChar; SnapLen: Longint; Promisc: Longint;
                          to_ms: Longint; ebuf: PAnsiChar): PPcap; cdecl;

  pcap_open_dead: function (LinkType: Longint; SnapLen: Longint): PPcap; cdecl;

  pcap_open_offline: function (FileName: PAnsiChar; ErrBuf: PAnsiChar): PPcap; cdecl;

//  function pcap_fopen_offline(para1:PFILE; para2:PAnsiChar): PPcap; cdecl;

  pcap_close: procedure (p :PPcap); cdecl;

  pcap_loop: function (p: PPcap; cnt: Longint; Callback: TPCapHandler; User: PAnsiChar): Longint; cdecl;

  pcap_dispatch: function (p: PPcap; cnt: Longint; Callback: TPCapHandler; User: PAnsiChar): Longint; cdecl;

  pcap_next: function (para1: PPcap; para2:PPcap_Pkthdr): PAnsiChar; cdecl;

  pcap_next_ex: function (para1: PPcap; para2:PPPcap_Pkthdr; para3: PPAnsiChar): Longint; cdecl;

  pcap_breakloop: procedure (para1:PPcap); cdecl;

  pcap_stats: function (para1: PPcap; para2:PPcap_Stat): Longint; cdecl;

  pcap_setfilter: function (para1: PPcap; para2:PBPF_Program): Longint; cdecl;

  pcap_setdirection: function (para1: PPcap; para2:TDirection): Longint; cdecl;

  pcap_getnonblock: function (para1: PPcap; para2:PAnsiChar): Longint; cdecl;

  pcap_setnonblock: function (para1: PPcap; para2: Longint; para3:PAnsiChar): Longint; cdecl;

  pcap_perror: procedure (para1: PPcap; para2:PAnsiChar); cdecl;

  pcap_inject: function (para1: PPcap; para2:pointer; para3: TSize): Longint; cdecl;

  pcap_sendpacket: function (para1: PPcap; para2: PAnsiChar; para3:Longint): Longint; cdecl;

  pcap_strerror: function (para1:Longint): PAnsiChar; cdecl;

  pcap_geterr: function (para1:PPcap): PAnsiChar; cdecl;

  pcap_compile: function (para1: PPcap; para2:PBPF_Program; para3: PAnsiChar; para4: Longint; para5:DWord): Longint; cdecl;

  pcap_compile_nopcap: function (para1: Longint; para2: Longint; para3:PBPF_Program; para4: PAnsiChar; para5: Longint;
             para6:DWord): Longint; cdecl;

  pcap_freecode: procedure (para1:PBPF_Program); cdecl;

  pcap_datalink: function (para1:PPcap): Longint; cdecl;

  pcap_list_datalinks: function (para1: PPcap; para2:PPLongint): Longint; cdecl;

  pcap_set_datalink: function (para1: PPcap; para2:Longint): Longint; cdecl;

  pcap_datalink_name_to_val: function (para1:PAnsiChar): Longint; cdecl;

  pcap_datalink_val_to_name: function (para1:Longint): PAnsiChar; cdecl;

  pcap_datalink_val_to_description: function (para1:Longint): PAnsiChar; cdecl;

  pcap_snapshot: function (para1:PPcap): Longint; cdecl;

  pcap_is_swapped: function (para1:PPcap): Longint; cdecl;

  pcap_major_version: function (para1:PPcap): Longint; cdecl;

  pcap_minor_version: function (para1:PPcap): Longint; cdecl;

  { XXX  }
//  function pcap_file(para1:PPcap):PFILE; cdecl;

  pcap_fileno: function (para1:PPcap): Longint; cdecl;

  pcap_dump_open: function (para1: PPcap; para2:PAnsiChar):PPCapDumper; cdecl;

//  function pcap_dump_fopen(para1: PPcap; fp:PFILE):PPCapDumper; cdecl;

//  function pcap_dump_file(para1:PPCapDumper):PFILE; cdecl;

  pcap_dump_ftell: function (para1:PPCapDumper): Longint; cdecl;

  pcap_dump_flush: function (para1:PPCapDumper): Longint; cdecl;

  pcap_dump_close: procedure (para1:PPCapDumper); cdecl;

  pcap_dump: procedure (para1: PAnsiChar; para2:PPcap_Pkthdr; para3:PAnsiChar); cdecl;

  pcap_findalldevs: function (para1: PPPcap_If; para2:PAnsiChar): Longint; cdecl;

  pcap_findalldevs_ex: function (source: PAnsichar; auth: Pointer; {pcap_rmtauth} alldevs: PPPcap_If; errbuf: PAnsiChar): Longint; cdecl;

  pcap_freealldevs: procedure (para1:PPcap_If); cdecl;

  pcap_lib_version: function : PAnsiChar; cdecl;

  { XXX this guy lives in the bpf tree  }
  bpf_filter: function (para1:Pbpf_insn; para2: PAnsiChar; para3: Int64; para4: Int64): Int64; cdecl;

  bpf_validate: function (f:Pbpf_insn; len:Longint): Longint; cdecl;

  bpf_image: function (para1:Pbpf_insn; para2:Longint): PAnsiChar; cdecl;

  bpf_dump: procedure (para1:PBPF_Program; para2:Longint); cdecl;

{$ifdef WINDOWS}

  pcap_setbuff: function(p: PPcap; dim:Longint): Longint; cdecl;

  pcap_setmode: function(p: PPcap; mode:Longint): Longint; cdecl;

  pcap_setmintocopy: function(p: PPcap; size:Longint): Longint; cdecl;

{$endif}

{$ifdef unix}

  pcap_get_selectable_fd: function (para1:PPcap): Longint; cdecl;

{$endif}

function InitPicap(const APath: string): boolean;

implementation

uses
  SysUtils;

var
  HandlePicap: THandle = 0;

function InitPicap(const APath: string): boolean;
begin
  if HandlePicap > 0 then
    exit(true);

  {$ifdef FPC}
    {$ifdef WINDOWS}
    HandlePicap := LoadLibraryW(PWideChar(APath + PCAP_LIB_NAME));
    {$else}
    HandlePicap := LoadLibrary(APath + PCAP_LIB_NAME);
    {$endif}
  {$else}
  HandlePicap := LoadLibrary(PWideChar(APath + PCAP_LIB_NAME));
  {$endif}
  //var err := GetLastError;
  if HandlePicap = 0 then
  	exit(false);

  pcap_lookupdev := GetProcAddress(HandlePicap, 'pcap_lookupdev');

  pcap_lookupnet := GetProcAddress(HandlePicap, 'pcap_lookupnet');

  pcap_open_live := GetProcAddress(HandlePicap, 'pcap_open_live');

  pcap_open_dead := GetProcAddress(HandlePicap, 'pcap_open_dead');

  pcap_open_offline := GetProcAddress(HandlePicap, 'pcap_open_offline');

  pcap_close := GetProcAddress(HandlePicap, 'pcap_close');

  pcap_loop := GetProcAddress(HandlePicap, 'pcap_loop');

  pcap_dispatch := GetProcAddress(HandlePicap, 'pcap_dispatch');

  pcap_next := GetProcAddress(HandlePicap, 'pcap_next');

  pcap_next_ex := GetProcAddress(HandlePicap, 'pcap_next_ex');

  pcap_breakloop := GetProcAddress(HandlePicap, 'pcap_breakloop');

  pcap_stats := GetProcAddress(HandlePicap, 'pcap_stats');

  pcap_setfilter := GetProcAddress(HandlePicap, 'pcap_setfilter');

  pcap_setdirection := GetProcAddress(HandlePicap, 'pcap_setdirection');

  pcap_getnonblock := GetProcAddress(HandlePicap, 'pcap_getnonblock');

  pcap_setnonblock := GetProcAddress(HandlePicap, 'pcap_setnonblock');

  pcap_perror := GetProcAddress(HandlePicap, 'pcap_perror');

  pcap_inject := GetProcAddress(HandlePicap, 'pcap_inject');

  pcap_sendpacket := GetProcAddress(HandlePicap, 'pcap_sendpacket');

  pcap_strerror := GetProcAddress(HandlePicap, 'pcap_strerror');

  pcap_geterr := GetProcAddress(HandlePicap, 'pcap_geterr');

  pcap_compile := GetProcAddress(HandlePicap, 'pcap_compile');

  pcap_compile_nopcap := GetProcAddress(HandlePicap, 'pcap_compile_nopcap');

  pcap_freecode := GetProcAddress(HandlePicap, 'pcap_freecode');

  pcap_datalink := GetProcAddress(HandlePicap, 'pcap_datalink');

  pcap_list_datalinks := GetProcAddress(HandlePicap, 'pcap_list_datalinks');

  pcap_set_datalink := GetProcAddress(HandlePicap, 'pcap_set_datalink');

  pcap_datalink_name_to_val := GetProcAddress(HandlePicap, 'pcap_datalink_name_to_val');

  pcap_datalink_val_to_name := GetProcAddress(HandlePicap, 'pcap_datalink_val_to_name');

  pcap_datalink_val_to_description := GetProcAddress(HandlePicap, 'pcap_datalink_val_to_description');

  pcap_snapshot := GetProcAddress(HandlePicap, 'pcap_snapshot');

  pcap_is_swapped := GetProcAddress(HandlePicap, 'pcap_is_swapped');

  pcap_major_version := GetProcAddress(HandlePicap, 'pcap_major_version');

  pcap_minor_version := GetProcAddress(HandlePicap, 'pcap_minor_version');

  pcap_fileno := GetProcAddress(HandlePicap, 'pcap_fileno');

  pcap_dump_open := GetProcAddress(HandlePicap, 'pcap_dump_open');

  pcap_dump_ftell := GetProcAddress(HandlePicap, 'pcap_dump_ftell');

  pcap_dump_flush := GetProcAddress(HandlePicap, 'pcap_dump_flush');

  pcap_dump_close := GetProcAddress(HandlePicap, 'pcap_dump_close');

  pcap_dump := GetProcAddress(HandlePicap, 'pcap_dump');

  pcap_findalldevs := GetProcAddress(HandlePicap, 'pcap_findalldevs');

  pcap_findalldevs_ex := GetProcAddress(HandlePicap, 'pcap_findalldevs_ex');

  pcap_freealldevs := GetProcAddress(HandlePicap, 'pcap_freealldevs');

  pcap_lib_version := GetProcAddress(HandlePicap, 'pcap_lib_version');

  { XXX this guy lives in the bpf tree  }
  bpf_filter := GetProcAddress(HandlePicap, 'bpf_filter');

  bpf_validate := GetProcAddress(HandlePicap, 'bpf_validate');

  bpf_image := GetProcAddress(HandlePicap, 'bpf_image');

  bpf_dump := GetProcAddress(HandlePicap, 'bpf_dump');

{$ifdef WINDOWS}

  pcap_setbuff := GetProcAddress(HandlePicap, 'pcap_setbuff');

  pcap_setmode := GetProcAddress(HandlePicap, 'pcap_setmode');

  pcap_setmintocopy := GetProcAddress(HandlePicap, 'pcap_setmintocopy');

{$endif}

{$ifdef unix}

  pcap_get_selectable_fd := GetProcAddress(HandlePicap, '');

{$endif}
  Result := Assigned(pcap_lookupdev) and Assigned(pcap_findalldevs);
end;

end.
