﻿unit Cinema.Scenario.View;

{$ifdef FPC}
  {$mode Delphi}
{$endif}

{ вьюшка сценария }

interface

{$if declared(FireMonkeyVersion)}
  {$define FMX}
{$endif}

uses
    Generics.Collections
  , Graphics
  , Classes
  , Types
  , Controls
  , Forms
  , ExtCtrls
  , FPCanvas
  , System.UITypes
  //, System.Math.Vectors
  , Cinema.Types
  , Cinema.Scenario.Interfaces
  ;

const
  csDefaultWorkStateColor0: array[TNodeState] of TColor = (
    TColorRec.Gray, TColorRec.Greenyellow, TColorRec.Greenyellow, TColorRec.Skyblue, TColorRec.Skyblue
  );
  csDefaultWorkStateColor1: array[TNodeState] of TAlphaColor = (
    TColorRec.Darkgray, TColorRec.Orange, TColorRec.Green, TColorRec.Skyblue, TColorRec.Mediumblue
  );

  csDefaultLogMessageColor: array[TNodeLogState] of TColor = (TColorRec.Green, TColorRec.Orange, TColorRec.Red);

type

  TColorByte = record
    class operator Multiply (const Color: TColorByte; K: Single): TColorByte;
    class operator Explicit (const Color: Cardinal): TColorByte;
    class operator Explicit (const Color: TColorByte): Cardinal;
    class operator Add (const Color1, Color2: TColorByte): TColorByte;
    class operator Subtract (const Color1, Color2: TColorByte): TColorByte;
    case byte of
    0: (Color: Cardinal);
    1: (R, G, B, A: Byte);
  end;

  TViewScenario = class;
  TViewChange = (vcSelection, vcConnectionSelection, vcChangeNode, vcPosition, vcSize, vcDelete, vcConnection, vcCreatePlugin, vcCreateLayer);

  TViewConnection = class
  private
    FPipe: IScenarioConnection;
    FHolst: TCanvas;
    FIsSelected: boolean;
    FRectOutputName: TRect;
    FRectInputName: TRect;
    // стрелка
    FTriangleTo: array of TPoint;
    class constructor Create;
    class destructor Destroy;
    procedure SetIsSelected(const Value: boolean);
    procedure CreateTriangle;
    function GetIsSourceSelected: boolean;
    function GetIsDestinationSelected: boolean;
  protected
    property IsSourceSelected: boolean read GetIsSourceSelected;
    property IsDestinationSelected: boolean read GetIsDestinationSelected;
  public
    constructor Create(AHolst: TCanvas; const APipe: IScenarioConnection); virtual;
    destructor Destroy; override;
    procedure Draw;

    property Pipe: IScenarioConnection read FPipe;
    property Holst: TCanvas read FHolst;
    property IsSelected: boolean read FIsSelected write SetIsSelected;
  end;

  { TViewNode }

  TViewNode<T: IScenarioNode> = class
  private
    FHolst: TBitmap;
    FNode: T;
    FScenarioView: TViewScenario;
    FIsSelected: boolean;
    FBeginDragPosition: TPoint;
    FIsMouseDown: boolean;
    function GetCaption: string;
    procedure SetCaption(const Value: string);
    procedure SetIsSelected(const Value: boolean);
    class constructor Create;
    class destructor Destroy;
  protected
    procedure OnChange(AObject: TObject; AViewChange: TViewChange); virtual;
    procedure DoSetIsSelected(const Value: boolean); virtual;
  public
    constructor Create(AHolst: TBitmap; const ANode: T; AViewScenario: TViewScenario);
    destructor Destroy; override;
    procedure MouseDown(X, Y: Integer; Shift: TShiftState); virtual;
    procedure MouseMove(X, Y: Integer); virtual;
    procedure MouseUp; virtual;
    procedure Drag(const DeltaPosition: TPoint);
    procedure Draw; virtual; abstract;
    procedure BeginDrag;
    property Node: T read FNode;
    property Holst: TBitmap read FHolst;
    property IsSelected: boolean read FIsSelected write SetIsSelected;
    property IsMouseDown: boolean read FIsMouseDown write FIsMouseDown;
    property ScenarioView: TViewScenario read FScenarioView;
    property Caption: string read GetCaption write SetCaption;
  end;

  TViewNodeNode = TViewNode<IScenarioNode>;

  TOnViewNodeNotify = procedure(ANode: TViewNode<IScenarioNode>) of object;

  { TViewDataStreamProcessor }

  TViewDataStreamProcessor<T: IDataStreamProcessor> = class(TViewNode<IDataStreamProcessor>)
  private
    FProcessor: T;
  protected
    procedure DoSetIsSelected(const Value: boolean); override;
  public
    constructor Create(AHolst: TBitmap; const ANode: T; AViewScenario: TViewScenario);
    destructor Destroy; override;
    procedure AfterConstruction; override;
    procedure CreatePipesView;
    procedure UpdatePipesView;
    procedure FreePipesView;

    property Processor: T read FProcessor;
  end;

  TViewDataStreamProcessor = TViewDataStreamProcessor<IDataStreamProcessor>;

  { TViewPlugin }

  TViewPlugin = class(TViewDataStreamProcessor<IScenarioPlugin>)
  private
    function GetConfig: string;
    procedure SetConfig(const Value: string);
  protected
    procedure OnChange(AObject: TObject; AViewChange: TViewChange); override;
  public
    procedure Draw; override;
    procedure MouseDown(X, Y: Integer; Shift: TShiftState); override;
    property Config: string read GetConfig write SetConfig;
  end;

  { TViewLayer }

  TViewLayer = class(TViewDataStreamProcessor<IScenarioLayer>)
  private
    FMouseDownPosition: TPoint;
    FMousePosition: TPoint;
    FIsDragging: boolean;
    function GetIsOpened: boolean;
    procedure SetIsDragging(AValue: boolean);
    procedure SetIsOpened(const AValue: boolean);
  protected
    procedure OnChange(AObject: TObject; AViewChange: TViewChange); override;
    property MouseDownPosition: TPoint read FMouseDownPosition;
    property MousePosition: TPoint read FMousePosition;
    procedure DoSetIsSelected(const Value: boolean); override;
    property IsDragging: boolean read FIsDragging write SetIsDragging;
  public
    destructor Destroy; override;
    procedure Draw; override;
    procedure Show;
    procedure Hide;
    procedure BeginDragSelected;
    procedure DragSelected(const ADeltaPosition: TPoint);
    procedure MouseDown(X, Y: Integer; Shift: TShiftState); override;
    procedure MouseMove(X, Y: Integer); override;
    procedure MouseUp; override;
    procedure ClearSelection; virtual;
    function GetChildrenAtPoint(X, Y: Integer): IScenarioNode;
    function GetPipeAtPoint(X, Y: Integer): IScenarioConnection;
    function HitToFrame(X, Y: Integer): boolean;
    procedure CheckSelectArea(const ARect: TRect);

    property IsOpened: boolean read GetIsOpened write SetIsOpened;
  end;

  TOnSelectPluginNotify = procedure(APlugin: TViewPlugin) of object;

  { TViewScenario }

  TViewScenario = class(TViewLayer)
  private
    FBitmap: TBitmap;
    FScenario: ICinemaScenario;
    //[Weak]
    FCurrentLayer: IScenarioLayer;
    FSelectedNode: IScenarioNode;
    FOnSelectPlugin: TOnSelectPluginNotify;
    FSelectPlugin: TViewPlugin;
    FComponent: TScrollBox;
    FIsModified: boolean;
    FOnChangeNode: TNotifyEvent;
    FSelectedProcessor: IDataStreamProcessor;
    FPosition: TPoint;
    FOnSelectNode: TOnViewNodeNotify;
    FOnShowLayer: TNotifyEvent;
    FUpdateCount: Integer;
    FBeforeUpdateSelectedNode: IScenarioNode;
    function GetIsMouseDown: boolean;
    procedure SetCurrentLayer(const AValue: IScenarioLayer);
    procedure SetIsModified(const AValue: boolean);
    procedure SetSelectedNode(const AValue: IScenarioNode);
    function AddNewNode(X, Y: Integer; ANodeType: TGuid): IScenarioNode;
    function GetEventListener: TEventListener;
    procedure SetEventListener(const Value: TEventListener);
    procedure SetOnSelectPlugin(const Value: TOnSelectPluginNotify);
    procedure DrawSelectedArea;
    procedure BeginUpdate;
    procedure EndUpdate;
  protected
    procedure MouseUpFromNodeSelection;
    procedure OnChange(AObject: TObject; AViewChange: TViewChange); override;
    property Bitmap: TBitmap read FBitmap;
  public
    constructor Create(AHolst: TScrollBox; const AScenario: ICinemaScenario);
    destructor Destroy; override;
    procedure BeforeDestruction; override;
    procedure Draw(const APosition: TPoint); reintroduce;
    procedure MouseDown(X, Y: Integer; Shift: TShiftState); override;
    procedure MouseMove(X, Y: Integer); override;
    procedure MouseUp; override;
    function AddPlugin(X, Y: Integer; const ACaption, AName: string; const AData: string = ''): TViewPlugin;
    function AddLayer(X, Y: Integer; const ACaption: string; IsOpened, IsThread: boolean; const AData: string = ''): TViewLayer;
    function CreateConnection(AFromProcessor, AToProcessor: TViewDataStreamProcessor<IDataStreamProcessor>; ANumber: Integer; ANumberInput: Integer): TViewConnection;
    procedure DeleteConnection(AConnection: TViewConnection);
    procedure Delete(ANode: TViewNode<IScenarioNode>);
    procedure InsertFromData(X, Y: Integer; const AData: string);

    property Scenario: ICinemaScenario read FScenario;
    property CurrentLayer: IScenarioLayer read FCurrentLayer write SetCurrentLayer;
    property SelectedProcessor: IDataStreamProcessor read FSelectedProcessor;
    property SelectPlugin: TViewPlugin read FSelectPlugin;
    property SelectedNode: IScenarioNode read FSelectedNode write SetSelectedNode;
    property IsModified: boolean read FIsModified write SetIsModified;
    property Position: TPoint read FPosition write FPosition;
    property OnSelectPlugin: TOnSelectPluginNotify read FOnSelectPlugin write SetOnSelectPlugin;
    property OnSelectNode: TOnViewNodeNotify read FOnSelectNode write FOnSelectNode;
    property OnChangeNode: TNotifyEvent read FOnChangeNode write FOnChangeNode;
    property OnShowLayer: TNotifyEvent read FOnShowLayer write FOnShowLayer;
    property EventListener: TEventListener read GetEventListener write SetEventListener;
  end;

implementation

uses
    SysUtils
  , Math
  , fpjson
  , jsonparser
  , Cinema.Utils
  , Cinema.Interfaces
  ;

{ TViewNode<T> }

class constructor TViewNode<T>.Create;
begin
end;

class destructor TViewNode<T>.Destroy;
begin
end;

procedure TViewNode<T>.BeginDrag;
begin
  FBeginDragPosition := FNode.Position^;
end;

constructor TViewNode<T>.Create(AHolst: TBitmap; const ANode: T; AViewScenario: TViewScenario);
begin
  FHolst := AHolst;
  FNode := ANode;
  if Assigned(FNode) then
    FNode.View := Self;
  FScenarioView := AViewScenario;
end;

destructor TViewNode<T>.Destroy;
begin
  IsSelected := false;
  if Assigned(FNode.View) then
    FNode.View := nil;
  FNode := nil;
  inherited;
end;

procedure TViewNode<T>.Drag(const DeltaPosition: TPoint);
var
  point: TPoint;
begin
  point := FBeginDragPosition + DeltaPosition;
  FNode.Position := @point;
  OnChange(Self, vcPosition);
end;

function TViewNode<T>.GetCaption: string;
begin
  Result := FNode.Caption;
end;

procedure TViewNode<T>.MouseDown(X, Y: Integer; Shift: TShiftState);
begin
  if not FIsMouseDown then
  begin
    FIsMouseDown := true;
  end;
end;

procedure TViewNode<T>.MouseMove(X, Y: Integer);
begin

end;

procedure TViewNode<T>.MouseUp;
begin
  FIsMouseDown := false;
end;

procedure TViewNode<T>.OnChange(AObject: TObject; AViewChange: TViewChange);
begin
  if Assigned(FNode.Parent) then
    FScenarioView.OnChange(Self, AViewChange);
end;

procedure TViewNode<T>.DoSetIsSelected(const Value: boolean);
begin
  if FIsSelected = Value then
    exit;

  FIsSelected := Value;
  OnChange(Self, vcSelection);
end;

procedure TViewNode<T>.SetCaption(const Value: string);
begin
  FNode.Caption := PWideChar(Value);
  OnChange(Self, vcChangeNode);
end;

procedure TViewNode<T>.SetIsSelected(const Value: boolean);
begin
  DoSetIsSelected(Value);
end;

{ TViewLayer }

procedure TViewLayer.BeginDragSelected;
var
  i: Integer;
  child: IScenarioNode;
begin
  for i := 0 to FProcessor.CountChildren - 1 do
  begin
    child := IScenarioNode(FProcessor.Children[i]);
    if Assigned(child.View) then
    begin
      if TViewNode<IScenarioNode>(child.View).IsSelected then
        TViewNode<IScenarioNode>(child.View).BeginDrag
      else
      if child.View.InheritsFrom(TViewLayer) and TViewLayer(child.View).IsOpened then
        TViewLayer(child.View).BeginDragSelected;
    end;
  end;
end;

procedure TViewLayer.ClearSelection;
var
  i: Integer;
  child: IScenarioNode;
begin
  for i := 0 to FProcessor.CountChildren - 1 do
  begin
    child := IScenarioNode(FProcessor.Children[i]);
    if Assigned(child.View) then
    begin
      if TViewNode<IScenarioNode>(child.View).IsSelected then
        TViewNode<IScenarioNode>(child.View).IsSelected := false;
      if child.View.InheritsFrom(TViewLayer) then
        TViewLayer(child.View).ClearSelection;
    end;
  end;
end;

destructor TViewLayer.Destroy;
begin
  Hide;
  inherited;
end;

procedure TViewLayer.DragSelected(const ADeltaPosition: TPoint);
var
  i: Integer;
  child: IScenarioNode;
begin
  for i := 0 to FProcessor.CountChildren - 1 do
  begin
    child := IScenarioNode(FProcessor.Children[i]);
    if Assigned(child.View) then
    begin
      if TViewNode<IScenarioNode>(child.View).IsSelected then
        TViewNode<IScenarioNode>(child.View).Drag(ADeltaPosition)
      else
      if child.View.InheritsFrom(TViewLayer) and TViewLayer(child.View).IsOpened then
        TViewLayer(child.View).DragSelected(ADeltaPosition);
    end;

  end;
end;

procedure TViewLayer.Draw;
var
  r, rectBorder: TRect;
  i: Integer;
  d: Integer;
  ts: TTextStyle;
begin
  inherited;
  if Assigned(FNode.Parent) then
  begin
    r := FNode.Rect^;
    // TODO: закрытый/открытый (IsThread), плюс учесть State, HealthState, состояние журнала событий

    if Processor.IsOpened then
    begin
      rectBorder := r;
      d := OPEN_LAYER_BORDER_SIZE div 2;
      rectBorder.Width := rectBorder.Width - OPEN_LAYER_BORDER_SIZE;
      rectBorder.Height := rectBorder.Height - OPEN_LAYER_BORDER_SIZE;
      rectBorder.Offset(d, d);
      Holst.Canvas.Pen.Width := OPEN_LAYER_BORDER_SIZE;
      Holst.Canvas.Pen.Color := TColorRec.Skyblue;
      Holst.Canvas.Frame(rectBorder);
      if IsSelected then
      begin
        Holst.Canvas.Pen.Color := TColorRec.Red;
        Holst.Canvas.Pen.Width := 2;
        Holst.Canvas.Frame(r);
      end;

      ts.Alignment := TAlignment.taRightJustify;
      rectBorder.Left := rectBorder.Left + d;
      rectBorder.Top := rectBorder.Top + d;
      rectBorder.Width := rectBorder.Width - d - 1;
      rectBorder.Height := rectBorder.Height - d - 1;
      Holst.Canvas.Brush.Style := TFPBrushStyle.bsClear;
      Holst.Canvas.Pen.Width := 1;
      Holst.Canvas.TextRect(rectBorder, rectBorder.Left, rectBorder.Top, FNode.Caption, ts);
    end else
    begin
      // TODO
      //FStrokeBrush.Thickness := 1;
      //Holst.Canvas.DrawRect(r, 3, 3, [TCorner.TopLeft, TCorner.TopRight, TCorner.BottomLeft, TCorner.BottomRight], 0.5, FStrokeBrush);
      //FBrush.Color := TAlphaColorRec.White;
      //Holst.Canvas.Fill := FBrush;
      //Holst.Canvas.FillText(r, FNode.Caption, true, 1.0, [TFillTextFlag.RightToLeft], TTextAlign.Center);
    end;

  end;

  for i := 0 to FProcessor.CountChildren - 1 do
  begin
    if Assigned(IScenarioNode(FProcessor.Children[i]).View) then
      TViewNodeNode(IScenarioNode(FProcessor.Children[i]).View).Draw;
  end;
end;

function TViewLayer.GetChildrenAtPoint(X, Y: Integer): IScenarioNode;
begin
  Result := IScenarioNode(FProcessor.GetChildrenAtPoint(X, Y, true));
end;

function TViewLayer.GetIsOpened: boolean;
begin
  Result := (FNode as IScenarioLayer).IsOpened;
end;

procedure TViewLayer.SetIsDragging(AValue: boolean);
begin
  if FIsDragging = AValue then
    exit;
  FIsDragging := AValue;
  if FIsDragging then
    BeginDragSelected;
end;

function TViewLayer.GetPipeAtPoint(X, Y: Integer): IScenarioConnection;
var
  i, j: Integer;
  child: IDataStreamProcessor;
  pipe: IScenarioConnection;
  node: IScenarioNode;
begin

  for i := 0 to FProcessor.CountChildren - 1 do
  begin
    node := IScenarioNode(FProcessor.Children[i]);
    if node.QueryInterface(IDataStreamProcessor, child) <> S_OK then
      continue;

    for j := 0 to child.CountInputConnections - 1 do
    begin
      pipe := IScenarioConnection(child.InputConnections[j]);
      if Assigned(pipe) and pipe.HitTest(X, Y) then
        exit(pipe);
    end;

    for j := 0 to child.CountOutputConnections - 1 do
    begin
      pipe := IScenarioConnection(child.OutputConnections[j]);
      if Assigned(pipe) and pipe.HitTest(X, Y) then
        exit(pipe);
    end;

  end;

  Result := nil;
end;

function TViewLayer.HitToFrame(X, Y: Integer): boolean;
var
  r: TRect;
begin
  r := Node.Rect^;
  Result := not IsOpened or (
    Rect(r.Left, r.Top, r.Left + OPEN_LAYER_BORDER_SIZE, R.Bottom).Contains(Point(X, Y)) or
    Rect(r.Right - OPEN_LAYER_BORDER_SIZE, r.Top, r.Right, R.Bottom).Contains(Point(X, Y)) or
    Rect(r.Left, r.Top, r.Right, r.Top + OPEN_LAYER_BORDER_SIZE).Contains(Point(X, Y)) or
    Rect(r.Left, r.Bottom - OPEN_LAYER_BORDER_SIZE, r.Right, r.Bottom).Contains(Point(X, Y))
    );
end;

procedure TViewLayer.CheckSelectArea(const ARect: TRect);
var
  i: Integer;
  child: IScenarioNode;
  r: TRect;
begin
  for i := 0 to FProcessor.CountChildren - 1 do
  begin
    child := IScenarioNode(FProcessor.Children[i]);
    if Assigned(child.View) then
    begin
      r := TViewNodeNode(child.View).Node.Rect^;
      if child.View.InheritsFrom(TViewLayer) and TViewLayer(child.View).IsOpened then
      begin
        r.Inflate(OPEN_LAYER_BORDER_SIZE, OPEN_LAYER_BORDER_SIZE, -OPEN_LAYER_BORDER_SIZE, -OPEN_LAYER_BORDER_SIZE);
        TViewLayer(child.View).IsSelected := r.IntersectsWith(ARect) and not r.Contains(ARect);
        TViewLayer(child.View).CheckSelectArea(ARect);
      end else
        TViewNodeNode(child.View).IsSelected := r.IntersectsWith(ARect);
    end;
  end;
end;

procedure TViewLayer.Hide;
var
  i: Integer;
begin
  ClearSelection;
  for i := 0 to FProcessor.CountChildren - 1 do
  begin
    if Assigned(IScenarioNode(FProcessor.Children[i]).View) then
      TObject(IScenarioNode(FProcessor.Children[i]).View).Free;
  end;

end;

procedure TViewLayer.MouseDown(X, Y: Integer; Shift: TShiftState);
var
  child: IScenarioNode;
  pipe: IScenarioConnection;
begin
  inherited;
  FMouseDownPosition := TPoint.Create(X, Y);
  if IsOpened then
  begin
    if Assigned(FScenarioView.CurrentLayer) and (FScenarioView.Scenario.Guid^ <> Self.Node.Guid^) and HitToFrame(X, Y) then
    begin
      if not IsSelected then
        IsSelected := true
      else
      if ssShift in Shift then
        IsSelected := false
      else
        IsDragging := true;
    end
    else
    begin
      if IsSelected then
        IsSelected := false;
      child := GetChildrenAtPoint(X, Y);
      if Assigned(child) then
      begin
        if Assigned(child.View) then
        begin
          if not TViewNodeNode(child.View).IsSelected then
          begin
            if not (ssShift in Shift) then
              ClearSelection;
          end
          else
            IsDragging := true;
          TViewNodeNode(child.View).MouseDown(X, Y, Shift);
        end;
      end else
      begin
        pipe := GetPipeAtPoint(X, Y);
        if Assigned(pipe) and Assigned(pipe.View) and not TViewConnection(pipe.View).IsSelected then
          TViewConnection(pipe.View).IsSelected := true
        else
        if not (ssShift in Shift) then
          ClearSelection;
      end;

    end;
  end else
  if IsSelected then
  begin
    if ssShift in Shift then
      IsSelected := false;
  end else
    IsSelected := true;
end;

procedure TViewLayer.MouseMove(X, Y: Integer);
var
  node: IScenarioNode;
begin
  inherited;
  FMousePosition := Point(X, Y);
  node := GetChildrenAtPoint(X, Y);
  if Assigned(node) and Assigned(node.View) and (node.Guid <> Processor.Guid) then
    TViewLayer(node.View).MouseMove(X, Y);
end;

procedure TViewLayer.MouseUp;
var
  i, j: Integer;
  child, parentChild: IScenarioNode;
  canChangeParent: boolean;
  globalRect: TRect;
  parentRect: TRect;
  layer: IScenarioLayer;
  parentParentLayer: IScenarioLayer;
begin
  inherited;
  IsDragging := false;
  // закрытый блок или поток куда можно перекинуть элемент
  parentParentLayer := IScenarioLayer(FProcessor.Parent);
  canChangeParent := (not FProcessor.IsOpened or FProcessor.IsThread) and Assigned(parentParentLayer);
  parentRect := FProcessor.Rect^;
  for i := FProcessor.CountChildren - 1 downto 0 do
  begin
    child := IScenarioNode(FProcessor.Children[i]);
    if Assigned(child.View) then
    begin
      if TViewNode<IScenarioNode>(child.View).IsMouseDown then
        TViewNode<IScenarioNode>(child.View).MouseUp
      else
      if child.View.InheritsFrom(TViewLayer) then
        TViewLayer(child.View).MouseUp;
    end;

    if canChangeParent then
    begin
      globalRect := child.Rect^;
      //globalRect.Offset(parentRect.Location);
      if not parentRect.Contains(globalRect.Location) then
      begin
        // поиск нового родителя
        for j := 0 to parentParentLayer.CountChildren - 1 do
        begin
          parentChild := IScenarioNode(parentParentLayer.Children[j]);
          if (parentChild = child) or not (parentChild.QueryInterface(IScenarioLayer, layer) = S_OK) then
            continue;
          if TRect(parentChild.Rect^).Contains(globalRect.Location) then
          begin
            // TODO: перекидывание во внутрь блока, когда они появятся
            child.Parent := layer;
          end;
        end;
      end;
    end;

  end;
end;

procedure TViewLayer.OnChange(AObject: TObject; AViewChange: TViewChange);
var
  i: Integer;
  node: IScenarioNode;
  processor: IDataStreamProcessor;
begin
  inherited;
  if (AViewChange = vcPosition) or (AViewChange = vcSize) then
  begin
    UpdatePipesView;
    for i := 0 to FProcessor.CountChildren - 1 do
    begin
      node := IScenarioNode(FProcessor.Children[i]);
      if Assigned(node.View) then
      begin
        if (node.QueryInterface(IDataStreamProcessor, processor) = S_OK) and Assigned(processor.View) then
        begin
          TViewDataStreamProcessor<IDataStreamProcessor>(processor.View).UpdatePipesView;
        end;
      end;
    end;
  end;
end;

procedure TViewLayer.DoSetIsSelected(const Value: boolean);
begin
  if Value = IsSelected then
    exit;
  //if Value and not HitToFrame(FMousePosition.X, FMousePosition.Y) then
  //  exit;
  inherited DoSetIsSelected(Value);
end;

procedure TViewLayer.SetIsOpened(const AValue: boolean);
var
  layer: IScenarioLayer;
begin
  if FNode.QueryInterface(IScenarioLayer, layer) = S_OK then
    layer.IsOpened := AValue;
end;

procedure TViewLayer.Show;
var
  i: Integer;
  plugin: IScenarioPlugin;
  layer: IScenarioLayer;
  viewLayer: TViewLayer;
  node: IScenarioNode;
begin
  for i := 0 to FProcessor.CountChildren - 1 do
  begin
    node := IScenarioNode(FProcessor.Children[i]);
    if not Assigned(node.View) then
    begin
      if node.QueryInterface(IScenarioPlugin, plugin) = S_OK then
      begin
        TViewPlugin.Create(FScenarioView.Bitmap, plugin, ScenarioView);
      end else
      if node.QueryInterface(IScenarioLayer, layer) = S_OK then
      begin
        viewLayer := TViewLayer.Create(FScenarioView.Bitmap, layer, ScenarioView);
        viewLayer.Show;
      end;
    end;
  end;
end;

{ TViewPlugin }

procedure TViewPlugin.Draw;
var
  r: TRect;
  i: Integer;
  wrostLogState: TNodeLogState;
  color1, color2: TColor;
  textStyle: TTextStyle;
begin
  r := FNode.Rect^;
  case FNode.State of
    nsNone: begin
      color1 := TColorRec.DarkGray;
      color2 := TColorRec.LightGray;
    end;
    nsStarting, nsWorking, nsStopping: begin
      color1 := $58CB27;
      color2 := TColorRec.Green;
    end;
    nsNoData: begin
      color1 := TColorRec.DeepSkyBlue;
      color2 := TColorRec.SkyBlue;
    end;
  end;

  wrostLogState := TNodeLogState.lsInfo;

  {%H-}case FNode.LogState of
    lsWarn: begin
      wrostLogState := TNodeLogState.lsWarn;
      color1 := TColorRec.Yellow;
      color2 := TColorRec.Orange;
    end;
    lsError: begin
      wrostLogState := TNodeLogState.lsError;
      color1 := TColorRec.IndianRed;
      color2 := TColorRec.Red;
    end;
  end;

  // TODO: градиент от зелёного к тёмно-красному
  if FNode.HealthState < 33 then
  begin
    color1 := TColorRec.IndianRed;
    color2 := TColorRec.Red;
  end else
  if FNode.HealthState < 66 then
  begin
    if wrostLogState <> TNodeLogState.lsError then
    begin
      color1 := TColorRec.Yellow;
      color2 := TColorRec.Orange;
    end;
  end;

  Holst.Canvas.Pen.Width := 1;
  Holst.Canvas.GradientFill(r, color1, color2, TGradientDirection.gdVertical);

  if IsSelected then
    Holst.Canvas.Pen.Color := TColorRec.Red
  else
    Holst.Canvas.Pen.Color := TColorRec.Black;

  Holst.Canvas.Frame(r);

  if Node.State <> TNodeState.nsNone then
    Holst.Canvas.Pen.Color := TColorRec.Black
  else
    Holst.Canvas.Pen.Color := TColorRec.White;

  textStyle.Alignment := TAlignment.taCenter;
  textStyle.Layout := TTextLayout.tlCenter;
  textStyle.Clipping := true;
  textStyle.Wordbreak := true;
  textStyle.SingleLine := false;
  Holst.Canvas.Brush.Style := TFPBrushStyle.bsClear;
  r.Left := r.Left + 2;
  r.Top := r.Top + 2;
  r.Right := r.Right - 2;
  r.Bottom := r.Bottom - 2;
  Holst.Canvas.TextRect(r, r.Left, r.Top, Node.Caption, textStyle);

  for i := 0 to FProcessor.CountOutputConnections - 1 do
    if Assigned(IScenarioConnection(FNode.OutputConnections[i])) and Assigned(IScenarioConnection(FNode.OutputConnections[i]).View) then
      TViewConnection(IScenarioConnection(FNode.OutputConnections[i]).View).Draw;
end;

procedure TViewPlugin.MouseDown(X, Y: Integer; Shift: TShiftState);
begin
  inherited MouseDown(X, Y, Shift);
  if IsSelected then
  begin
    if ssShift in Shift then
      IsSelected := false;
  end else
    IsSelected := true;
end;

function TViewPlugin.GetConfig: string;
begin
  Result := WideToString(Processor.Configuration);
end;

procedure TViewPlugin.OnChange(AObject: TObject; AViewChange: TViewChange);
begin
  inherited;
  UpdatePipesView;
end;

procedure TViewPlugin.SetConfig(const Value: string);
begin
  Processor.Configuration := PWideChar(StringToWide(Value));
  OnChange(Self, TViewChange.vcChangeNode);
end;

{ TViewScenario }

function TViewScenario.AddLayer(X, Y: Integer; const ACaption: string; IsOpened, IsThread: boolean; const AData: string = ''): TViewLayer;
var
  r: TRect;
  scenarioLayer: IScenarioLayer;
  pos: TPoint;
  guid: TGuid;
begin

  scenarioLayer := AddNewNode(X, Y, IScenarioLayer) as IScenarioLayer;
  if Assigned(scenarioLayer) then
  begin

    if AData.IsEmpty then
    begin
      guid := TGuid.NewGuid;
      scenarioLayer.GUID := @guid;
      scenarioLayer.IsOpened := IsOpened;
      scenarioLayer.IsThread := IsThread;
      scenarioLayer.Caption := PWideChar(StringToWide(ACaption));
    end else
    begin
      scenarioLayer.Load(PWideChar(StringToWide(AData)), IScenarioLayer(scenarioLayer.Parent));
    end;

    if IScenarioLayer(scenarioLayer.Parent).IsOpened then
    begin
      // возмем абсолютную площдь слоя на территории текущего просматриваемого слоя которому он принадлежит
      r := IScenarioLayer(scenarioLayer.Parent).Rect^;
      pos := Point(round(X) - r.Left, round(Y) - r.Top);
    end else
      pos := Point(round(X), round(Y));

    scenarioLayer.Position := @pos;

    Result := TViewLayer.Create(FBitmap, scenarioLayer, Self);
    OnChange(Result, TViewChange.vcCreateLayer);
  end else
    Result := nil;
end;

function TViewScenario.AddNewNode(X, Y: Integer; ANodeType: TGuid): IScenarioNode;
var
  node: IScenarioNode;
  parentLayer: IScenarioLayer;
begin
  // поищем слой куда ткнул пользователь что бы ему закинуть плагин
  node := IScenarioNode(FCurrentLayer.GetChildrenAtPoint(round(X), round(Y), true));
  if Assigned(node) then
  begin
    if (node.QueryInterface(IScenarioLayer, parentLayer) <> S_OK) or not parentLayer.IsOpened then
      parentLayer := IScenarioLayer(node.Parent);
  end else
    parentLayer := FCurrentLayer;

  Result := IScenarioNode(parentLayer.CreateChild(ANodeType));
end;

function TViewScenario.AddPlugin(X, Y: Integer; const ACaption, AName: string; const AData: string = ''): TViewPlugin;
var
  scenarioPlugin: IScenarioPlugin;
  r: TRect;
  pos: TPoint;
  guid: TGuid;
begin
  scenarioPlugin := AddNewNode(X, Y, IScenarioPlugin) as IScenarioPlugin;
  if Assigned(scenarioPlugin) then
  begin

    if not AData.IsEmpty then
    begin
      scenarioPlugin.Load(PWideChar(StringToWide(AData)), IScenarioLayer(scenarioPlugin.Parent));
    end else
    begin
      guid := TGUID.NewGuid;
      scenarioPlugin.GUID := @guid;
      scenarioPlugin.Name := PWideChar(StringToWide(AName));
      scenarioPlugin.Caption := PWideChar(StringToWide(ACaption));
    end;

    if IScenarioLayer(scenarioPlugin.Parent).IsOpened then
    begin
      // возмем абсолютную площдь слоя на территории текущего просматриваемого слоя которому он принадлежит
      r := IScenarioLayer(scenarioPlugin.Parent).Rect^;
      pos := Point(round(X) - r.Left, round(Y) - r.Top);
    end else
      pos := Point(round(X), round(Y));

    scenarioPlugin.Position := @pos;

    Result := TViewPlugin.Create(FBitmap, scenarioPlugin, Self);
    OnChange(Result, TViewChange.vcCreatePlugin);
  end else
    Result := nil;
end;

procedure TViewScenario.BeforeDestruction;
begin

  if Assigned(FCurrentLayer) then
    CurrentLayer := nil;
  inherited;
end;

constructor TViewScenario.Create(AHolst: TScrollBox; const AScenario: ICinemaScenario);
begin
  FScenario := AScenario;
  FBitmap := TBitmap.Create;
  inherited Create(FBitmap, AScenario as IScenarioLayer, Self);
  FComponent := AHolst;
  FBitmap.PixelFormat := TPixelFormat.pf32bit;
  //FBitmap.TransparentColor := clWhite;
  //FBitmap.TransparentMode := TTransparentMode.tmAuto;
  //FBitmap.Transparent := true;
end;

destructor TViewScenario.Destroy;
begin
  FBitmap.Free;
  inherited Destroy;
end;

function TViewScenario.CreateConnection(AFromProcessor, AToProcessor: TViewDataStreamProcessor<IDataStreamProcessor>; ANumber: Integer; ANumberInput: Integer): TViewConnection;
var
  conn, connDeleted: IScenarioConnection;
begin
  connDeleted := IScenarioConnection(AFromProcessor.Processor.OutputConnections[ANumber]);
  conn := IScenarioConnection(AFromProcessor.Processor.CreateOutputConnection(ANumber, ANumberInput, AToProcessor.Processor));
  if Assigned(conn) then
  begin
    Result := TViewConnection.Create(FBitmap.Canvas, conn);
    OnChange(Result, vcConnection);
  end else
  begin
    if Assigned(connDeleted) then
      connDeleted.View.Free;
    Result := nil;
  end;
end;

procedure TViewScenario.Delete(ANode: TViewNode<IScenarioNode>);
begin
  OnChange(ANode, vcDelete);
  if ANode.Node = FSelectedNode then
    SelectedNode := nil;
  ANode.Node.Parent := nil;
  ANode.Free;
end;

procedure TViewScenario.DeleteConnection(AConnection: TViewConnection);
var
  conn: IScenarioConnection;
begin
  OnChange(AConnection, vcDelete);
  conn := AConnection.Pipe;
  AConnection.Free;
  IDataStreamProcessor(conn.Source).RemoveOutputConnection(conn);
end;

procedure TViewScenario.Draw(const APosition: TPoint);
begin
  FPosition := APosition;
  FBitmap.SetSize(FComponent.ClientWidth, FComponent.ClientHeight);
  FBitmap.Canvas.Pen.Color := clWhite;
  FBitmap.Canvas.Brush.Style := TFPBrushStyle.bsSolid;
  FBitmap.Canvas.FillRect(0, 0, FBitmap.Width, FBitmap.Height);
  if Assigned(FCurrentLayer) and Assigned(FCurrentLayer.View) then
  begin
    inherited Draw;
    if IsMouseDown and not IsDragging then
      DrawSelectedArea;

    FComponent.Canvas.Draw(0, 0, FBitmap);
  end;
end;

function TViewScenario.GetEventListener: TEventListener;
begin
  if Assigned(FScenario) then
    Result := FScenario.EventListener
  else
    Result := nil;
end;

procedure TViewScenario.InsertFromData(X, Y: Integer; const AData: string);
var
  res: Boolean;
  json: TJsonCinema;
  guidStr: string;
  guid: TGuid;
begin
  json := TJSONObject.ParseJSONValue(StringToWide('{' + AData + '}'));
  try
    res := json.TryGetValue<string>('TGuid', guidStr);
    if not res then
      exit;
    guid := StringToGUID(guidStr);
    if guid = IScenarioPlugin then
      AddPlugin(X, Y, '', '', AData)
    else
    if guid = IScenarioLayer then
      AddLayer(X, Y, '', false, false, AData);
  finally
    json.Free;
  end;

end;

procedure TViewScenario.MouseDown(X, Y: Integer; Shift: TShiftState);
begin
  BeginUpdate;
  try
    inherited;
  finally
    EndUpdate;
  end;
end;

procedure TViewScenario.MouseMove(X, Y: Integer);
var
  delta: TPoint;
begin
  inherited;
  if not Assigned(FCurrentLayer) then
    exit;

  if IsMouseDown then
  begin
    if IsDragging then
    begin
      delta := Point(X - MouseDownPosition.X, Y - MouseDownPosition.Y);
      if (delta.X = 0) and (delta.Y = 0) then
        exit;

      // drag selected
      TViewLayer(FCurrentLayer.View).DragSelected(delta);
    end else
      Draw(Point(X, Y)); // перерисовка не вызывается через LCL в этом случае, поэтому отрисовываем сами
  end;

end;

procedure TViewScenario.MouseUp;
begin
  FIsDragging := false;
  inherited;
  Draw(FMousePosition); // перерисовка не вызывается через LCL в этом случае, поэтому отрисовываем сами
end;

procedure TViewScenario.MouseUpFromNodeSelection;
begin
  FIsDragging := false;
end;

procedure TViewScenario.OnChange(AObject: TObject; AViewChange: TViewChange);
begin
  inherited;

  if AViewChange = TViewChange.vcSelection then
  begin
    if Assigned(AObject) and (TViewNode<IScenarioNode>(AObject).IsSelected) then
      SelectedNode := TViewNode<IScenarioNode>(AObject).Node
    else
      SelectedNode := nil;
  end else
  if (AViewChange in [TViewChange.vcChangeNode, TViewChange.vcPosition, TViewChange.vcCreatePlugin, TViewChange.vcCreateLayer, TViewChange.vcDelete]) then
  begin
    IsModified := true;
    if Assigned(FOnChangeNode) then
      FOnChangeNode(AObject);
  end;

end;

procedure TViewScenario.SetCurrentLayer(const AValue: IScenarioLayer);
begin
  if FCurrentLayer = AValue then
    exit;

  if Assigned(FSelectedNode) and Assigned(FSelectedNode.View) then
    TViewNodeNode(FSelectedNode.View).IsSelected := false;

  if Assigned(FCurrentLayer) then
  begin
    FCurrentLayer.View.Free;
    FCurrentLayer.View := nil;
  end;

  FCurrentLayer := AValue;
  if Assigned(FCurrentLayer) then
  begin

    // пока запретим показ открытых слоёв, т.к. их внутренности и так видны
    while (FCurrentLayer.IsOpened) and Assigned(FCurrentLayer.Parent) do
      FCurrentLayer := IScenarioLayer(FCurrentLayer.Parent);

    TViewLayer.Create(FBitmap, FCurrentLayer, Self);
    TViewLayer(FCurrentLayer.View).Show;
  end;

  if Assigned(FOnShowLayer) then
    FOnShowLayer(Self);
end;

function TViewScenario.GetIsMouseDown: boolean;
begin
  if Assigned(FCurrentLayer) then
    Result := TViewNodeNode(FCurrentLayer.View).IsMouseDown
  else
    Result := false;
end;

procedure TViewScenario.SetEventListener(const Value: TEventListener);
begin
  if Assigned(FScenario) then
    FScenario.EventListener := Value;
end;

procedure TViewScenario.SetIsModified(const AValue: boolean);
begin
  if FIsModified = AValue then
    exit;
  FIsModified := AValue;
end;

procedure TViewScenario.SetOnSelectPlugin(const Value: TOnSelectPluginNotify);
begin
  FOnSelectPlugin := Value;
end;

procedure TViewScenario.DrawSelectedArea;
var
  rectSel: TRect;
begin
  //FBrush.Kind := TBrushKind.Solid;
  //FBrush.Color := TAlphaColorRec.Blue;
  rectSel := TRect.Create(min(MouseDownPosition.X, MousePosition.X), min(MouseDownPosition.Y, MousePosition.Y), max(MouseDownPosition.X, MousePosition.X), max(MouseDownPosition.Y, MousePosition.Y));

  if (rectSel.Width <= 0) or (rectSel.Height <= 0) then
    exit;
  // TODO: how to do transparent
  //Holst.Canvas.Brush.Color := TColorRec.Blue;// or $80FFFFFF);
  //Holst.Canvas.FillRect(rectSel);

  Holst.Canvas.Pen.Color := TColorRec.Blue;
  Holst.Canvas.Brush.Style := TFPBrushStyle.bsClear;
  Holst.Canvas.Pen.Width := 1;
  Holst.Canvas.Frame(rectSel);
  TViewLayer(FCurrentLayer.View).CheckSelectArea(rectSel);
end;

procedure TViewScenario.BeginUpdate;
begin
  if FUpdateCount = 0 then
  begin
    FBeforeUpdateSelectedNode := FSelectedNode;
  end;
  inc(FUpdateCount);
end;

procedure TViewScenario.EndUpdate;
begin
  dec(FUpdateCount);
  if  FUpdateCount = 0 then
  begin
    if Assigned(FOnSelectPlugin) and (FSelectedNode <> FBeforeUpdateSelectedNode) then
      FOnSelectPlugin(FSelectPlugin);
    FBeforeUpdateSelectedNode := nil;
  end;
end;

procedure TViewScenario.SetSelectedNode(const AValue: IScenarioNode);
begin
  if FSelectedNode = AValue then
    exit;

  FSelectedNode := AValue;

  if Assigned(FOnSelectNode) then
  begin
    if Assigned(FSelectedNode)  then
      FOnSelectNode(TViewNodeNode(FSelectedNode.View))
    else
      FOnSelectNode(nil);
  end;

  if Assigned(AValue) then
    AValue.QueryInterface(IDataStreamProcessor, FSelectedProcessor)
  else
    FSelectedProcessor := nil;

  if Assigned(FSelectedNode) and (TObject(FSelectedNode.View).InheritsFrom(TViewPlugin)) then
    FSelectPlugin := TViewPlugin(FSelectedNode.View)
  else
    FSelectPlugin := nil;

  if Assigned(FOnSelectPlugin) and (FUpdateCount = 0) then
    FOnSelectPlugin(FSelectPlugin);
end;

{ TViewConnection }

class constructor TViewConnection.Create;
begin
  //FStrokeBrush := TBrush.Create;
  //FStrokeBrush.Style := TFPBrushStyle.bsSolid;
  //FStrokeBrush.Color := TAlphaColorRec.Black;
  //FBrush := TBrush.Create;
  //FBrush.Style := TFPBrushStyle.bsSolid;
  //FBrush.Color := TAlphaColorRec.Black;
end;

constructor TViewConnection.Create(AHolst: TCanvas; const APipe: IScenarioConnection);
begin
  FHolst := AHolst;
  Pointer(FPipe) := Pointer(APipe);
  FPipe.View := Self;
  SetLength(FTriangleTo, 4);
  CreateTriangle;
end;

procedure TViewConnection.CreateTriangle;
const
  csTriHeight = 10;
  csTriWidth = 2;
var
  dx, dy: Integer;
  k: Double;
  s, c: Double;
  overlap: TRect;
  selected: boolean;
  name: string;
  textOutputWidth: Integer;
  nameInput: string;
  textInputWidth: Integer;
begin
  if not Assigned(FPipe.Destination) or not Assigned(FPipe.Source) then
    exit;

  overlap := RectOverlap(IDataStreamProcessor(FPipe.Source).Rect^, IDataStreamProcessor(FPipe.Destination).Rect^);
  FTriangleTo[0] := FPipe.PointDst^;
  FTriangleTo[3] := FTriangleTo[0];

  name := '';
  selected := IsSelected or IsSourceSelected or IsDestinationSelected;
  if selected then
    name := FPipe.NameOutput;

  if name.IsEmpty then
    textOutputWidth := Holst.TextWidth(IntToStr(FPipe.Number)) + 2
  else
    textOutputWidth := Holst.TextWidth(name) + 2;

  nameInput := '';
  if selected then
    nameInput := FPipe.NameInput;

  if nameInput.IsEmpty then
    textInputWidth := Holst.TextWidth(IntToStr(FPipe.NumberInput)) + 2
  else
    textInputWidth := Holst.TextWidth(nameInput) + 2;

  if (FPipe.Source = FPipe.Destination) or ((overlap.Width > 0) and (overlap.Height > 0)) then
  begin  // на самого себя или два объекта перекрываются
    FTriangleTo[1].X := FTriangleTo[0].X + csTriWidth;
    FTriangleTo[2].X := FTriangleTo[0].X - csTriWidth;
    FTriangleTo[1].Y := FTriangleTo[0].Y - csTriHeight;
    FTriangleTo[2].Y := FTriangleTo[0].Y - csTriHeight;

    FRectOutputName.Location := overlap.CenterPoint;
    FRectOutputName.Height := 14;
    if name.IsEmpty then
      FRectOutputName.Width := 32
    else
      FRectOutputName.Width := textOutputWidth;
    exit;
  end;

  if (overlap.Width > 0) then
  begin // прямая проходит параллельно оси Y

    FRectOutputName.Location := Point(FPipe.PointSrc.X - textOutputWidth - 2, FPipe.PointSrc.Y + IfThen(FPipe.PointSrc.Y < FPipe.PointDst.Y, 2, -14));
    FRectInputName.Location := Point(FPipe.PointDst.X + 2, FPipe.PointDst.Y + IfThen(FPipe.PointSrc.Y < FPipe.PointDst.Y, -14, 2));
    FTriangleTo[1].X := FTriangleTo[0].X + csTriWidth;
    FTriangleTo[2].X := FTriangleTo[0].X - csTriWidth;
    if FTriangleTo[0].Y > FPipe.PointSrc.Y then
    begin
      FTriangleTo[1].Y := FTriangleTo[0].Y - csTriHeight;
      FTriangleTo[2].Y := FTriangleTo[0].Y - csTriHeight;
    end else
    begin
      FTriangleTo[1].Y := FTriangleTo[0].Y + csTriHeight;
      FTriangleTo[2].Y := FTriangleTo[0].Y + csTriHeight;
    end;

  end else
  if (overlap.Height > 0) then
  begin // прямая проходит параллельно оси X

    FRectOutputName.Location := Point(FPipe.PointSrc.X + IfThen(FPipe.PointSrc.X < FPipe.PointDst.X, 2, -textOutputWidth), FPipe.PointSrc.Y + 1);
    FRectInputName.Location := Point(FPipe.PointDst.X + IfThen(FPipe.PointSrc.X < FPipe.PointDst.X, textInputWidth, 2), FPipe.PointDst.Y + 1);
    FTriangleTo[1].Y := FTriangleTo[0].Y + csTriWidth;
    FTriangleTo[2].Y := FTriangleTo[0].Y - csTriWidth;
    if FTriangleTo[0].X > FPipe.PointSrc.X then
    begin
      FTriangleTo[1].X := FTriangleTo[0].X - csTriHeight;
      FTriangleTo[2].X := FTriangleTo[0].X - csTriHeight;
    end else
    begin
      FTriangleTo[1].X := FTriangleTo[0].X + csTriHeight;
      FTriangleTo[2].X := FTriangleTo[0].X + csTriHeight;
    end;

  end else
  begin

    if FTriangleTo[0].X - FPipe.PointSrc.X > Epsilon then
      k := (FTriangleTo[0].Y  - FPipe.PointSrc.Y)/(FTriangleTo[0].X - FPipe.PointSrc.X)
    else
      k := 1;

    dx := Round(abs(sqrt((csTriHeight*csTriHeight) / (1 + k * k))));
    dy := Round(abs(dx * k));
    if FTriangleTo[0].Y > FPipe.PointSrc.Y then
    begin
      FTriangleTo[1].Y := FTriangleTo[0].Y - dy;
      FTriangleTo[2].Y := FTriangleTo[0].Y - dy;
    end else
    begin
      FTriangleTo[1].Y := FTriangleTo[0].Y + dy;
      FTriangleTo[2].Y := FTriangleTo[0].Y + dy;
    end;

    if FTriangleTo[0].X > FPipe.PointSrc.X then
    begin
      FTriangleTo[1].X := FTriangleTo[0].X - dx;
      FTriangleTo[2].X := FTriangleTo[0].X - dx;
    end else
    begin
      FTriangleTo[1].X := FTriangleTo[0].X + dx;
      FTriangleTo[2].X := FTriangleTo[0].X + dx;
    end;

    //тангенс угла наклона к горизонтали + угол поворота
    SinCos(2/12 + arctan2(FTriangleTo[1].Y - FTriangleTo[0].Y, FTriangleTo[1].X - FTriangleTo[0].X), s, c);
    FTriangleTo[1].Y := Round(FTriangleTo[0].Y + csTriHeight * s);
    FTriangleTo[1].X := Round(FTriangleTo[0].X + csTriHeight * c);

    SinCos(arctan2(FTriangleTo[2].Y - FTriangleTo[0].Y, FTriangleTo[2].X - FTriangleTo[0].X) - csTriWidth/csTriHeight, s, c);
    FTriangleTo[2].Y := Round(FTriangleTo[0].Y + csTriHeight * s);
    FTriangleTo[2].X := Round(FTriangleTo[0].X + csTriHeight * c);

    FRectOutputName.Location := Point(FPipe.PointSrc.X + IfThen(FPipe.PointSrc.X < FPipe.PointDst.X, 2, -textOutputWidth), FPipe.PointSrc.Y + IfThen(FPipe.PointSrc.Y < FPipe.PointDst.Y, 2, -14));
    FRectInputName.Location := Point(FPipe.PointDst.X + IfThen(FPipe.PointSrc.X < FPipe.PointDst.X, textInputWidth, 2), FPipe.PointDst.Y + IfThen(FPipe.PointSrc.Y < FPipe.PointDst.Y, -14, 2));
  end;

  FRectOutputName.Height := 14;
  FRectOutputName.Width := textOutputWidth;
  FRectInputName.Height := 14;
  FRectInputName.Width := textInputWidth;
end;

destructor TViewConnection.Destroy;
begin
  FPipe.View := nil;
  Pointer(FPipe) := Pointer(nil);
  inherited;
end;

class destructor TViewConnection.Destroy;
begin
  //FStrokeBrush.Free;
  //FBrush.Free;
end;

procedure TViewConnection.Draw;
var
  name: string;
  selected: boolean;
begin
  selected := IsSelected or IsSourceSelected or IsDestinationSelected;
  if selected then
  begin
    FHolst.Pen.Color := TColorRec.Red;
  end else
  begin
    FHolst.Pen.Color := TColorRec.Black;
  end;

  FHolst.Pen.Width := 2;
  // линия
  FHolst.Line(Point(FPipe.PointSrc.X, FPipe.PointSrc.Y), Point(FPipe.PointDst.X, FPipe.PointDst.Y));

  // стрелка
  if (FPipe.PointDst.X <> trunc(FTriangleTo[0].X)) or (FPipe.PointDst.Y <> trunc(FTriangleTo[0].Y)) then
    CreateTriangle;

  FHolst.Pen.Width := 1;
  FHolst.Polyline(FTriangleTo);

  name := '';
  if selected then
    name := FPipe.NameOutput;

  if name.IsEmpty and (FPipe.Number <> 0) then
    name := IntToStr(FPipe.Number);

  // номер/название связи
  if not name.IsEmpty then
  begin
    FHolst.Pen.Color := IfThen(selected, TColorRec.Red, TColorRec.Black);
    Holst.TextRect(FRectOutputName, FRectOutputName.Left, FRectOutputName.Top, name);
  end;

  name := '';
  if selected then
    name := FPipe.NameInput;

  if name.IsEmpty and (FPipe.NumberInput <> 0) then
    name := IntToStr(FPipe.NumberInput);

  // номер/название связи
  if not name.IsEmpty then
  begin
    FHolst.Pen.Color := IfThen(selected, TColorRec.Red, TColorRec.Black);
    Holst.TextRect(FRectInputName, FRectInputName.Left, FRectInputName.Top, name);
  end;
end;

function TViewConnection.GetIsDestinationSelected: boolean;
var
  proc: TViewDataStreamProcessor;
begin
  if not Assigned(FPipe.Destination) or not Assigned(IDataStreamProcessor(FPipe.Destination).View) then
    exit(False);
  proc := Pointer(IDataStreamProcessor(FPipe.Destination).View);

  Result := proc.IsSelected;
end;

function TViewConnection.GetIsSourceSelected: boolean;
var
  proc: TViewDataStreamProcessor;
begin
  if not Assigned(FPipe.Source) or not Assigned(IDataStreamProcessor(FPipe.Source).View) then
    exit(false);
  proc := Pointer(IDataStreamProcessor(FPipe.Source).View);
  Result := proc.IsSelected;
end;

procedure TViewConnection.SetIsSelected(const Value: boolean);
begin
  if FIsSelected = Value then
    exit;
  FIsSelected := Value;
  CreateTriangle;
  if Assigned(IDataStreamProcessor(FPipe.Source).View) then
    TViewDataStreamProcessor(IDataStreamProcessor(FPipe.Source).View).OnChange(Self, vcConnectionSelection)
  else
  if Assigned(IDataStreamProcessor(FPipe.Destination).View) then
    TViewDataStreamProcessor(IDataStreamProcessor(FPipe.Destination).View).OnChange(Self, vcConnectionSelection);
end;

{ TViewDataStreamProcessor<T> }

procedure TViewDataStreamProcessor<T>.AfterConstruction;
begin
  inherited;
  CreatePipesView;
end;

constructor TViewDataStreamProcessor<T>.Create(AHolst: TBitmap; const ANode: T; AViewScenario: TViewScenario);
begin
  inherited Create(AHolst, ANode, AViewScenario);
  FProcessor := ANode;
end;

procedure TViewDataStreamProcessor<T>.CreatePipesView;
var
  i: Integer;
  pipe: IScenarioConnection;
begin
  for i := 0 to FProcessor.CountInputConnections - 1 do
  begin
    pipe := IScenarioConnection(FProcessor.InputConnections[i]);
    if Assigned(pipe.View) or not Assigned(pipe.Source) then
      continue;

    if Assigned(IDataStreamProcessor(pipe.Source).View) then
      TViewConnection.Create(FHolst.Canvas, pipe);
  end;

  for i := 0 to FProcessor.CountOutputConnections - 1 do
  begin
    pipe := IScenarioConnection(FProcessor.OutputConnections[i]);
    if not Assigned(pipe) or Assigned(pipe.View) then
      continue;

    if Assigned(IDataStreamProcessor(pipe.Source).View) then
      TViewConnection.Create(FHolst.Canvas, pipe);
  end;
end;

procedure TViewDataStreamProcessor<T>.FreePipesView;
var
  i: Integer;
  pipe: IScenarioConnection;
begin
  for i := 0 to FNode.CountInputConnections - 1 do
  begin
    pipe := IScenarioConnection(FNode.InputConnections[i]);
    if not Assigned(pipe.View) then
      continue;

    TViewConnection(pipe.View).Free;
  end;

  for i := 0 to FNode.CountOutputConnections - 1 do
  begin
    pipe := IScenarioConnection(FNode.OutputConnections[i]);
    if not Assigned(pipe) or not Assigned(pipe.View) then
      continue;

    TViewConnection(pipe.View).Free;
  end;
end;

procedure TViewDataStreamProcessor<T>.DoSetIsSelected(const Value: boolean);
var
  i: Integer;
  pipe: IScenarioConnection;
begin
  inherited;
  for i := 0 to FNode.CountInputConnections - 1 do
  begin
    pipe := IScenarioConnection(FNode.InputConnections[i]);
    if Assigned(pipe) and Assigned(pipe.View) then
      TViewConnection(pipe.View).IsSelected := Value;
  end;
  for i := 0 to FNode.CountOutputConnections - 1 do
  begin
    pipe := IScenarioConnection(FNode.OutputConnections[i]);
    if Assigned(pipe) and Assigned(pipe.View) then
      TViewConnection(pipe.View).IsSelected := Value;
  end;
end;

procedure TViewDataStreamProcessor<T>.UpdatePipesView;
var
  i: Integer;
  pipe: IScenarioConnection;
begin

  for i := 0 to FNode.CountInputConnections - 1 do
  begin
    pipe := IScenarioConnection(FNode.InputConnections[i]);
    pipe.Calculate;
  end;

  for i := 0 to FNode.CountOutputConnections - 1 do
  begin
    pipe := IScenarioConnection(FNode.OutputConnections[i]);
    // выходное соединение с номером i может отсутствовать, поэтому необходима проверка
    if Assigned(pipe) then
      pipe.Calculate;
  end;

end;

destructor TViewDataStreamProcessor<T>.Destroy;
begin
  FreePipesView;
  inherited;
end;

{ TColorByte }

class operator TColorByte.Add(const Color1, Color2: TColorByte): TColorByte;
begin
  Result.R := min(255, IfThen(Color2.R < Color1.R, 0, Color1.R + Color2.R));
  Result.G := min(255, IfThen(Color2.G < Color1.G, 0, Color1.G + Color2.G));
  Result.B := min(255, IfThen(Color2.B < Color1.B, 0, Color1.B + Color2.B));
  Result.A := Color1.A;
end;

class operator TColorByte.Explicit(const Color: TColorByte): Cardinal;
begin
  Result := Color.Color;
end;

class operator TColorByte.Explicit(const Color: Cardinal): TColorByte;
begin
  Result.Color := Color;
end;

class operator TColorByte.Multiply(const Color: TColorByte; K: Single): TColorByte;
begin
  Result.R := round(Color.R * K);
  Result.G := round(Color.G * K);
  Result.B := round(Color.B * K);
  Result.A := Color.A;
end;

class operator TColorByte.Subtract(const Color1, Color2: TColorByte): TColorByte;
begin
  Result.R := IfThen(Color2.R > Color1.R, 0, Color1.R - Color2.R);
  Result.G := IfThen(Color2.G > Color1.G, 0, Color1.G - Color2.G);
  Result.B := IfThen(Color2.B > Color1.B, 0, Color1.B - Color2.B);
  Result.R := Color1.R;
end;

end.
