﻿unit Cinema.Scenario;

{$ifdef FPC}
	{$mode delphi}
	{$warn 4046 off}
{$endif}

interface

uses
  Classes,
  Types,
  {$ifdef FPC}
  fpjson,
  {$else}
  System.JSON,
  {$endif}
  SyncObjs,
  Generics.Collections,
  //System.Threading,  // виснет при финализации юнита
  Cinema.FPC.Threading, // альтернатива System.Threading
  Cinema.Log,
  Cinema.Types,
  Cinema.Collections,
  Cinema.Log.Interfaces,
  Cinema.Interfaces,
  Cinema.Scenario.Interfaces,
  Cinema.PluginsDictionary
  ;

type

  TDataStreamProcessor = class;
  TScenarioLayer = class;

  TSystemMessages = (smSendOnNoExistsConnection, smPluginNotFound);

  TQueueItemType = (itData, itCreateStream, itDestroyStream);
  TQueueItem = record
    QueueItemType: TQueueItemType;
    DataStream: Pointer;
  end;

  TQueueRam = TQueueRamFIFO<TQueueItem>;

  { TScenarioConnection }

  TScenarioConnection = class(TInterfacedObject, IScenarioConnection)
  private
    FOwner: TScenarioLayer;
    FSource: IDataStreamProcessor;
    FDestination: IDataStreamProcessor;
    FNumber: Integer;
    FNumberInput: Integer;
    FQueue: TQueueRam;
    FView: TObject;
    FAlignFrom: TConnectionAlign;
    FAlignTo: TConnectionAlign;
    FRect: TRect;
    FPointSrc: TPoint;
    FPointDst: TPoint;
    FStringBuffer: WideString;
    FUidSource: TGuid;
    FUidDestination: TGuid;
    FNameOutput: WideString;
    FNameInput: WideString;
    function DoGetSource: IDataStreamProcessor;
    procedure DoSetSource(const AValue: IDataStreamProcessor);
    function DoGetDestination: IDataStreamProcessor;
    procedure DoSetDestination(const AValue: IDataStreamProcessor);
  public
    constructor Create(AOwner: TScenarioLayer); virtual;
    function GetSource: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif}; stdcall;
    procedure SetSource(const AValue: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif}); stdcall;
    function GetDestination: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif}; stdcall;
    procedure SetDestination(const AValue: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif}); stdcall;
    function GetNumber: Integer; stdcall;
    procedure SetNumber(const AValue: Integer); stdcall;
    function GetNumberInput: Integer; stdcall;
    procedure SetNumberInput(const AValue: Integer); stdcall;
    function GetQueue: Pointer; stdcall;
    procedure SetQueue(const AValue: Pointer); stdcall;
    function Save: PWideChar; stdcall;
    function Load(AValues: PWideChar; const AParentLayer: IScenarioLayer): boolean; stdcall;
    function GetView: TObject; stdcall;
    procedure SetView(const AValue: TObject); stdcall;
    function HitTest(X, Y: Integer): boolean; stdcall;
    procedure Calculate; stdcall;
    function GetAlignFrom: TConnectionAlign; stdcall;
    function GetAlignTo: TConnectionAlign; stdcall;
    function GetPointDst: PPoint; stdcall;
    function GetPointSrc: PPoint; stdcall;
    function GetRect: PRect; stdcall;
    function GetNameOutput: PWideChar; stdcall;
    function GetNameInput: PWideChar; stdcall;
    procedure LoadFromJson(AValues: TJsonCinema; const AParentLayer: IScenarioLayer);

    property Source: {$ifdef NARI} Pointer read GetSource write SetSource {$else} IDataStreamProcessor read GetSource write SetSource {$endif};
    property Destination: {$ifdef NARI} Pointer read GetDestination write SetDestination {$else} IDataStreamProcessor read GetDestination write SetDestination {$endif};
    property Number: Integer read GetNumber write SetNumber;
    property NumberInput: Integer read GetNumberInput write SetNumberInput;
    property Queue: Pointer read GetQueue write SetQueue;
    property AlignFrom: TConnectionAlign read GetAlignFrom;
    property AlignTo: TConnectionAlign read GetAlignTo;
    property Rect: PRect read GetRect;
    property PointSrc: PPoint read GetPointSrc;
    property PointDst: PPoint read GetPointDst;
    property NameOutput: PWideChar read GetNameOutput;
    property NameInput: PWideChar read GetNameInput;
    property View: TObject read GetView write SetView;

  end;

  TScenarioNodeClass = class of TScenarioNode;

  { TScenarioNode }

  TScenarioNode = class(TInterfacedObject, IScenarioNode)
  strict private
    procedure DoSetParent_(const AValue: IScenarioLayer);
  private
    FOwner: TScenarioLayer;
    FCaption: WideString;
    FGuid: TGuid;
    FParent: TScenarioLayer;
    FPosition: TPoint;
    FState: TNodeState;
    FView: TObject;
    FMessagesCounter: array[TNodeLogState] of Integer;
    FNodeKey: WideString;
    FIsSaving: boolean;
    FIsLoading: boolean;
    FLogState: TNodeLogState;
    FHealthState: TNodeHalthDegreeState;
    FLogStateMessage: WideString;
    FHealthStateMessage: WideString;
    FWidth: Integer;
    FHeight: Integer;
    FRect: TRect;
  protected
    FStringBuffer: WideString;

    function DoClone(const AResult: IScenarioNode): IScenarioNode; virtual;
    function DoSendNodeMessage(ASourceNode: TScenarioNode; AMsg: WideString; AMessageType: TMessageType): boolean; virtual;
    procedure DoSetLogState(AMessage: WideString; AValue: TNodeLogState); virtual;
    procedure DoSetParent(AValue: TScenarioLayer);
    function DoGetThreadLayer: IScenarioLayer;
    function DoGetParent: IScenarioLayer;

    procedure OnChangeNodeState(ASourceNode: TScenarioNode); virtual;
    procedure UpdateNodeKey; virtual;
    procedure ResetStringBuffer; virtual;
    function GetNodeType: TGuid; virtual;
    procedure LoadFormJson(const AValues: TJsonCinema; const AParentLayer: IScenarioLayer); virtual;
    function ContainsInAnotherThread(ANode: TScenarioNode): boolean; inline;
    function DoGetRect: PRect; virtual;

    property IsLoading: boolean read FIsLoading write FIsLoading;
    property IsSaving: boolean read FIsSaving write FIsSaving;
  public
    constructor Create(AOwner: TScenarioLayer); virtual;
    destructor Destroy; override;
    function Clone(const AResult: IScenarioNode): {$ifdef NARI} Pointer {$else} IScenarioNode {$endif}; stdcall;
    function GetCaption: PWideChar; stdcall;
    procedure SetCaption(const ACaption: PWideChar); stdcall;
    function GetParent: {$ifdef NARI} Pointer {$else} IScenarioLayer {$endif}; stdcall;
    procedure SetParent(const AValue: {$ifdef NARI} Pointer {$else} IScenarioLayer {$endif}); stdcall;
    function GetGuid: System.PGuid; stdcall;
    procedure SetGuid(AValue: System.PGuid); stdcall;
    function GetState: TNodeState; stdcall;
    procedure SetState(const AValue: TNodeState); virtual; stdcall;
    procedure SendNodeMessage(AMsg: PWideChar; AMessageType: TMessageType); stdcall;
    function GetHealthState: TNodeHalthDegreeState; stdcall;
    function GetHealthStateMessage: PWideChar; stdcall;
    procedure ChangeHealthState(AMsg: PWideChar; AHealthState: TNodeHalthDegreeState); virtual; stdcall;
    function GetLogState: TNodeLogState; stdcall;
    procedure SetLogState(AMessage: PWideChar; AValue: TNodeLogState); stdcall;
    /// <summary> Установка статуса лога в состояние "Без ошибок" </summary>
    procedure ResetLogState; stdcall;
    function GetLogStateMessage: PWideChar; stdcall;
    function GetThreadLayer: {$ifdef NARI} Pointer {$else} IScenarioLayer {$endif}; stdcall;
    function GetPosition: PPoint; stdcall;
    procedure SetPosition(AValue: PPoint); virtual; stdcall;
    function Save: PWideChar; virtual; stdcall;
    function Load(AValues: PWideChar; const AParentLayer: IScenarioLayer): boolean; virtual; stdcall;
    function GetHeight: Integer; stdcall;
    function GetWidth: Integer; stdcall;
    procedure Resize(AWidth, AHeight: Integer); virtual; stdcall;
    function GetView: TObject; stdcall;
    procedure SetView(const AValue: TObject); stdcall;
    function GetRect: PRect; stdcall;
    function GetNodeKey: PWideChar; stdcall;

    property Owner: TScenarioLayer read FOwner;

    property Caption: PWideChar read GetCaption write SetCaption;
    property Parent: IScenarioLayer read DoGetParent write DoSetParent_;
    property Guid: System.PGuid read GetGuid write SetGuid;
    property State: TNodeState read GetState write SetState;
    /// <summary> Состояние лога; характеризует наличие наихудшего типа сообщения в логе </summary>
    property LogState: TNodeLogState read GetLogState;
    property LogStateMessage: PWideChar read GetLogStateMessage;
    /// <summary> Текущее качество выполняемой работы </summary>
    property HealthState: TNodeHalthDegreeState read GetHealthState;
    property HealthStateMessage: PWideChar read GetHealthStateMessage;
    property Position: PPoint read GetPosition write SetPosition;
    property Width: Integer read GetWidth;
    property Height: Integer read GetHeight;
    property Rect: PRect read GetRect;
    property NodeKey: PWideChar read GetNodeKey;
    property View: TObject read GetView write SetView;
  end;

  PRealTimeDataStream = ^TRealTimeDataStream;
  TRealTimeDataStream = record
    Parent: PRealTimeDataStream;
    NumberInput: Integer;
    Owner: TDataStreamProcessor;
    Attributes: PDataStreamAttribute;
    Processor: Pointer;
    Queue: TQueueRam;
  end;

  { TDataStreamProcessor }

  TDataStreamProcessor = class(TScenarioNode, IDataStreamProcessor)
  private
    FOutputConnections: TArray<IScenarioConnection>;
    FInputConnections: TList<IScenarioConnection>;
    FCountInputStreams: Int64;
    FCountOutputStreams: Int64;
    FCountLostBlocks: Integer;
    FCountLostData: Int64;
    FCountInputData: Int64;
    FLastTimeReqInputSpeed: UInt64;
    FLastTimeReqOutputSpeed: UInt64;
    FLastTimeProcessData: UInt64;
    FOutputSpeed: int64;
    FInputSpeed: int64;
    FQueuesItemsCount: UInt32;
    FQueuesSize: UInt32;
    procedure CalculateOutputConnections;
    procedure IncCountInputStreams;
    procedure DecCountInputStreams;
    procedure IncCountOutputStreams;
    procedure DecCountOutputStreams;
  protected
    FCountOutputData: Int64;
    property LastTimeProcessData: UInt64 read FLastTimeProcessData;
    function DoCreateDataStream(AOwner: TDataStreamProcessor; AParentStream: TDataStream; AInputNumber: Integer): PRealTimeDataStream; inline;
    function DoAddDataStreamAttribute(ADataStream: PRealTimeDataStream; AID: Integer; AData: Pointer; ADataSize: Integer): PDataStreamAttribute;
    function DoClone(const AResult: IScenarioNode): IScenarioNode; override;
    function DoGetOutputConnections(AIndex: Integer): IScenarioConnection;
    procedure DoSetOutputConnections(AIndex: Integer; const APipe: IScenarioConnection);
    function DoGetInputConnections(AIndex: Integer): IScenarioConnection;
    function DoCreateOutputConnection(ANumberConnection, ANumberInput: Integer; const ADestination: IDataStreamProcessor): IScenarioConnection;
    function DoHasOutputConnection(const ADestination: IDataStreamProcessor): IScenarioConnection;
  public
    constructor Create(AOwner: TScenarioLayer); override;
    destructor Destroy; override;

    procedure SetPosition(AValue: PPoint); override; stdcall;
    procedure Resize(AWidth, AHeight: Integer); override; stdcall;

    function GetOutputConnections(AIndex: Integer): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}; stdcall;
    procedure SetOutputConnections(AIndex: Integer; const APipe: {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}); stdcall;
    function GetCountOutputConnections: Integer; stdcall;
    function GetCountInputConnections: Integer; stdcall;
    function GetInputConnections(AIndex: Integer): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}; stdcall;
    procedure AddInputConnection(const AConnection: IScenarioConnection); stdcall;
    procedure AddOutputConnection(const AConnection: IScenarioConnection; AAsNumber: Integer); stdcall;
    procedure RemoveInputConnection(const AConnection: IScenarioConnection); stdcall;
    procedure RemoveOutputConnection(const AConnection: IScenarioConnection); stdcall;
    function CreateOutputConnection(ANumberConnection, ANumberInput: Integer; const ADestination: IDataStreamProcessor): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}; stdcall;
    function HasOutputConnection(const ADestination: IDataStreamProcessor): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif};
    function GetCountInputDataStreams: Int64; stdcall;
    function GetCountOutputDataStreams: Int64; stdcall;
    function GetInputSpeed: UInt64; stdcall;
    function GetOutputSpeed: UInt64; stdcall;
    procedure SetCountInputDataStreams(const Value: Int64); stdcall;
    procedure SetCountOutputDataStreams(const Value: Int64); stdcall;
    procedure SetInputSpeed(const Value: UInt64); stdcall;
    procedure SetOutputSpeed(const Value: UInt64); stdcall;

    function GetCountLostBlocks: Integer; stdcall;
    function GetCountLostData: Int64; stdcall;

    function CreateDataStream(AParentStream: TDataStream; AInputNumber: Integer): TDataStream; virtual; stdcall;
    procedure ProcessData(ADataStream: TDataStream; AData: PByte; ASize: Int64); virtual; stdcall;
    procedure DestroyDataStream(ADataStream: TDataStream); virtual; stdcall;

    function GetHasInputQueues: boolean; stdcall;
    function GetQueuesItemsCount: UInt32; stdcall;
    function GetQueuesSize: UInt32; stdcall;
    procedure SetQueuesItemsCount(const Value: UInt32); stdcall;
    procedure SetQueuesSize(const Value: UInt32); stdcall;

    property OutputConnections[AIndex: Integer]: IScenarioConnection read DoGetOutputConnections write DoSetOutputConnections;
    property CountOutputConnections: Integer read GetCountOutputConnections;
    property InputConnections[AIndex: Integer]: IScenarioConnection read DoGetInputConnections;
    property CountInputConnections: Integer read GetCountInputConnections;

    property CountInputDataStreams: Int64 read GetCountInputDataStreams write SetCountInputDataStreams;
    property CountOutputDataStreams: Int64 read GetCountOutputDataStreams write SetCountOutputDataStreams;

    property InputSpeed: UInt64 read GetInputSpeed write SetInputSpeed;
    property OutputSpeed: UInt64 read GetOutputSpeed write SetOutputSpeed;

    property HasInputQueues: boolean read GetHasInputQueues;
    property QueuesItemsCount: UInt32 read GetQueuesItemsCount write SetQueuesItemsCount;
    property QueuesSize: UInt32 read GetQueuesSize write SetQueuesSize;
  end;

  { TScenarioPlugin }

  TScenarioPlugin = class(TDataStreamProcessor, IScenarioPlugin, ICinema)
  private
    FName: WideString;
    FAsStart: boolean;
    FConfiguration: WideString;
    FPlugin: TPlugin;
    FCinemaPlugin: ICinemaPlugin;
    FTables: TList<IPluginTable>;
    FHavingSystemMessages: array[TSystemMessages] of boolean;
    FNeedUpdateTables: Boolean;
    FBufferForAttribute: TArray<byte>;
    FStringBuf: WideString;
    FLoadBeforeStartScenario: boolean;
    procedure AddTable(const ATable: IPluginTable);
  protected
    procedure CreatePlugin; inline;
    function GetNodeType: TGuid; override;
    procedure LoadFormJson(const AValues: TJsonCinema; const AParentLayer: IScenarioLayer); override;
    procedure DoDestroyStream(ADataStream: PRealTimeDataStream); inline;
    procedure DoSendData(ADataStream: PRealTimeDataStream; AData: PByte; ASize: Int64); inline;
    function DoGetTable(AIndex: Integer): IPluginTable;
    function DoGetCinemaPlugin: ICinemaPlugin;
  public
    constructor Create(AOwner: TScenarioLayer); override;
    destructor Destroy; override;
    procedure RequestRedrawView;
    procedure RedrawView;

    procedure SetState(const AValue: TNodeState); override; stdcall;
    function GetCountTables: Integer; stdcall;
    function GetTable(AIndex: Integer): {$ifdef NARI} Pointer {$else} IPluginTable {$endif}; stdcall;
    function GetName: PWideChar; stdcall;
    procedure SetName(const AName: PWideChar); stdcall;
    function GetConfiguration: PWideChar; stdcall;
    procedure SetConfiguration(const AValue: PWideChar); stdcall;
    function GetPluginDescription: PPluginDescription; stdcall;
    function GetLoadBeforeStartScenario: boolean; stdcall;
    procedure SetLoadBeforeStartScenario(AValue: boolean); stdcall;
    function GetAsStart: boolean; stdcall;
    procedure SetAsStart(AValue: boolean); stdcall;
    function GetCinemaPlugin: {$ifdef NARI} Pointer {$else} ICinemaPlugin {$endif}; stdcall;
    function Save: PWideChar; override; stdcall;
    function Load(AValues: PWideChar; const AParentLayer: IScenarioLayer): boolean; override; stdcall;

    function CreateDataStream(AParentStream: TDataStream; AInputNumber: Integer): TDataStream; override; stdcall;
    procedure ProcessData(ADataStream: TDataStream; AData: PByte; ASize: Int64); override; stdcall;
    procedure DestroyDataStream(ADataStream: TDataStream); override; stdcall;

    {$region 'ICinema'}
    /// <summary> метод для журналирования событий от плагина </summary>
    procedure SendMessage(AMsg: PWideChar; AMessageType: TMessageType); stdcall;
    /// <summary> текущее состояние качества выполняемой работы </summary>
    procedure SetHealthState(AMsg: PWideChar; AHealthState: TNodeHalthDegreeState); stdcall;
    /// <summary> создать таблицу для плагина </summary>
    function CreateTable(AName: PWideChar; ACaption: PWideChar): Pointer; stdcall;
    /// <summary> метод создания потока для отправки данных плагином </summary>
    function CreateStream(AConnectionNumber: Integer; AParentDataStream: TDataStream): TDataStream; virtual; stdcall;
    /// <summary> отправить данные по ранее созданному потоку </summary>
    procedure SendData(ADataStream: TDataStream; AData: PByte; ASize: Int64); virtual; stdcall;
    /// <summary> метод уничтожения потока </summary>
    procedure DestroyStream(ADataStream: TDataStream); virtual; stdcall;
    /// <summary> добавить атрибут канала </summary>
    function AddDataStreamAttribute(ADataStream: TDataStream; AID: Integer; AData: Pointer; ADataSize: Integer): PDataStreamAttribute; stdcall;
    /// <summary> найти атрибут по идентификатору </summary>
    function FindAttribute(ADataStream: TDataStream; AID: Integer): PDataStreamAttribute; stdcall;
    ///  <summary> найти атрибут по идентификатору </summary>
    function FindAttributeValuePtr(ADataStream: TDataStream; AID: Integer; ADefaultValuePtr: Pointer): Pointer; stdcall;
    ///  <summary> найти атрибут типа NativeInt по идентификатору </summary>
    function FindAttributeNativeInt(ADataStream: TDataStream; AID: Integer; ADefault: NativeInt): NativeInt; stdcall;
    ///  <summary> упаковка всех атрибутов потока в буфер </summary>
    function PackAttributes(ADataStreamSource: TDataStream; AFilters: PInteger; AHighIdInFilters: Int32; ALength: PInteger): PByte; stdcall;
    ///  <summary> распаковка всех атрибутов из буфера в поток </summary>
    procedure UnPackAttributes(ADataStreamDest: TDataStream; AAttributes: PByte; ALength: UInt32); stdcall;
    ///  <summary> получить ссылку на все атрибуты </summary>
    function GetAttributes(ADataStream: TDataStream): PDataStreamAttribute; stdcall;
    ///  <summary> выставить атрибуты из списка </summary>
    procedure SetAttributes(ADataStream: TDataStream; AAtributes: PDataStreamAttribute); stdcall;
    /// <summary> если плагин может выполнять работу без длительного возврата захваченного потока слоя,
    ///  тогда разработчик по возможности должен предусмотреть прерывание текущей работы плагина для того
    ///  что бы при желании пользователя остановить сценарий, можно было корректно завершть его выполнение;
    ///  данный метод позволяет узнать плагину что поток был прерван и необходимо завершать работу </summary>
    function IsThreadTerminated: boolean; stdcall;
    ///  <summary> имя и путь выполняемого сценария </summary>
    function GetScenarioName: PWideChar; stdcall;
    ///  <summary> поучить путь до модуля плагина </summary>
    function GetPluginPath: PWideChar; stdcall;
    ///  <summary> поучить путь до приложения в котором выполняется сценарий </summary>
    function GetAppPath: PWideChar; stdcall;
    ///  <summary> TODO </summary>
    procedure SetForceAsStart(AValue: boolean); stdcall;
    /// <summary> Получить полный путь до каталога со сценариями </summary>
    function GetRtlScenariosPath: PWideChar; stdcall;
    /// <summary> Добавить глобальный атрибут, заменяет на новое значение если такой уже есть </summary>
    procedure AddGlobalAttribute(AID: Integer; AData: Pointer; ADataSize: Integer); stdcall;
    function FindGloabalAttribute(AID: Integer): IValueHolder; stdcall;
    procedure RemoveGlobalAttribute(AID: Integer); stdcall;
    procedure AddGlobalNamedAttribute(AName: PWideChar; AData: Pointer; ADataSize: Integer); stdcall;
    function FindGloabalNamedAttribute(AName: PWideChar): IValueHolder; stdcall;
    procedure RemoveGlobalNamedAttribute(AName: PWideChar); stdcall;
    procedure ClearGlobalAttributes; stdcall;
    function GetVersionScenario: PWideChar; stdcall;
    {$endregion 'ICinema'}

    property Name: PWideChar read GetName write SetName;
    property Configuration: PWideChar read GetConfiguration write SetConfiguration;

    property CountTables: Integer read GetCountTables;
    property Table[AIndex: Integer]: IPluginTable read DoGetTable;
    property PluginDescription: PPluginDescription read GetPluginDescription;
    property NeedUpdateTables: Boolean read FNeedUpdateTables;
    property Plugin: TPlugin read FPlugin;
    property LoadBeforeStartScenario: boolean read GetLoadBeforeStartScenario write SetLoadBeforeStartScenario;
    property AsStart: boolean read GetAsStart write SetAsStart;
    property CinemaPlugin: {$ifdef NARI} Pointer {$else} ICinemaPlugin {$endif} read GetCinemaPlugin;
  end;

  TPluginTable = class(TInterfacedObject, IPluginTable)
  private
    type
      TColumn = record
        Name: WideString;
        Caption: WideString;
        ColType: TCinemaFieldType;
      end;
      // в WideString отсутствует подсчёт сылок, поэтому для уменьшения оверхэда используем его для ячеек таблицы
      TRow = TArray<WideString>;
  private
    FColumns: TList<TColumn>;
    FRows: TList<TRow>;
    FCaption: WideString;
    FName: WideString;
    FTextResult: WideString;
    FCs: SyncObjs.TCriticalSection;
    FOwner: TScenarioPlugin;
    procedure UpdateRowCountCells;
    procedure DeleteColumn(AIndex: int32);
    function DoAddColumn(const AName, ACaption: WideString; AType: TCinemaFieldType): Integer;
    procedure DoUpdateColumn(AIndex: int32; const AName, ACaption: WideString; AType: TCinemaFieldType);
    function DoAddRow: int32;
    procedure DoSetCell(AColumn: Integer; ARow: Integer; const AValue: WideString); inline;
  public
    constructor Create(AOwner: TScenarioPlugin; ACaption, AName: PWideChar);
    destructor Destroy; override;
    function GetName: PWideChar; stdcall;
    function GetCaption: PWideChar; stdcall;
    function AddColumn(AName, ACaption: PWideChar; AType: TCinemaFieldType): Integer; stdcall;
    function AddRow: Integer; stdcall;
    function GetCountColumn: Integer; stdcall;
    function GetCountRows: Integer; stdcall;
    procedure SetCell(AColumn: Integer; ARow: Integer; AValue: PWideChar); stdcall;
    function GetCell(AColumn: Integer; ARow: Integer): PWideChar; stdcall;
    function GetText: PWideChar; stdcall;
    procedure SetText(AText: PWideChar); stdcall;
    function GetJson: PWideChar; stdcall;
    function GetColumnCaption(AIndex: Integer): PWideChar; stdcall;
    procedure SetNeedReload(AValue: Boolean); stdcall;

    property Name: PWideChar read GetName;
    property Caption: PWideChar read GetCaption;
    property CountColumn: Integer read GetCountColumn;
    property CountRows: Integer read GetCountRows;
    property Text: PWideChar read GetText write SetText;
    property Json: PWideChar read GetJson;
    property NeedReload: Boolean write SetNeedReload;
  end;

  TCinemaScenario = class;

  { TScenarioLayer }

  TScenarioLayer = class(TDataStreamProcessor, IScenarioLayer)
  private
    class var FNodeImplementators: TDictionary<TGUID, TScenarioNodeClass>;
    class var FNodeJsonContainerNames: TDictionary<TScenarioNodeClass, WideString>;
    class constructor Create;
    class destructor Destroy;
  private
    type

      TWorkThread = class(TThread)
      private
        FOwner: TScenarioLayer;
      protected
        procedure Execute; override;
      public
        constructor Create(AOwner: TScenarioLayer);
        destructor Destroy; override;
      end;

      TWorkTerminationThread = class(TThread)
      private
        FOwner: TScenarioLayer;
      protected
        procedure Execute; override;
      public
        constructor Create(AOwner: TScenarioLayer);
      end;

      TQueueKey = record
        IDSource: TGuid;
        IDDest: TGuid;
      end;

  private
    FScenario: TCinemaScenario;
    FIsThread: Boolean;
    FChildren: TList<IScenarioNode>;
    FChildrenPlugins: TList<TScenarioPlugin>;
    FChildrenLayers: TList<TScenarioLayer>;
    FWorkThread: TWorkThread;
    FDictGuidChildren: TDictionary<TGuid, TInterfacedObject>;
    // стартовые плагины получающие данные из своих источников
    FStartPlugins: TList<TScenarioPlugin>;
    // плагины получающие данные из входных очередей
    FQueuePlugins: TList<TScenarioPlugin>;
    FForceStartPluginsOn: TList<TScenarioPlugin>;
    FForceStartPluginsOff: TList<TScenarioPlugin>;
    FStartOutputConnections: TList<PRealTimeDataStream>;
    FQueues: TList<TQueueRam>;
    FQueuesDict: TDictionary<TQueueKey, TQueueRam>;
    FIsOpened: boolean;
    FLog: ICinemaLog;
    FNeedUpdateTables: Boolean;
    FRequestsRedrawPlugin: TList<TScenarioPlugin>;
    FTimer: UInt64;
    FLastRedrawViewPlugins: UInt64;
    FLastUpdateNoDataState: UInt64;
    FRequareChangedStates: Boolean;
    FEventOnQueue: TEvent;
    FStopTask: ITask;
    FDateTimeStart: TDateTime;
    FLastTimeReadQueue: UInt64;
    FRectChilds: TRect;

    procedure AfterStart;
    procedure Execute;
    function GetCinemaScenario: ICinemaScenario;
    procedure DestroyStartProcessers;
    procedure DeleteQueues;
    function GetQueue(AQueueKey: TQueueKey): TQueueRam;

    procedure CalculateStates;
    procedure RedrawViewPlugins;
    procedure SendChangedStates;
    procedure SetDataWaiting(AWaiting: boolean);
    procedure OnWriteQueue(AQueue: TQueueRam);
    procedure DoStop;
    function HaveRunStopTask: boolean;
    function GetQueueSize: Integer;
    procedure FlushStartPluginsOn;
    procedure FlushStartPluginsOff;
    procedure CreateStartProcessor(APlugin: TScenarioPlugin);
    procedure DestroyStartProcessor(APlugin: TScenarioPlugin);
    procedure ServiceRoutines; inline;
  protected
    procedure OnChangeNodeState(ASourceNode: TScenarioNode); override;
    function DoSendNodeMessage(ASourceNode: TScenarioNode; AMsg: WideString; AMessageType: TMessageType): boolean; override;
    function DoGetChildren(AIndex: Integer): IScenarioNode;
    function DoGetChildByGuid(const AGuid: TGuid): IScenarioNode;
    function DoGetChildrenAtPoint(X, Y: Integer; AIncludOpenedLayers: boolean): IScenarioNode;
    function DoGetPipeAtPoint(X, Y: Integer): IScenarioConnection;
    function DoCreateChild(const ANodeType: TGUID): IScenarioNode;
    function DoGetLog: ICinemaLog;

    procedure AddChild(const AChild: TScenarioNode); virtual;
    procedure RemoveChild(const AChild: TScenarioNode); virtual;
    procedure UpdateNodeKey; override;
    procedure ResetStringBuffer; override;
    property NeedUpdateTables: Boolean read FNeedUpdateTables;
    function GetNodeType: TGuid; override;
    procedure LoadFormJson(const AValues: TJsonCinema; const AParentLayer: IScenarioLayer); override;
    procedure OnUpdateGuid(const AChild: TScenarioNode; OldGuid, NewGuid: TGuid);
    // событие установки/сброса как стартовый по инициативе плагина
    procedure OnSetForceAsStartRealtime(APlugin: TScenarioPlugin);
    property QueueSize: Integer read GetQueueSize;
  public
    constructor Create(AOwner: TScenarioLayer); override;
    destructor Destroy; override;
    class procedure RegisterNode(const ANodeType: TGUID; const ANodeImplementation: TScenarioNodeClass; const AJsonContainerName: WideString);
    procedure UpdateNoDataState;
    procedure LinkConnections;
    procedure RequestRedrawView(APlugin: TScenarioPlugin);

    procedure Lock; stdcall;
    function TryLock: boolean; stdcall;
    procedure Unlock; stdcall;
    function GetChildren(AIndex: Integer): {$ifdef NARI} Pointer {$else} IScenarioNode {$endif}; stdcall;
    function GetCountChildren: Integer; stdcall;
    function GetRectChilds: PRect; stdcall;
    function GetIsThread: Boolean; stdcall;
    procedure SetIsThread(const AValue: Boolean);  stdcall;
    function GetIsThreadTerminated: boolean; stdcall;
    function GetIsOpened: boolean; stdcall;
    procedure SetIsOpened(const AValue: boolean); stdcall;
    function GetChildByGuid(const AGuid: TGuid): {$ifdef NARI} Pointer {$else} IScenarioNode {$endif}; stdcall;
    function GetChildrenAtPoint(X, Y: Integer; AIncludOpenedLayers: boolean): {$ifdef NARI} Pointer {$else} IScenarioNode {$endif}; stdcall;
    function GetPipeAtPoint(X, Y: Integer): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}; stdcall;
    function CreateChild(const ANodeType: TGUID): {$ifdef NARI} Pointer {$else} IScenarioNode {$endif}; stdcall;
    procedure Clear; stdcall;

    function AskQueues: boolean; stdcall;
    function Save: PWideChar; override; stdcall;

    function TryStart: boolean; virtual; stdcall;
    function TryStop: boolean; virtual; stdcall;
    function TryTerminate: boolean; virtual; stdcall;

    procedure GetChangedStates; virtual; stdcall;

    function GetLog: {$ifdef NARI} Pointer {$else} ICinemaLog {$endif}; stdcall;

    property Children[AIndex: Integer]: {$ifdef NARI} Pointer {$else} IScenarioNode {$endif} read GetChildren;
    property CountChildren: Integer read GetCountChildren;
    property IsThread: Boolean read GetIsThread write SetIsThread;
    property IsOpened: boolean read GetIsOpened write SetIsOpened;
    property RectChilds: PRect read GetRectChilds;
    property Log: ICinemaLog read DoGetLog;

    property Timer: UInt64 read FTimer write FTimer;
    property Scenario: TCinemaScenario read FScenario;
  end;

  { TCinemaScenario }

  TCinemaScenario = class(TScenarioLayer, ICinemaScenario)
  private
    type
      TStateEvent = class(TInterfacedObject, IStateEvent)
      private
        FSource: WideString;
        FSourceId: TGuid;
        FState: Integer;
        FHealthState: Integer;
        FHealthStateMessage: WideString;
      public
        constructor Create(ASourceNode: TScenarioNode);
        function GetSource: PWideChar; stdcall;
        function GetSourceId: System.PGuid; stdcall;
        function GetState: Integer; stdcall;
        function GetHealthState: Integer; stdcall;
        function GetHealthStateMessage: PWideChar; stdcall;

        property Source: PWideChar read GetSource;
        property SourceId: System.PGuid read GetSourceId;
        property State: Integer read GetState;
        property HealthState: Integer read GetHealthState;
        property HealthStateMessage: PWideChar read GetHealthStateMessage;
      end;

      TMessageEvent = class (TInterfacedObject, IMessageEvent)
      private
        FMsg: WideString;
        FEventType: int32;
        FSource: WideString;
        FSourceId: TGuid;
      public
        constructor Create(ASource, AMsg: WideString; ASourceGuid: TGuid; AMessageType: TMessageType);
        function GetSource: PWideChar; stdcall;
        function GetSourceId: System.PGuid; stdcall;
        function GetMsg: PWideChar; stdcall;
        function GetEventType: Integer; stdcall;

        property Msg: PWideChar read GetMsg;
        property EventType: int32 read GetEventType;
        property Source: PWideChar read GetSource;
        property SourceId: System.PGuid read GetSourceId;
      end;

  private
    FVersion: UnicodeString;
    FDescription: UnicodeString;
    FScenarioName: UnicodeString;
    FScenarioFullPath: UnicodeString;
    FTagObject: Pointer;
    FOnCreateNode: TScenarioNodeEventNotify;
    FEventListener: TEventListener;
    FExternalEventListener: IExternalEventListener;
    FGloabalAttributes: TList<PDataStreamAttribute>;
    FGloabalNamedAttributes: TDictionary<string, IValueHolder>;
    FPluginsDictionary: TPluginsDictionary;
    function GetPluginsDictionary: TPluginsDictionary;
  protected
    procedure RemoveChild(const AChild: TScenarioNode); override;
    procedure AddChild(const AChild: TScenarioNode); override;
    procedure OnChangeNodeState(ASourceNode: TScenarioNode); override;
    function DoSendNodeMessage(ASourceNode: TScenarioNode; AMsg: WideString; AMessageType: TMessageType): boolean; override;
    function GetNodeType: TGuid; override;
    procedure LoadFormJson(const AValues: TJsonCinema; const AParentLayer: IScenarioLayer); override;
  public
    constructor Create(AOwner: TScenarioLayer); override;
    destructor Destroy; override;
    function Save: PWideChar; override; stdcall;

    function TryStart: boolean; override; stdcall;
    function TryStop: boolean; override; stdcall;
    function TryTerminate: boolean; override; stdcall;

    function LoadScenario(AFileName: PWideChar): boolean; stdcall;
    function SaveScenario(AFileName: PWideChar): boolean; stdcall;
    procedure AddPluginsSearchPath(ASearchPath: PWideChar); stdcall;
    function GetScenarioName: PWideChar; stdcall;
    function GetScenarioFullPath: PWideChar; stdcall;
    function GetEventListener: TEventListener; stdcall;
    procedure SetEventListener(const Value: TEventListener); stdcall;
    function GetExternalEventListener: IExternalEventListener; stdcall;
    procedure SetExternalEventListener(AExternalEventListener: IExternalEventListener); stdcall;
    function GetOnCreateNode: TScenarioNodeEventNotify; stdcall;
    procedure SetOnCreateNode(const AValue: TScenarioNodeEventNotify); stdcall;
    procedure GetChangedStates; override; stdcall;

    function GetTagObject: Pointer; stdcall;
    procedure SetTagObject(AValue: Pointer); stdcall;

    procedure AddGlobalAttribute(AID: Integer; AData: Pointer; ADataSize: Integer); stdcall;
    function FindGloabalAttribute(AID: Integer): IValueHolder; stdcall;
    procedure RemoveGlobalAttribute(AID: Integer); stdcall;
    procedure AddGlobalNamedAttribute(AName: PWideChar; AData: Pointer; ADataSize: Integer); stdcall;
    function FindGloabalNamedAttribute(AName: PWideChar): IValueHolder; stdcall;
    procedure RemoveGlobalNamedAttribute(AName: PWideChar); stdcall;
    procedure ClearGlobalAttributes; stdcall;

    function GetVersion: PWideChar; stdcall;
    procedure SetVersion(Value: PWideChar); stdcall;

    property ScenarioName: PWideChar read GetScenarioName;
    property ScenarioFullPath: PWideChar read GetScenarioFullPath;
    property EventListener: TEventListener read GetEventListener write SetEventListener;
    property ExternalEventListener: IExternalEventListener read GetExternalEventListener write SetExternalEventListener;
    property OnCreateNode: TScenarioNodeEventNotify read GetOnCreateNode write SetOnCreateNode;
    // <summary> Любой сторонний объект с которым ассоциируется сценарий </summary>
    property TagObject: Pointer read GetTagObject write SetTagObject;

    property PluginsDictionary: TPluginsDictionary read GetPluginsDictionary;
    property Version: PWideChar read GetVersion write SetVersion;
  end;

  TValueHolder = class(TInterfacedObject, IValueHolder)
  private
    FData: PByte;
    FSize: Integer;
  public
    constructor Create(AData: PByte; ASize: Integer);
    destructor Destroy; override;
    function GetValue: PByte; stdcall;
    function GetSize: Integer; stdcall;
    property Value: PByte read GetValue;
    property Size: Integer read GetSize;
  end;

var
  LocalHostName: WideString;
  AppPath: WideString;

implementation

uses
{$IFDEF MSWINDOWS}
  Windows,
{$ENDIF}
  SysUtils,
	Math,
{$ifdef FPC}
 	base64,
  Cinema.Monitor.FPC,
{$else}
  System.IOUtils,
  System.NetEncoding,
  System.StrUtils,
{$endif}
  Cinema.Configuration,
  Cinema.Exceptions,
  Cinema.Utils,
  Cinema.Attributes.Constants,
  Cinema.Net.SocketCommon
  ;

procedure InitLocalHost;
begin
  AppPath := WideString(ParamStr(0));
  LocalHostName := WideString(GetLocalHostName);
end;

function CreateDataStreamAttribute(AList: PDataStreamAttribute; AID: Integer; AData: Pointer; ADataSize: Integer): PDataStreamAttribute;
begin
  new(Result);
  if ADataSize > 0 then
  begin
    Result.Attribute := SysGetMem(ADataSize+2);
    Result.Attribute[ADataSize] := #0;
    Result.Attribute[ADataSize+1] := #0;
    move(PByte(AData)^, Result.Attribute^, ADataSize);
  end else
  begin
    Result.Attribute := nil;
  end;
  Result.ID := AID;
  Result.Size := ADataSize;
  if Assigned(AList) then
    Result.Next := AList
  else
    Result.Next := nil;
end;

{ TCinemaScenario }

procedure TCinemaScenario.LoadFormJson(const AValues: TJsonCinema; const AParentLayer: IScenarioLayer);
var
  uid: string;
begin
  if AValues.TryGetValue<string>('uid', uid) then
    FGuid := StringToGuid(string(uid));
  FVersion := AValues.GetValue<widestring>('version', '');
  FDescription := AValues.GetValue<widestring>('description', '');
  if FCaption = '' then
    Caption := PWideChar(WideString(ExtractFileName(ChangeFileExt(string(FScenarioFullPath), ''))));
end;

function TCinemaScenario.LoadScenario(AFileName: PWideChar): boolean; stdcall;
var
  i: int32;
  jsnValue: TJsonCinema;
  layerValue: TJsonArray;
  layer: TScenarioLayer;
  ilayer: IScenarioLayer;
  item: TJsonCinema;
begin

  FIsLoading := true;

  try
    jsnValue := GetJson(AFileName);
  except
  end;

  try
    Clear;
    FScenarioFullPath := AFileName;
    Caption := PWideChar(WideString(ChangefileExt(ExtractFileName(string(AFileName)), '')));
    if not Assigned(jsnValue) then
      exit(false);

    Result := true;
    LoadFormJson(jsnValue, nil);
    FScenarioFullPath := AFileName;
    // топовый слой
    if jsnValue.TryGetValue<TJsonArray>('layers', layerValue) then
    begin
      for i := 0 to layerValue.Count - 1 do
      begin
      	item := layerValue.Items[i];
        layer := TScenarioLayer.Create(Self);
        layer.LoadFormJson(item, Self);
      end;

      // в конце ещё раз пробегаемся по новым слоям и заставляем слинковать связи по загруженным гуидам
      for i := 0 to CountChildren - 1 do
      begin
        if Supports(DoGetChildren(i), IScenarioLayer, ilayer) then
          (ilayer as TScenarioLayer).LinkConnections;
      end;

    end else
      Result := false;

  finally
    FIsLoading := false;
    jsnValue.Free;
  end;
end;

procedure TCinemaScenario.OnChangeNodeState(ASourceNode: TScenarioNode);
var
  event: IStateEvent;
begin
  inherited;
  if Assigned(FEventListener) then
    FEventListener.OnChangeNodeState(ASourceNode);

  if Assigned(FExternalEventListener) then
  begin
    event := TStateEvent.Create(ASourceNode);
    FExternalEventListener.OnChangeNodeStateEvent(event);
  end;
end;

procedure TCinemaScenario.RemoveChild(const AChild: TScenarioNode);
begin
  inherited;

end;

procedure TCinemaScenario.RemoveGlobalAttribute(AID: Integer); stdcall;
var
  attr: PDataStreamAttribute;
begin
  if FGloabalAttributes.Count > AID then
  begin
    attr := FGloabalAttributes[AID];
    if Assigned(attr) then
    begin
      if attr.Size > 0 then
        SysFreeMem(attr.Attribute);
      dispose(attr);
    end;
    FGloabalAttributes[AID] := nil;
  end;
end;

procedure TCinemaScenario.RemoveGlobalNamedAttribute(AName: PWideChar); stdcall;
begin
  System.TMonitor.Enter(FGloabalNamedAttributes);
  try
    if FGloabalNamedAttributes.ContainsKey(AName) then
      FGloabalNamedAttributes.Remove(AName);
  finally
    System.TMonitor.Exit(FGloabalNamedAttributes);
  end;
end;

function TCinemaScenario.Save: PWideChar; stdcall;
begin
  inherited;
  FIsSaving := true;
  try
    FStringBuffer := '{' + sLineBreak +
      '"version": "' + FVersion + '",' + sLineBreak +
      '"cinema": "scenario",' + sLineBreak +
      '"description": "' + FDescription + '",' + sLineBreak +
      FStringBuffer + sLineBreak + '}';
    Result := PWideChar(FStringBuffer);
  finally
    FIsSaving := false;
  end;
end;

function TCinemaScenario.SaveScenario(AFileName: PWideChar): boolean; stdcall;
var
  s: TStringStream;
  data: string;
begin
  FScenarioFullPath := AFileName;

  if FScenarioFullPath = '' then
    exit(false);

  FIsSaving := true;
  s := TStringStream.Create;
  try
    Caption := ScenarioName;
    data := WideToString(Save);
    s.WriteString(data);
    s.SaveToFile(WideToString(FScenarioFullPath));
  finally
    s.Free;
    FIsSaving := false;
  end;
  Result := true;
end;

procedure TCinemaScenario.AddChild(const AChild: TScenarioNode);
begin
  inherited;
  if Assigned(FOnCreateNode) then
    FOnCreateNode(AChild);
end;

procedure TCinemaScenario.AddGlobalAttribute(AID: Integer; AData: Pointer; ADataSize: Integer); stdcall;
var
  attr: PDataStreamAttribute;
begin
  System.TMonitor.Enter(FGloabalAttributes);
  try
    if AID >= FGloabalAttributes.Count then
      FGloabalAttributes.Count := AID + 1;

    attr := FGloabalAttributes.Items[AID];
    if Assigned(attr) then
    begin
      if attr.Size > 0 then
        SysFreeMem(attr.Attribute);
      dispose(attr);
    end;

    FGloabalAttributes.Items[AID] := CreateDataStreamAttribute(nil, AID, AData, ADataSize);
  finally
    System.TMonitor.Exit(FGloabalAttributes);
  end;
end;

procedure TCinemaScenario.AddGlobalNamedAttribute(AName: PWideChar; AData: Pointer; ADataSize: Integer); stdcall;
begin
  System.TMonitor.Enter(FGloabalNamedAttributes);
  try
    if FGloabalNamedAttributes.ContainsKey(AName) then
      FGloabalNamedAttributes.Remove(AName);
    FGloabalNamedAttributes.Add(AName, TValueHolder.Create(AData, ADataSize));
  finally
    System.TMonitor.Exit(FGloabalNamedAttributes);
  end;
end;

procedure TCinemaScenario.AddPluginsSearchPath(ASearchPath: PWideChar); stdcall;
begin
  if not Assigned(FPluginsDictionary) then
  begin
    FPluginsDictionary := TPluginsDictionary.Create(CinemaPlugins.FileNameDictionary);
  end;
  FPluginsDictionary.AddPluginsSearchPath(ASearchPath);
end;

procedure TCinemaScenario.ClearGlobalAttributes; stdcall;
var
  attr: PDataStreamAttribute;
begin
  System.TMonitor.Enter(FGloabalAttributes);
  try
    for attr in FGloabalAttributes do
    begin
      if not Assigned(attr) then
        continue;

      if attr.Size > 0 then
        SysFreeMem(attr.Attribute);
      dispose(attr);
    end;
    FGloabalAttributes.Count := 0;
  finally
    System.TMonitor.Exit(FGloabalAttributes);
  end;

  System.TMonitor.Enter(FGloabalNamedAttributes);
  try
    FGloabalNamedAttributes.Clear;
  finally
    System.TMonitor.Exit(FGloabalNamedAttributes);
  end;
end;

constructor TCinemaScenario.Create(AOwner: TScenarioLayer);
begin
  FGloabalAttributes := TList<PDataStreamAttribute>.Create;
  FGloabalNamedAttributes := TDictionary<string, IValueHolder>.Create;
  inherited Create(AOwner);
  IsOpened := true;
  FScenario := Self;
end;

destructor TCinemaScenario.Destroy;
begin
  ClearGlobalAttributes;
  FGloabalAttributes.Free;
  FGloabalNamedAttributes.Free;
  FPluginsDictionary.Free;
  inherited;
end;

procedure TCinemaScenario.GetChangedStates; stdcall;
begin
  inherited;
end;

function TCinemaScenario.GetEventListener: TEventListener; stdcall;
begin
  Result := FEventListener;
end;

function TCinemaScenario.GetExternalEventListener: IExternalEventListener;
  stdcall;
begin
  Result := FExternalEventListener;
end;

function TCinemaScenario.GetNodeType: TGuid;
begin
  Result := ICinemaScenario;
end;

function TCinemaScenario.GetOnCreateNode: TScenarioNodeEventNotify; stdcall;
begin
  Result := FOnCreateNode;
end;

function TCinemaScenario.GetPluginsDictionary: TPluginsDictionary;
begin
  if Assigned(FPluginsDictionary) then
    Result := FPluginsDictionary
  else
    Result := CinemaPlugins;
end;

function TCinemaScenario.DoSendNodeMessage(ASourceNode: TScenarioNode; AMsg: WideString; AMessageType: TMessageType): boolean;
var
  event: IMessageEvent;
begin
  Result := inherited;
  if Result or (AMessageType > MESSAGE_ERROR) then
  begin
    if Assigned(FEventListener) then
      FEventListener.OnChangeLog(ASourceNode);

    if Assigned(FExternalEventListener) then
    begin
      event := TMessageEvent.Create(ASourceNode.Caption, AMsg, ASourceNode.GUID^, AMessageType);
      FExternalEventListener.OnAcceptNewMessageEvent(event);
    end;
  end;
end;

function TCinemaScenario.FindGloabalAttribute(AID: Integer): IValueHolder; stdcall;
var
  attr: PDataStreamAttribute;
begin
  System.TMonitor.Enter(FGloabalAttributes);
  try
    Result := nil;
    if AID < FGloabalAttributes.Count then
    begin
      attr := FGloabalAttributes.Items[AID];
      if Assigned(attr) and (attr.Size > 0) then
        Result := TValueHolder.Create(PByte(attr.Attribute), attr.Size);
    end;
  finally
    System.TMonitor.Exit(FGloabalAttributes);
  end;
end;

function TCinemaScenario.FindGloabalNamedAttribute(AName: PWideChar): IValueHolder; stdcall;
begin
  System.TMonitor.Enter(FGloabalNamedAttributes);
  try
    FGloabalNamedAttributes.TryGetValue(AName, Result);
  finally
    System.TMonitor.Exit(FGloabalNamedAttributes);
  end;
end;

procedure TCinemaScenario.SetEventListener(const Value: TEventListener); stdcall;
begin
  FEventListener := Value;
end;

procedure TCinemaScenario.SetExternalEventListener(AExternalEventListener: IExternalEventListener); stdcall;
begin
  FExternalEventListener := AExternalEventListener;
end;

procedure TCinemaScenario.SetOnCreateNode(const AValue: TScenarioNodeEventNotify); stdcall;
begin
  FOnCreateNode := AValue;
end;

procedure TCinemaScenario.SetTagObject(AValue: Pointer); stdcall;
begin
  FTagObject := AValue;
end;

procedure TCinemaScenario.SetVersion(Value: PWideChar); stdcall;
begin
  FVersion := Value;
end;

function TCinemaScenario.TryStart: boolean; stdcall;
begin
  if State <> TNodeState.nsNone then
    exit(true);
  Result := inherited;
  if Result then
  begin
    AfterStart;
    DoSendNodeMessage(Self, 'Сценарий запущен', MESSAGE_INFO)
  end;
end;

function TCinemaScenario.TryStop: boolean; stdcall;
begin
  if State = TNodeState.nsNone then
    exit(true);
  Result := inherited;
  if Result then
    DoSendNodeMessage(Self, 'Сценарий остановлен', MESSAGE_INFO);
end;

function TCinemaScenario.TryTerminate: boolean; stdcall;
begin
  if State = TNodeState.nsNone then
    exit(true);
  Result := inherited;
  if Result then
    DoSendNodeMessage(Self, 'Сценарий остановлен', MESSAGE_INFO);
end;

function TCinemaScenario.GetScenarioFullPath: PWideChar; stdcall;
begin
  Result := PWideChar(FScenarioFullPath);
end;

function TCinemaScenario.GetScenarioName: PWideChar; stdcall;
begin
  FScenarioName := ChangeFileExt(ExtractFileName(FScenarioFullPath), UnicodeString(''));
  Result := PWideChar(FScenarioName);
end;

function TCinemaScenario.GetTagObject: Pointer; stdcall;
begin
  Result := FTagObject;
end;

function TCinemaScenario.GetVersion: PWideChar; stdcall;
begin
  Result := PWideChar(FVersion);
end;

{ TScenarioNode }

function TScenarioNode.GetHealthState: TNodeHalthDegreeState; stdcall;
begin
  Result := FHealthState;
end;

function TScenarioNode.GetHealthStateMessage: PWideChar; stdcall;
begin
  Result := PWideChar(FHealthStateMessage);
end;

function TScenarioNode.GetHeight: Integer; stdcall;
begin
  Result := FHeight;
end;

function TScenarioNode.GetLogState: TNodeLogState; stdcall;
begin
  Result := FLogState;
end;

function TScenarioNode.GetLogStateMessage: PWideChar; stdcall;
begin
  Result := PWideChar(FLogStateMessage);
end;

function TScenarioNode.GetNodeKey: PWideChar; stdcall;
begin
  Result := PWideChar(FNodeKey);
end;

function TScenarioNode.GetNodeType: TGuid;
begin
  Result := IScenarioNode;
end;

function TScenarioNode.GetState: TNodeState; stdcall;
begin
  Result := FState;
end;

procedure TScenarioNode.ChangeHealthState(AMsg: PWideChar; AHealthState: TNodeHalthDegreeState); stdcall;
begin
  FHealthState := AHealthState;
  FHealthStateMessage := AMsg;
  OnChangeNodeState(Self);
end;

function TScenarioNode.Clone(const AResult: IScenarioNode): {$ifdef NARI} Pointer {$else} IScenarioNode {$endif}; stdcall;
var
  n: IScenarioNode;
begin
  n := DoClone(AResult);
  Result := {$ifdef NARI}Pointer{$endif}(n);
end;

function TScenarioNode.ContainsInAnotherThread(ANode: TScenarioNode): boolean;
begin
  Result := (Owner <> ANode.Owner) and ((Assigned(ANode.Owner) and ANode.Owner.IsThread) or Owner.IsThread);
end;

function TScenarioNode.DoGetRect: PRect;
var
  layer: TScenarioLayer;
begin
  FRect := TRect.Create(FPosition, FWidth, FHeight);
  layer := FParent;

  while Assigned(layer) and layer.IsOpened do
  begin
    FRect.Offset(layer.Position^);
    layer := layer.FParent;
  end;
  Result := @FRect;
end;

constructor TScenarioNode.Create(AOwner: TScenarioLayer);
begin
  FNodeKey := '';
  FOwner := AOwner;
  FHealthState := HEALTH_STATE_OK;
end;

destructor TScenarioNode.Destroy;
begin
  Pointer(FParent) := nil;
  inherited Destroy;
end;

function TScenarioNode.GetCaption: PWideChar; stdcall;
begin
  Result := PWideChar(FCaption);
end;

function TScenarioNode.GetGuid: System.PGuid; stdcall;
begin
  if FGuid.IsEmpty then
    FGuid := TGuid.NewGuid;
  Result := @FGuid;
end;

function TScenarioNode.GetParent: {$ifdef NARI} Pointer {$else} IScenarioLayer {$endif}; stdcall;
var
  p: IScenarioLayer;
begin
  p := DoGetParent;
  Result := {$ifdef NARI}Pointer{$endif}(p);
end;

function TScenarioNode.GetPosition: PPoint; stdcall;
begin
  Result := @FPosition;
end;

function TScenarioNode.GetRect: PRect; stdcall;
begin
  Result := DoGetRect;
end;

function TScenarioNode.GetThreadLayer: {$ifdef NARI} Pointer {$else} IScenarioLayer {$endif}; stdcall;
var
  layer: IScenarioLayer;
begin
  layer := DoGetThreadLayer;
  Result := {$ifdef NARI}Pointer{$endif}(layer);
end;

function TScenarioNode.GetView: TObject; stdcall;
begin
  Result := FView;
end;

function TScenarioNode.GetWidth: Integer; stdcall;
begin
  Result := FWidth;
end;

function TScenarioNode.Load(AValues: PWideChar; const AParentLayer: IScenarioLayer): boolean; stdcall;
var
  json: TJsonCinema;
begin
  Result := true;
  json := TJSONObject.ParseJSONValue('{' + AValues + '}');
  if Assigned(json) then
    LoadFormJson(json, AParentLayer);
end;

procedure TScenarioNode.LoadFormJson(const AValues: TJsonCinema; const AParentLayer: IScenarioLayer);
var
  uid: UnicodeString;
  g: TGuid;
begin
  FIsLoading := true;
  try
    if AValues.TryGetValue<UnicodeString>('uid', uid) then
    begin
      g := StringToGuid(string(uid));
      Guid := @g;
    end;
    FPosition.X := AValues.GetValue<Integer>('x', FPosition.X);
    FPosition.Y := AValues.GetValue<Integer>('y', FPosition.Y);
    FWidth := AValues.GetValue<Integer>('width', FWidth);
    FHeight := AValues.GetValue<Integer>('height', FHeight);
    FCaption := AValues.GetValue<WideString>('caption');
  finally
    FIsLoading := false;
  end;
end;

procedure TScenarioNode.OnChangeNodeState(ASourceNode: TScenarioNode);
begin
  if Assigned(FOwner) then
    FOwner.OnChangeNodeState(ASourceNode);
end;

procedure TScenarioNode.ResetLogState; stdcall;
var
  state: TNodeLogState;
begin
  FLogState := TNodeLogState.lsInfo;
  FLogStateMessage := '';
  for state := Low(TNodeLogState) to High(TNodeLogState) do
    FMessagesCounter[state] := 0;
  OnChangeNodeState(Self);
end;

procedure TScenarioNode.ResetStringBuffer;
begin
  FStringBuffer := '';
end;

procedure TScenarioNode.Resize(AWidth, AHeight: Integer); stdcall;
begin
  if (FWidth = AWidth) and (FHeight = AHeight) then
    exit;

  FWidth := AWidth;
  FHeight := AHeight;

end;

function TScenarioNode.Save: PWideChar; stdcall;
begin
  FIsSaving := true;
  try
    FStringBuffer :=
      WideString('"TGuid": "' + GetNodeType.ToString + '",' + sLineBreak +
      '"uid": "' + Guid.ToString + '",' + sLineBreak +
      '"caption": "' + WideToString(FCaption) + '",' + sLineBreak +
      '"x": ' + IntToStr(FPosition.X) + ',' + sLineBreak +
      '"y": ' + IntToStr(FPosition.Y) + ',' + sLineBreak +
      '"width": ' + IntToStr(FWidth) + ',' + sLineBreak +
      '"height": ' + IntToStr(height))
      ;
  finally
    FIsSaving := false;
  end;
  Result := PWideChar(FStringBuffer);
end;

procedure TScenarioNode.DoSetParent_(const AValue: IScenarioLayer);
begin
  DoSetParent(AValue as TScenarioLayer);
end;

function TScenarioNode.DoClone(const AResult: IScenarioNode): IScenarioNode;
begin
  AResult.Caption := Caption;
  Result := AResult;
end;

function TScenarioNode.DoSendNodeMessage(ASourceNode: TScenarioNode;
  AMsg: WideString; AMessageType: TMessageType): boolean;
begin
  Result := false;

  if (AMessageType < 0) or (AMessageType > MESSAGE_ERROR) then
  begin
    if Assigned(FParent) then
      if (FParent as TScenarioNode).DoSendNodeMessage(ASourceNode, AMsg, AMessageType) then
        Result := true;
    exit;
  end;

  if ContainsInAnotherThread(ASourceNode) then
    TMonitor.Enter(Self);

  try

    inc(FMessagesCounter[TNodeLogState(AMessageType)]);

    // если сообщение нового типа и оно по типу хуже текущего статуса лога - меняем на новый статус, т.е.
    // вегда выбираем первое наихудшее сообщение
    if (FMessagesCounter[TNodeLogState(AMessageType)] = 1) and (FLogState < TNodeLogState(AMessageType)) then
    begin
      Result := true;
      DoSetLogState(AMsg, TNodeLogState(AMessageType));
    end;

  finally
    if ContainsInAnotherThread(ASourceNode) then
      System.TMonitor.Exit(Self);
  end;

  if Assigned(FParent) then
    if (FParent as TScenarioNode).DoSendNodeMessage(ASourceNode, AMsg, AMessageType) then
      Result := true;
end;

procedure TScenarioNode.DoSetLogState(AMessage: WideString; AValue: TNodeLogState);
begin
  FLogStateMessage := AMessage;
  FLogState := AValue;
  if (MSG_TYPE_TO_HEALTH_DEGREE[AValue] < FHealthState) then
    ChangeHealthState(PWideChar(AMessage), MSG_TYPE_TO_HEALTH_DEGREE[AValue])
  else
    OnChangeNodeState(Self);
end;

procedure TScenarioNode.DoSetParent(AValue: TScenarioLayer);
begin
  if AValue = FParent then
    exit;

  if Assigned(FParent) then
  begin
    FParent.RemoveChild(Self);
    if (FParent = FOwner) and Assigned(AValue) then
      FOwner := AValue as TScenarioLayer;
  end;

  FParent := AValue;

  if Assigned(FParent) then
    FParent.AddChild(Self);

  UpdateNodeKey;
end;

function TScenarioNode.DoGetThreadLayer: IScenarioLayer;
var
  layerI: IScenarioLayer;
  layer: TScenarioLayer;
begin
  Result := nil;
  if Supports(Self, IScenarioLayer, layerI) then
    layer := layerI as TScenarioLayer
  else
    layer := FParent;
  while Assigned(layer) do
  begin
    if layer.IsThread then
      exit(layer);
    layer := layer.FParent;
  end;
end;

function TScenarioNode.DoGetParent: IScenarioLayer;
begin
  Result := FParent;
end;

procedure TScenarioNode.SendNodeMessage(AMsg: PWideChar; AMessageType: TMessageType); stdcall;
begin
  DoSendNodeMessage(Self, AMsg, AMessageType);
end;

procedure TScenarioNode.SetCaption(const ACaption: PWideChar); stdcall;
begin
  FCaption := ACaption;
  UpdateNodeKey;
end;

procedure TScenarioNode.SetGuid(AValue: System.PGuid); stdcall;
begin
  FOwner.OnUpdateGuid(Self, FGuid, AValue^);
  FGuid := AValue^;
end;

procedure TScenarioNode.SetLogState(AMessage: PWideChar; AValue: TNodeLogState); stdcall;
begin
  if FLogState = AValue then
    exit;
  DoSetLogState(AMessage, AValue);
end;

procedure TScenarioNode.SetState(const AValue: TNodeState); stdcall;
begin
  if FState = AValue then
    exit;
  FState := AValue;
  OnChangeNodeState(Self);
end;

procedure TScenarioNode.SetParent(const AValue: {$ifdef NARI} Pointer {$else} IScenarioLayer {$endif}); stdcall;
begin
  DoSetParent_({$ifdef NARI}IScenarioLayer{$endif}(AValue));
end;

procedure TScenarioNode.SetPosition(AValue: PPoint); stdcall;
begin
  if AValue^ = FPosition then
    exit;
  FPosition := AValue^;
end;

procedure TScenarioNode.SetView(const AValue: TObject); stdcall;
begin
  if FView = AValue then
    exit;
  FView := AValue;
end;

procedure TScenarioNode.UpdateNodeKey;
begin
  if Assigned(FParent) then
    FNodeKey := FParent.NodeKey + '/' + FCaption
  else
    FNodeKey := FCaption;
end;

{ TDataStreamProcessor }

function TDataStreamProcessor.GetCountInputConnections: Integer; stdcall;
begin
  Result := FInputConnections.Count;
end;

function TDataStreamProcessor.GetCountInputDataStreams: Int64; stdcall;
begin
  Result := FCountInputStreams;
end;

function TDataStreamProcessor.GetInputSpeed: UInt64; stdcall;
begin
  Result := FInputSpeed;
  if TThread.GetTickCount64 - FLastTimeReqInputSpeed >= 3000 then
    FInputSpeed := 0;
end;

function TDataStreamProcessor.GetCountLostBlocks: Integer; stdcall;
begin
  Result := FCountLostBlocks;
end;

function TDataStreamProcessor.GetCountLostData: Int64; stdcall;
begin
  Result := FCountLostData;
end;

function TDataStreamProcessor.GetCountOutputDataStreams: Int64; stdcall;
begin
  Result := FCountOutputStreams;
end;

function TDataStreamProcessor.GetHasInputQueues: boolean; stdcall;
var
  i: Integer;
  conn: IScenarioConnection;
  layerSrc: TScenarioLayer;
  layerDst: TScenarioLayer;
begin
  Result := false;
  for i := 0 to FInputConnections.Count - 1 do
  begin
    conn := FInputConnections.Items[i];
    if not Assigned(conn.Source) or not Assigned(conn.Destination) then
      continue;
    layerSrc := IScenarioLayer((IDataStreamProcessor(conn.Source) as TDataStreamProcessor).GetThreadLayer) as TScenarioLayer;
    layerDst := IScenarioLayer((IDataStreamProcessor(conn.Destination) as TDataStreamProcessor).GetThreadLayer) as TScenarioLayer;
    if Assigned(layerSrc) and Assigned(layerDst) and (layerSrc.GUID <> layerDst.GUID) and (layerSrc.IsThread and layerDst.IsThread) then
    begin
      Result := true;
      break;
    end;
  end;
end;

function TDataStreamProcessor.GetOutputSpeed: UInt64; stdcall;
begin
  Result := FOutputSpeed;
  if TThread.GetTickCount64 - FLastTimeReqOutputSpeed >= 3000 then
    FOutputSpeed := 0;
end;

function TDataStreamProcessor.GetQueuesItemsCount: UInt32; stdcall;
begin
  Result := FQueuesItemsCount;
end;

function TDataStreamProcessor.GetQueuesSize: UInt32; stdcall;
begin
  Result := FQueuesSize;
end;

function TDataStreamProcessor.HasOutputConnection(const ADestination: IDataStreamProcessor): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif};
var
  conn: IScenarioConnection;
begin
  conn := DoHasOutputConnection(ADestination);
  Result := {$ifdef NARI}Pointer{$endif}(conn);
end;

function TDataStreamProcessor.GetCountOutputConnections: Integer; stdcall;
begin
  Result := Length(FOutputConnections);
end;

function TDataStreamProcessor.GetInputConnections(AIndex: Integer): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}; stdcall;
var
  conn: IScenarioConnection;
begin
  conn := DoGetInputConnections(AIndex);
  Result := {$ifdef NARI}Pointer{$endif}(conn);
end;

procedure TDataStreamProcessor.AddInputConnection(const AConnection: IScenarioConnection); stdcall;
begin
  FInputConnections.Add(AConnection);
end;

procedure TDataStreamProcessor.AddOutputConnection(const AConnection: IScenarioConnection; AAsNumber: Integer); stdcall;
begin
  if length(FOutputConnections) <= AAsNumber then
  begin
    SetLength(FOutputConnections, AAsNumber + 1);
  end;
  FOutputConnections[AAsNumber] := AConnection;
end;

procedure TDataStreamProcessor.CalculateOutputConnections;
var
  pipe: IScenarioConnection;
begin
  for pipe in FOutputConnections do
  begin
    if not Assigned(pipe) then
      continue;
    pipe.Calculate;
  end;

  for pipe in FInputConnections do
    pipe.Calculate;
end;

function TDataStreamProcessor.DoClone(const AResult: IScenarioNode): IScenarioNode;
//var
//  i: Integer;
begin
  // todo
//  for i := 0 to CountOutputConnections - 1 do
//    TDataStreamProcessor(AResult).AddOutputConnection();

  Result := AResult;
end;

function TDataStreamProcessor.DoGetOutputConnections(AIndex: Integer): IScenarioConnection;
begin
  if Length(FOutputConnections) > AIndex then
    Result := FOutputConnections[AIndex]
  else
    Result := nil;
end;

procedure TDataStreamProcessor.DoSetOutputConnections(AIndex: Integer; const APipe: IScenarioConnection);
begin
  if AIndex >= Length(FOutputConnections) then
  begin
    SetLength(FOutputConnections, AIndex + 1);
  end;
  FOutputConnections[AIndex] := APipe;
end;

function TDataStreamProcessor.DoGetInputConnections(AIndex: Integer): IScenarioConnection;
begin
  Result := FInputConnections.Items[AIndex];
end;

function TDataStreamProcessor.DoCreateOutputConnection(ANumberConnection, ANumberInput: Integer; const ADestination: IDataStreamProcessor): IScenarioConnection;
var
  objEqual: boolean;
begin
  // удаляем связь если опять на этот же элемент создаётся
  if (ANumberConnection < Length(FOutputConnections)) and Assigned(FOutputConnections[ANumberConnection]) then
  begin
    if Assigned(FOutputConnections[ANumberConnection].Destination) then
    begin
      objEqual := (FOutputConnections[ANumberConnection].Destination = ADestination);
      IDataStreamProcessor(FOutputConnections[ANumberConnection].Destination).RemoveInputConnection(FOutputConnections[ANumberConnection]);
    end else
      objEqual := false;

    FOutputConnections[ANumberConnection] := nil;

    if objEqual then
    begin
      while (Length(FOutputConnections) > 0) and (FOutputConnections[Length(FOutputConnections)-1] = nil) do
        SetLength(FOutputConnections, Length(FOutputConnections) - 1);
      exit(nil);
    end;
  end;

  Result := TScenarioConnection.Create(Owner);
  Result.Number := ANumberConnection;
  Result.NumberInput := ANumberInput;
  Result.Source := Self as IDataStreamProcessor;
  Result.Destination := ADestination;
end;

function TDataStreamProcessor.DoHasOutputConnection(const ADestination: IDataStreamProcessor): IScenarioConnection;
var
  i: Int32;
begin
  for i := 0 to length(FOutputConnections) - 1 do
  begin
    if Assigned(FOutputConnections[i]) and Assigned(FOutputConnections[i].Destination) and (IDatastreamProcessor(FOutputConnections[i].Destination).Guid^ = ADestination.Guid^) then
      exit(FOutputConnections[i]);
  end;
  Result := nil;
end;

constructor TDataStreamProcessor.Create(AOwner: TScenarioLayer);
begin
  inherited;
  FInputConnections := TList<IScenarioConnection>.Create;
  FLastTimeProcessData := TThread.GetTickCount64;
end;

function TDataStreamProcessor.CreateDataStream(AParentStream: TDataStream; AInputNumber: Integer): TDataStream; stdcall;
begin
  IncCountInputStreams;
  Result := DoCreateDataStream(Self, AParentStream, AInputNumber);
end;

function TDataStreamProcessor.CreateOutputConnection(ANumberConnection, ANumberInput: Integer; const ADestination: IDataStreamProcessor): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}; stdcall;
var
  conn: IScenarioConnection;
begin
  conn := DoCreateOutputConnection(ANumberConnection, ANumberInput, ADestination);
  Result := {$ifdef NARI}Pointer{$endif}(conn);
end;

destructor TDataStreamProcessor.Destroy;
begin
  FInputConnections.Free;
  inherited;
end;

procedure TDataStreamProcessor.DestroyDataStream(ADataStream: TDataStream);
  stdcall;
begin
  DecCountInputStreams;
end;

procedure TDataStreamProcessor.SetPosition(AValue: PPoint); stdcall;
begin
  inherited;
  CalculateOutputConnections;
end;

procedure TDataStreamProcessor.SetQueuesItemsCount(const Value: UInt32);
  stdcall;
begin
  FQueuesItemsCount := Value;
end;

procedure TDataStreamProcessor.SetQueuesSize(const Value: UInt32); stdcall;
begin
  FQueuesSize := Value;
end;

procedure TDataStreamProcessor.DecCountInputStreams;
begin
  dec(FCountInputStreams);
end;

procedure TDataStreamProcessor.DecCountOutputStreams;
begin
  dec(FCountOutputStreams);
end;

function TDataStreamProcessor.DoAddDataStreamAttribute(ADataStream: PRealTimeDataStream; AID: Integer; AData: Pointer; ADataSize: Integer): PDataStreamAttribute;
begin
  Result := CreateDataStreamAttribute(ADataStream.Attributes, AID, AData, ADataSize);
  Result.DataStreamOwner := ADataStream;
  ADataStream.Attributes := Result;
end;

function TDataStreamProcessor.DoCreateDataStream(AOwner: TDataStreamProcessor; AParentStream: TDataStream; AInputNumber: Integer): PRealTimeDataStream;
begin
  new(Result);
  Result.Parent := AParentStream;
  Result.Owner := AOwner;
  Result.Attributes := nil;
  Result.Queue := nil;
  Result.NumberInput := AInputNumber;
  Result.Processor := nil;
  /// создадим служебный пустой атрибут разделитель списка между узлами
  DoAddDataStreamAttribute(Result, -1, nil, 0);
end;

function TDataStreamProcessor.GetOutputConnections(AIndex: Integer): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}; stdcall;
var
  conn: IScenarioConnection;
begin
  conn := DoGetOutputConnections(AIndex);
  Result := {$ifdef NARI}Pointer{$endif}(conn);
end;

procedure TDataStreamProcessor.IncCountInputStreams;
begin
  inc(FCountInputStreams);
end;

procedure TDataStreamProcessor.IncCountOutputStreams;
begin
  inc(FCountOutputStreams);
end;

procedure TDataStreamProcessor.ProcessData(ADataStream: TDataStream;
  AData: PByte; ASize: Int64); stdcall;
var
  delta: uint64;
begin
  if ASize > 0 then
  begin
    inc(FCountInputData, ASize);
    FLastTimeProcessData := TThread.GetTickCount64;
    if (FState <> TNodeState.nsWorking) and (FState <> TNodeState.nsStopping) then
      State := TNodeState.nsWorking;

    delta := FLastTimeProcessData - FLastTimeReqInputSpeed;
    if delta > 0 then
    begin
      FInputSpeed := round(1000*FCountInputData/delta);
      if delta >= 50 then
      begin
        FCountInputData := 0;
        FLastTimeReqInputSpeed := FLastTimeProcessData;
      end;
    end;
  end;
end;

procedure TDataStreamProcessor.RemoveInputConnection(const AConnection: IScenarioConnection); stdcall;
begin
  FInputConnections.Remove(AConnection);
end;

procedure TDataStreamProcessor.RemoveOutputConnection(
  const AConnection: IScenarioConnection); stdcall;
var
  i: Integer;
begin
  for i := 0 to Length(FOutputConnections) - 1 do
  begin
    if FOutputConnections[i] = AConnection then
    begin
      FOutputConnections[i] := nil;
      IDataStreamProcessor(AConnection.Destination).RemoveInputConnection(AConnection);
      while (Length(FOutputConnections) > 0) and (FOutputConnections[Length(FOutputConnections)-1] = nil) do
        SetLength(FOutputConnections, Length(FOutputConnections) - 1);
      break;
    end;
  end;
end;

procedure TDataStreamProcessor.Resize(AWidth, AHeight: Integer); stdcall;
begin
  inherited;
  CalculateOutputConnections;
end;

procedure TDataStreamProcessor.SetCountInputDataStreams(const Value: Int64); stdcall;
begin
  FCountInputStreams := Value;
end;

procedure TDataStreamProcessor.SetCountOutputDataStreams(const Value: Int64); stdcall;
begin
  FCountOutputStreams := Value;
end;

procedure TDataStreamProcessor.SetInputSpeed(const Value: UInt64); stdcall;
begin
  FInputSpeed := Value;
end;

procedure TDataStreamProcessor.SetOutputConnections(AIndex: Integer; const APipe: {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}); stdcall;
begin
  DoSetOutputConnections(AIndex, {$ifdef NARI}IScenarioConnection{$endif}(APipe));
end;

procedure TDataStreamProcessor.SetOutputSpeed(const Value: UInt64); stdcall;
begin
  FOutputSpeed := Value;
end;

{ TScenarioPlugin }

procedure TScenarioPlugin.AddGlobalAttribute(AID: Integer; AData: Pointer; ADataSize: Integer); stdcall;
begin
  FOwner.Scenario.AddGlobalAttribute(AID, AData, ADataSize);
end;

procedure TScenarioPlugin.AddGlobalNamedAttribute(AName: PWideChar; AData: Pointer; ADataSize: Integer); stdcall;
begin
  FOwner.Scenario.AddGlobalNamedAttribute(AName, AData, ADataSize);
end;

procedure TScenarioPlugin.AddTable(const ATable: IPluginTable);
begin
  FTables.Add(ATable);
end;

function TScenarioPlugin.CreateStream(AConnectionNumber: Integer; AParentDataStream: TDataStream): TDataStream; stdcall;
var
  pipe: IScenarioConnection;
  qItem: TQueueItem;
begin
  pipe := OutputConnections[AConnectionNumber];
  if Assigned(pipe) and Assigned(pipe.Destination) then
  begin
    if Assigned(pipe.Queue) then
    begin
      Result := DoCreateDataStream(IDataStreamProcessor(pipe.Destination) as TDataStreamProcessor, AParentDataStream, pipe.NumberInput);
      PRealTimeDataStream(Result).Queue := pipe.Queue;
      // если поток данных будет приниматься из очереди, значит необходимо скопировать атрибуты из родительского,
      // т.к. доступ к ним нельзя выполнять из другого потока выполнения
      if Assigned(AParentDataStream) then
        SetAttributes(Result, PRealTimeDataStream(AParentDataStream).Attributes);
      qItem.QueueItemType := itCreateStream;
      qItem.DataStream := Result;
      PRealTimeDataStream(Result).Queue.Write(nil, 0, qItem);
    end else
    begin
      Result := IDataStreamProcessor(pipe.Destination).CreateDataStream(AParentDataStream, pipe.NumberInput);
      // соединим списки родительских и дочерних узлов
      if Assigned(AParentDataStream) then
        PRealTimeDataStream(Result).Attributes.Next := PRealTimeDataStream(AParentDataStream).Attributes;
    end;

    IncCountOutputStreams;

  end else
  begin
    Result := nil;
    if not FHavingSystemMessages[smSendOnNoExistsConnection] then
    begin
      FHavingSystemMessages[smSendOnNoExistsConnection] := true;
      SendMessage(PWideChar('Плагин "' + Caption + '" отправляет данные на несуществующую по сценарию связь с номером ' + WideString(IntToStr(AConnectionNumber))), MESSAGE_WARN);
    end;
  end;
end;

function TScenarioPlugin.CreateTable(AName: PWideChar; ACaption: PWideChar): Pointer; stdcall;
var
  table: IPluginTable;
begin
  table := TPluginTable.Create(Self, ACaption, AName);
  Result := table;
  AddTable(table);
end;

function TScenarioPlugin.FindAttribute(ADataStream: TDataStream; AID: Integer): PDataStreamAttribute; stdcall;
begin
  Result := PRealTimeDataStream(ADataStream).Attributes;
  while Assigned(Result) and (Result.ID <> AID) do
    Result := Result.Next;
end;

function TScenarioPlugin.FindAttributeNativeInt(ADataStream: TDataStream; AID: Integer; ADefault: NativeInt): NativeInt; stdcall;
var
  attr: PDataStreamAttribute;
begin
  attr := FindAttribute(ADataStream, AID);
  Result := ADefault;
  if Assigned(attr) and (attr.Size <= SizeOf(NativeInt)) then
    move(attr.Attribute^, Result, attr.Size);
end;

function TScenarioPlugin.FindAttributeValuePtr(ADataStream: TDataStream; AID: Integer; ADefaultValuePtr: Pointer): Pointer; stdcall;
var
  attr: PDataStreamAttribute;
begin
  attr := FindAttribute(ADataStream, AID);
  Result := ADefaultValuePtr;
  if Assigned(attr) then
    Result := attr.Attribute;
end;

function TScenarioPlugin.FindGloabalAttribute(AID: Integer): IValueHolder;
  stdcall;
begin
  Result := FOwner.Scenario.FindGloabalAttribute(AID);
end;

function TScenarioPlugin.FindGloabalNamedAttribute(AName: PWideChar
  ): IValueHolder; stdcall;
begin
  Result := FOwner.Scenario.FindGloabalNamedAttribute(AName);
end;

procedure TScenarioPlugin.SendMessage(AMsg: PWideChar; AMessageType: TMessageType); stdcall;
begin
  DoSendNodeMessage(Self, AMsg, AMessageType);
end;

procedure TScenarioPlugin.SetHealthState(AMsg: PWideChar; AHealthState: TNodeHalthDegreeState); stdcall;
begin
  ChangeHealthState(AMsg, AHealthState);
end;

procedure TScenarioPlugin.SetLoadBeforeStartScenario(AValue: boolean); stdcall;
begin
  FLoadBeforeStartScenario := AValue;
end;

function TScenarioPlugin.AddDataStreamAttribute(ADataStream: TDataStream; AID: Integer; AData: Pointer; ADataSize: Integer): PDataStreamAttribute; stdcall;
begin
  if not Assigned(ADataStream) then
    exit(nil);

  Result := DoAddDataStreamAttribute(ADataStream, AID, AData, ADataSize);
end;

procedure TScenarioPlugin.ClearGlobalAttributes; stdcall;
begin
  FOwner.Scenario.ClearGlobalAttributes;
end;

constructor TScenarioPlugin.Create(AOwner: TScenarioLayer);
begin
  inherited Create(AOwner);
  FWidth := 136;
  FHeight := 34;
  FTables := TList<IPluginTable>.Create;
end;

function TScenarioPlugin.CreateDataStream(AParentStream: TDataStream; AInputNumber: Integer): TDataStream; stdcall;
begin
  CreatePlugin;
  Result := inherited;
end;

procedure TScenarioPlugin.CreatePlugin;
var
  cinema: ICinema;
begin
  if not Assigned(FCinemaPlugin) then
  begin
    if FHavingSystemMessages[smPluginNotFound] then
      exit;

    State := TNodeState.nsStarting;
    try
      if supports(Self, ICinema, cinema) then
      begin
        try
          FCinemaPlugin := FPlugin.CreatePlugin(PWideChar(FConfiguration), cinema);
        except
          on e: Exception do
            SendMessage(PWideChar('Ошибка создания плагина "' + FCaption + '": ' + WideString(e.Message)), MESSAGE_ERROR);
        end;
        if not Assigned(FCinemaPlugin) then
        begin
          if not FHavingSystemMessages[smPluginNotFound] then
          begin
            FHavingSystemMessages[smPluginNotFound] := true;
            SendMessage(PWideChar('Плагин "' + FCaption + '" не найден'), MESSAGE_ERROR);
            exit;
          end;
        end;
      end else
        raise ECinema.Create('Error Message');
    finally
      if Assigned(FCinemaPlugin) then
        State := TNodeState.nsWorking
      else
        State := TNodeState.nsNone;
    end;
  end;
end;

destructor TScenarioPlugin.Destroy;
begin
  FTables.Free;
  inherited;
end;

procedure TScenarioPlugin.DestroyDataStream(ADataStream: TDataStream); stdcall;
var
  attr, a: PDataStreamAttribute;
begin
  if not Assigned(ADataStream) then
    exit;

  if FInputConnections.Count = 0 then
    State := TNodeState.nsStopping;

  if Assigned(PRealTimeDataStream(ADataStream).Processor) then
    FCinemaPlugin.DestroyDataStreamProcessor(PRealTimeDataStream(ADataStream).Processor);
  // удаление атрибутов канала принадлежащих текущему процессору
  attr := PRealTimeDataStream(ADataStream).Attributes;
  while Assigned(attr) and (attr.DataStreamOwner = ADataStream) do
  begin
    a := attr;
    attr := attr.Next;
    if a.Size > 0 then
      SysFreeMem(a.Attribute);
    dispose(a);
  end;
  dispose(PRealTimeDataStream(ADataStream));
  // если нет входов, значит стартовый плагин, а т.к. у него всегда единственный класс-просессор, то
  // освобождаем плагин, т.к. это означает что сценарий останавливается
  if (FInputConnections.Count = 0) then
  begin
    //FPlugin.DestroyPlugin(FCinemaPlugin);
    FCinemaPlugin := nil;
    State := TNodeState.nsNone;
  end;
  inherited;
end;

procedure TScenarioPlugin.DestroyStream(ADataStream: TDataStream); stdcall;
var
  qItem: TQueueItem;
begin
  if not Assigned(ADataStream) then
    exit;

  if Assigned(PRealTimeDataStream(ADataStream).Queue) then
  begin
    DecCountOutputStreams;
    qItem.QueueItemType := itDestroyStream;
    qItem.DataStream := ADataStream;
    PRealTimeDataStream(ADataStream).Queue.Write(nil, 0, qItem);
  end else
  begin
    DoDestroyStream(ADataStream);
  end;
end;

procedure TScenarioPlugin.DoDestroyStream(ADataStream: PRealTimeDataStream);
begin
  TScenarioPlugin(ADataStream.Owner).DestroyDataStream(ADataStream);
  DecCountOutputStreams;
end;

procedure TScenarioPlugin.DoSendData(ADataStream: PRealTimeDataStream; AData: PByte; ASize: Int64);
var
  delta: UInt64;
begin
  if ASize > 0 then
  begin
    inc(FCountOutputData, ASize);
    FLastTimeProcessData := TThread.GetTickCount64;
    delta := FLastTimeProcessData - FLastTimeReqOutputSpeed;
    if delta > 0 then
    begin
      FOutputSpeed := round(1000*FCountOutputData/delta);
      if delta >= 50 then
      begin
        FCountOutputData := 0;
        FLastTimeReqOutputSpeed := FLastTimeProcessData;
      end;
    end;
  end;

  if (FState <> TNodeState.nsWorking) and (FState <> TNodeState.nsStopping) then
    State := TNodeState.nsWorking;

  if not Assigned(ADataStream) then
  begin
    if not FHavingSystemMessages[smSendOnNoExistsConnection] then
    begin
      FHavingSystemMessages[smSendOnNoExistsConnection] := true;
      SendMessage(PWideChar('Плагин "' + Caption + '" отправляет данные по отсутствующему потоку'), MESSAGE_WARN);
    end;
    exit;
  end;

  TScenarioPlugin(ADataStream.Owner).ProcessData(ADataStream, AData, ASize);
end;

function TScenarioPlugin.DoGetTable(AIndex: Integer): IPluginTable;
begin
  Result := FTables.Items[AIndex];
end;

function TScenarioPlugin.DoGetCinemaPlugin: ICinemaPlugin;
begin
  Result := FCinemaPlugin;
end;

procedure TScenarioPlugin.SendData(ADataStream: TDataStream; AData: PByte; ASize: Int64); stdcall;
var
  qItem: TQueueItem;
begin
  if not Assigned(ADataStream) then
    exit;

  if Assigned(PRealTimeDataStream(ADataStream).Queue) then
  begin
    qItem.QueueItemType := itData;
    qItem.DataStream := ADataStream;
    PRealTimeDataStream(ADataStream).Queue.Write(AData, ASize, qItem);
  end else
    DoSendData(ADataStream, AData, ASize);
end;

function TScenarioPlugin.GetCountTables: Integer; stdcall;
begin
  Result := FTables.Count;
end;

function TScenarioPlugin.GetAsStart: boolean; stdcall;
begin
  Result := FAsStart;
end;

function TScenarioPlugin.GetLoadBeforeStartScenario: boolean; stdcall;
begin
  Result := FLoadBeforeStartScenario;
end;

function TScenarioPlugin.GetTable(AIndex: Integer): {$ifdef NARI} Pointer {$else} IPluginTable {$endif}; stdcall;
var
  table: IPluginTable;
begin
  table := DoGetTable(AIndex);
  Result := {$ifdef NARI}Pointer{$endif}(table);
end;

function TScenarioPlugin.GetVersionScenario: PWideChar; stdcall;
begin
  Result := FOwner.Scenario.Version;
end;

function TScenarioPlugin.IsThreadTerminated: boolean; stdcall;
begin
  Result := Parent.IsThreadTerminated;
end;

function TScenarioPlugin.GetAppPath: PWideChar; stdcall;
begin
  Result := @AppPath[1];
end;

function TScenarioPlugin.GetAttributes(ADataStream: TDataStream): PDataStreamAttribute; stdcall;
begin
  if Assigned(ADataStream) then
    Result := PRealTimeDataStream(ADataStream).Attributes
  else
    Result := nil;
end;

function TScenarioPlugin.GetCinemaPlugin: {$ifdef NARI} Pointer {$else} ICinemaPlugin {$endif}; stdcall;
var
  p: ICinemaPlugin;
begin
  p := DoGetCinemaPlugin;
  Result := {$ifdef NARI}Pointer{$endif}(p);
end;

function TScenarioPlugin.GetConfiguration: PWideChar; stdcall;
begin
  Result := PWideChar(FConfiguration);
end;

function TScenarioPlugin.GetName: PWideChar; stdcall;
begin
  Result := PWideChar(FName);
end;

function TScenarioPlugin.GetNodeType: TGuid;
begin
  Result := IScenarioPlugin;
end;

function TScenarioPlugin.GetPluginDescription: PPluginDescription; stdcall;
begin
  if Assigned(FPlugin) then
    Result := FPlugin.PluginDescription
  else
    Result := nil;
end;

function TScenarioPlugin.GetPluginPath: PWideChar; stdcall;
begin
  Result := nil;
  if Assigned(FPlugin) then
  begin
    FStringBuf := FPlugin.LibraryPath;
    if FStringBuf <> '' then
      Result := @FStringBuf[1];
  end
end;

function TScenarioPlugin.GetRtlScenariosPath: PWideChar; stdcall;
begin
  Result := PWideChar(TCinemaConfiguration.RtlScenariosPath);
end;

function TScenarioPlugin.GetScenarioName: PWideChar; stdcall;
begin
  FStringBuf := FOwner.GetCinemaScenario.ScenarioFullPath;
  if FStringBuf <> '' then
    Result := @FStringBuf[1]
  else
    Result := '';
end;

function TScenarioPlugin.Load(AValues: PWideChar; const AParentLayer: IScenarioLayer): boolean; stdcall;
begin
  Result := inherited;
end;

procedure TScenarioPlugin.LoadFormJson(const AValues: TJsonCinema; const AParentLayer: IScenarioLayer);
var
  config: TJsonCinema;
  s: WideString;
  value: UnicodeString;
  {$ifndef fpc}
  decoder: TBase64Encoding;
  {$endif}
begin
  inherited;
  FIsLoading := true;
  try
    Name := PWideChar(AValues.GetValue<UnicodeString>('name'));
    FLoadBeforeStartScenario := AValues.GetValue<Boolean>('loadBeforeStart', false);
    if (AValues.TryGetValue<TJsonCinema>('config', config)) and Assigned(config) then
    begin
      if (config.GetValue<boolean>('base64')) and (config.TryGetValue<UnicodeString>('value', value)) then
      begin
        {$ifdef fpc}
        	s := WideString(DecodeStringBase64(string(value)));
        {$else}
          decoder := TBase64Encoding.Create;
          try
            s := decoder.Decode(value);
          finally
            decoder.Free;
          end;
        {$endif}
      end else
        s := WideString(config.GetValue<TJsonCinema>('value').ToJSON);
      Configuration := PWideChar(s);
    end;
    Parent := AParentLayer;
  finally
    FIsLoading := false;
  end;
end;

function TScenarioPlugin.PackAttributes(ADataStreamSource: TDataStream; AFilters: PInteger; AHighIdInFilters: Int32; ALength: PInteger): PByte; stdcall;
var
  attribute: PDataStreamAttribute;
  owner: TScenarioPlugin;
  len: Integer;
  filter: Integer;
begin
  ALength^ := 0;
  if not Assigned(ADataStreamSource) then
    exit(nil);

  attribute := PRealTimeDataStream(ADataStreamSource).Attributes;
  while Assigned(attribute) do
  begin
    if attribute.ID < 0 then
    begin
      // упакуем название плагина так же
      owner := TScenarioPlugin(PRealTimeDataStream(attribute.DataStreamOwner).Owner);
      len := owner.Plugin.Name.Length*2;

      if Length(FBufferForAttribute) < ALength^ + len + 8 then
        SetLength(FBufferForAttribute, ALength^ + len + 8);

      move(attribute.ID, FBufferForAttribute[ALength^], 4);
      move(len, FBufferForAttribute[ALength^+4], 4);
      move(owner.Plugin.Name[1], FBufferForAttribute[ALength^+8], len);
      inc(ALength^, len + 8);
    end else
    if (attribute.Size > 0) then
    begin
      if Assigned(AFilters) and (attribute.ID <= AHighIdInFilters) then
        filter := PInteger(PByte(AFilters) + attribute.id*SizeOf(Int32))^
      else
        filter := -1;

      if filter <> attribute.ID then
      begin
        if Length(FBufferForAttribute) < ALength^ + attribute.Size + 8 then
          SetLength(FBufferForAttribute, ALength^ + attribute.Size + 8);

        move(attribute.ID, FBufferForAttribute[ALength^], 4);
        move(attribute.Size, FBufferForAttribute[ALength^+4], 4);
        move(attribute.Attribute^, FBufferForAttribute[ALength^+8], attribute.Size);
        inc(ALength^, attribute.Size + 8);
      end;
    end;
    attribute := attribute.Next;
  end;

  if ALength^ > 0 then
    Result := @FBufferForAttribute[0]
  else
    Result := nil;

end;

procedure TScenarioPlugin.ProcessData(ADataStream: TDataStream; AData: PByte; ASize: Int64); stdcall;
begin
  inherited;
  if Assigned(ADataStream) then
  begin
    if not Assigned(PRealtimeDataStream(ADataStream).Processor) then
      PRealtimeDataStream(ADataStream).Processor := FCinemaPlugin.CreateDataStreamProcessor(ADataStream, PRealtimeDataStream(ADataStream).NumberInput);
    FCinemaPlugin.ProcessDataStream(PRealTimeDataStream(ADataStream).Processor, AData, ASize);
  end;
end;

procedure TScenarioPlugin.RequestRedrawView;
begin
  if FNeedUpdateTables then
    exit;
  FNeedUpdateTables := true;
  Owner.RequestRedrawView(Self);
end;

function TScenarioPlugin.Save: PWideChar; stdcall;
var
  s: WideString;
  {$ifndef FPC}
  encoder: TBase64Encoding;
  {$endif}
begin
  inherited Save;
  FIsSaving := true;
  try
    {$ifdef FPC}
    s := WideString(EncodeStringBase64(string(FConfiguration)));
    {$else}
    encoder := TBase64Encoding.Create(0);
    try
      s := encoder.Encode(FConfiguration);
    finally
      encoder.Free;
    end;
    {$endif}
    FStringBuffer := FStringBuffer + ',' + sLineBreak +
      //'"id": ' + IntToStr(ID) + ',' + sLineBreak +
      '"name": "' + Name + '",' + sLineBreak +
      '"loadBeforeStart": "' + WideString(BoolToStr(LoadBeforeStartScenario, true)) + '",' + sLineBreak +
      '"config": {' + sLineBreak +
        '"base64": "true",' + sLineBreak +
        '"value": "' + s + '"' + sLineBreak +
      '}'
      ;
    Result := PWideChar(FStringBuffer);
  finally
    FIsSaving := false;
  end;
end;

procedure TScenarioPlugin.SetForceAsStart(AValue: boolean); stdcall;
begin
  AsStart := AValue;
  FOwner.OnSetForceAsStartRealtime(Self);
end;

procedure TScenarioPlugin.SetAttributes(ADataStream: TDataStream; AAtributes: PDataStreamAttribute); stdcall;
var
  attr: PDataStreamAttribute;
begin
  attr := AAtributes;
  while Assigned(attr) do
  begin
    AddDataStreamAttribute(ADataStream, attr.ID, attr.Attribute, attr.Size);
    attr := attr.Next;
  end;
end;

procedure TScenarioPlugin.SetConfiguration(const AValue: PWideChar); stdcall;
begin
  FConfiguration := AValue;
end;

procedure TScenarioPlugin.SetAsStart(AValue: boolean); stdcall;
begin
  FAsStart := AValue;
end;

procedure TScenarioPlugin.SetName(const AName: PWideChar); stdcall;
begin
  FName := AName;
  FPlugin := FOwner.Scenario.PluginsDictionary.GetPlugin(WideString(FName));
  if not Assigned(FPlugin) then
    SendMessage(PWideChar('Плагин "' + FName+ '" не найден'), MESSAGE_ERROR);
end;

procedure TScenarioPlugin.SetState(const AValue: TNodeState); stdcall;
var
  was: TNodeState;
  sm: TSystemMessages;
begin
  was := State;
  inherited;
  // определяем переход из не рабочего состояния в рабочее
  if (was = TNodeState.nsNone) and (State <> TNodeState.nsNone) then
  begin
    // удалим предыдущие таблицы если они были
    FTables.Clear;
    // очистим информацию об наличии системных событий
    for sm := Low(TSystemMessages) to High(TSystemMessages) do
      FHavingSystemMessages[sm] := false;
  end;

  // выгружаем плагин если переходим в не рабочее состояние
  if (AValue = TNodeState.nsNone) and (FCountInputStreams = 0) and (FCountOutputStreams = 0) then
    FCinemaPlugin := nil;
end;

procedure TScenarioPlugin.UnPackAttributes(ADataStreamDest: TDataStream; AAttributes: PByte; ALength: UInt32); stdcall;
var
  pos: Int32;
  id: Int32;
  size: Int32;
begin
  if not Assigned(ADataStreamDest) then
    exit;

  pos := 0;
  while UInt32(pos) < ALength do
  begin
    id := PInteger(@AAttributes[pos])^;
    size := PInteger(@AAttributes[pos+4])^;
    // защита от некачественных данных
    if (UInt32(pos + size + 8) > ALength) or (size < 0) then
      break;
    AddDataStreamAttribute(ADataStreamDest, id, @AAttributes[pos+8], size);
    inc(pos, size + 8);
  end;

end;

procedure TScenarioPlugin.RedrawView;
begin
  FNeedUpdateTables := false;
  FCinemaPlugin.RedrawView;
end;

procedure TScenarioPlugin.RemoveGlobalAttribute(AID: Integer); stdcall;
begin
  FOwner.Scenario.RemoveGlobalAttribute(AID);
end;

procedure TScenarioPlugin.RemoveGlobalNamedAttribute(AName: PWideChar); stdcall;
begin
  FOwner.Scenario.RemoveGlobalNamedAttribute(AName);
end;

{ TScenarioConnection }

procedure TScenarioConnection.Calculate; stdcall;
var
  rectSrc: TRect;
  rectDst: TRect;
  centerSrc: TPoint;
  centerDst: TPoint;
  k, b: single;
  overlap: TRect;

  function CalcPointIntersect(const ARect: TRect; var AAlign: TConnectionAlign): TPoint;
  begin
    Result.Y := Round(k * ARect.Left + b);
    if ((centerSrc.X < ARect.Left) or (centerDst.X < ARect.Left)) and (Result.Y >= ARect.Top) and (Result.Y <= ARect.Bottom) then
    begin //слева?
      Result.X := ARect.Left;
      AAlign := TConnectionAlign.caLeft;
    end else
    begin
      //справа?
      Result.Y := Round(k * ARect.Right + b);
      if ((centerSrc.X > ARect.Right) or (Result.X > ARect.Right)) and (Result.Y >= ARect.Top) and (Result.Y <= ARect.Bottom) then
      begin
        Result.X := ARect.Right;
        AAlign := TConnectionAlign.caRight;
      end else
      begin
        //сверху?
        Result.X := Round((ARect.Top - b)/k);
        AAlign := TConnectionAlign.caTop;
        if ((centerSrc.Y < ARect.Top) or (centerDst.Y < ARect.Top)) and (Result.X >= ARect.Left) and (Result.X <= ARect.Right) then
        begin
          Result.Y := ARect.Top;
        end else
        begin //снизу?
          Result.X := Round((ARect.Bottom - b)/k);
          Result.Y := ARect.Bottom;
          AAlign := TConnectionAlign.caBottom;
        end;
      end;
    end;
  end;

begin

  if not Assigned(Source) or not Assigned(Destination) then
    exit;

  rectSrc := (DoGetSource as TDataStreamProcessor).Rect^;
  rectDst := (DoGetDestination as TDataStreamProcessor).Rect^;

  centerSrc := rectSrc.CenterPoint;
  centerDst := rectDst.CenterPoint;

  FRect.Left := min(FPointSrc.X, FPointDst.X);
  FRect.Top := min(FPointSrc.Y, FPointDst.Y);
  FRect.Right := max(FPointSrc.X, FPointDst.X);
  FRect.Bottom := max(FPointSrc.Y, FPointDst.Y);

  overlap := RectOverlap(rectSrc, rectDst);

  if Source = Destination then
  begin // feedback
    FAlignFrom := TConnectionAlign.caBottom;
    FAlignTo := TConnectionAlign.caTop;
    FPointSrc := TPoint.Create(centerSrc.X, rectSrc.Bottom);
    FPointDst := TPoint.Create(centerSrc.X, rectSrc.Top);
  end else
  if (overlap.Width > 0) and (overlap.Height > 0) then
  begin // rects are overlaped
    FAlignFrom := TConnectionAlign.caBottom;
    FAlignTo := TConnectionAlign.caTop;
    FPointSrc := centerSrc;
    FPointDst := centerDst;
  end else
  if (overlap.Width > 0) then
  begin // overlap only on x
    if (centerSrc.y < centerDst.y) then
    begin
      FAlignFrom := TConnectionAlign.caBottom;
      FAlignTo := TConnectionAlign.caTop;
      FPointSrc := TPoint.Create((overlap.Left + overlap.Right) shr 1, rectSrc.Bottom);
      FPointDst := TPoint.Create(FPointSrc.X, rectDst.Top);
    end else
    begin
      FAlignFrom := TConnectionAlign.caTop;
      FAlignTo := TConnectionAlign.caBottom;
      FPointDst := TPoint.Create((overlap.Left + overlap.Right) shr 1, rectDst.Bottom);
      FPointSrc := TPoint.Create(FPointDst.X, rectSrc.Top);
    end;
  end else
  if (overlap.Height > 0) then
  begin // overlap only on y
    if (centerSrc.x < centerDst.x) then
    begin
      FAlignFrom := TConnectionAlign.caRight;
      FAlignTo := TConnectionAlign.caLeft;
      FPointSrc := TPoint.Create(rectSrc.Right, (overlap.Top + overlap.Bottom) shr 1);
      FPointDst := TPoint.Create(rectDst.Left, FPointSrc.Y);
    end else
    begin
      FAlignFrom := TConnectionAlign.caLeft;
      FAlignTo := TConnectionAlign.caRight;
      FPointSrc := TPoint.Create(rectSrc.Left, (overlap.Top + overlap.Bottom) shr 1);
      FPointDst := TPoint.Create(rectDst.Right, FPointSrc.Y);
    end;
  end else
  begin // rects do not overlap

    /// вычисляем уравнение прямой
    /// коэфициент наклона k
    ///
    /// y = k*x + b или A*x + B*y + C = 0;
    /// k = (Y2 - Y1) / (X2 - X1)
    /// k = A / B
    /// b = C / B
    k := (centerDst.Y  - centerSrc.Y)/(centerDst.X - centerSrc.X);
    /// коэфициент смещения b
    b := centerSrc.Y - centerSrc.X * k;
    FPointSrc := CalcPointIntersect(rectSrc, FAlignFrom);
    FPointDst := CalcPointIntersect(rectDst, FAlignTo);
  end;

end;

function TScenarioConnection.DoGetSource: IDataStreamProcessor;
begin
  if not Assigned(FSource) then
  begin
    if not FUidSource.IsEmpty then
      Source := FOwner.DoGetChildByGuid(FUidSource) as IDataStreamProcessor;
  end;
  Result := FSource;
end;

procedure TScenarioConnection.DoSetSource(const AValue: IDataStreamProcessor);
begin
  if Assigned(FSource) then
    FSource.RemoveOutputConnection(Self);
  FSource := AValue;
  if Assigned(FSource) then
  begin
    FSource.AddOutputConnection(Self, FNumber);
    if Assigned(FDestination) then
      Calculate;
  end;
end;

function TScenarioConnection.DoGetDestination: IDataStreamProcessor;
begin
  if not Assigned(FDestination) then
  begin
    if not FUidDestination.IsEmpty then
      Destination := FOwner.DoGetChildByGuid(FUidDestination) as IDataStreamProcessor;
  end;

  Result := FDestination;
end;

procedure TScenarioConnection.DoSetDestination(const AValue: IDataStreamProcessor);
begin
  if Assigned(FDestination) then
    FDestination.RemoveInputConnection(Self);
  FDestination := AValue;
  if Assigned(FDestination) then
  begin
    FDestination.AddInputConnection(Self);
    if Assigned(FSource) then
      Calculate;
  end;
end;

constructor TScenarioConnection.Create(AOwner: TScenarioLayer);
begin
  FOwner := AOwner;
end;

function TScenarioConnection.GetAlignFrom: TConnectionAlign; stdcall;
begin
  Result := FAlignFrom;
end;

function TScenarioConnection.GetAlignTo: TConnectionAlign; stdcall;
begin
  Result := FAlignTo;
end;

function TScenarioConnection.GetDestination: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif}; stdcall;
var
  dst: IDataStreamProcessor;
begin
  dst := DoGetDestination;
  Result := {$ifdef NARI}Pointer{$endif}(dst);
end;

function TScenarioConnection.GetNameInput: PWideChar; stdcall;
var
  plugin: IScenarioPlugin;
begin
  {$ifdef fpc}
  plugin := nil;
  {$endif}
  if (length(FNameInput) = 0) and (Destination <> nil) and (FDestination.QueryInterface(IScenarioPlugin, plugin) = S_OK) and Assigned(plugin.PluginDescription) then
  begin
    if (plugin.PluginDescription.CountNamedInputConnections > FNumberInput) then
      FNameInput := PNamedInputConnection(PByte(plugin.PluginDescription.NamedInputConnection) + SizeOf(TNamedInputConnection)*FNumberInput)^.Name;
  end;

  if length(FNameInput) = 0 then
    Result := nil
  else
    Result := PWideChar(FNameInput);
end;

function TScenarioConnection.GetNameOutput: PWideChar; stdcall;
var
  plugin: IScenarioPlugin;
begin
  {$ifdef fpc}
  plugin := nil;
  {$endif}
  if (length(FNameOutput) = 0) and (Source <> nil) and (FSource.QueryInterface(IScenarioPlugin, plugin) = S_OK) and Assigned(plugin.PluginDescription) then
  begin
    if (plugin.PluginDescription.CountNamedConnections > FNumber) then
      FNameOutput := PNamedOutputConnection(PByte(plugin.PluginDescription.NamedOutputConnection) + SizeOf(TNamedOutputConnection)*FNumber)^.Name;
  end;

  if length(FNameOutput) = 0 then
    Result := nil
  else
    Result := PWideChar(FNameOutput);
end;

function TScenarioConnection.GetNumber: Integer; stdcall;
begin
  Result := FNumber;
end;

function TScenarioConnection.GetNumberInput: Integer; stdcall;
begin
  Result := FNumberInput;
end;

function TScenarioConnection.GetPointDst: PPoint; stdcall;
begin
  Result := @FPointDst;
end;

function TScenarioConnection.GetPointSrc: PPoint; stdcall;
begin
  Result := @FPointSrc;
end;

function TScenarioConnection.GetQueue: Pointer; stdcall;
begin
  Result := FQueue;
end;

function TScenarioConnection.GetRect: PRect; stdcall;
begin
  Result := @FRect;
end;

function TScenarioConnection.GetSource: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif}; stdcall;
var
  src: IDataStreamProcessor;
begin
  src := DoGetSource;
  Result := {$ifdef NARI}Pointer{$endif}(src);
end;

function TScenarioConnection.GetView: TObject; stdcall;
begin
  Result := FView;
end;

function TScenarioConnection.HitTest(X, Y: Integer): boolean; stdcall;
var
  A0, B0, A1, B1: Integer;
  p00, p01: TPoint;
  k0, k1: single;
  trigger: byte;
begin
  if (
    ((FPointSrc.x < X) and (FPointDst.x < X)) or
    ((FPointSrc.x > X) and (FPointDst.x > X)) or
    ((FPointSrc.y < Y) and (FPointDst.y < Y)) or
    ((FPointSrc.y > Y) and (FPointDst.y > Y))
    ) then
      exit(false);

  A0 := FPointDst.x - FPointSrc.x;
  B0 := FPointDst.y - FPointSrc.y;
  A1 := 0;
  B1 := 0;
  trigger := 0;
  if abs(A0) > 5 then
    trigger := 1;
  if abs(B0) > 5 then
    inc(trigger, 2);
  if abs(A1) > 5 then
    inc(trigger, 4);
  if abs(B1) > 5 then
    inc(trigger, 8);

  Result := false;
  case trigger of
    0, 1, 2, 3, 4, 5, 8: begin // 0000
      // two points
    end;
    {1: begin  // 1000
      // one point and one parallel x line
    end;
    2: begin  // 0100
      // one point and one parallel y line
    end;
    3: begin  // 1100
      // one line and one point
    end;
    4: begin  // 0010
    end;
    5: begin  // 1010
      // two parallel (x) lines
    end;}
    6: begin  // 0110
      // first line parallel y, second - x
      Result := true;
    end;
    7: begin  // 1110
      // one (first) line and second parallel x
      Result := true;
    end;
    {8: begin  // 0001
      // one point and one line parallel y
    end;    }
    9: begin  // 1001
      // first line parallel x and second line parallel y
      Result := true;
    end;
    10: begin // 0101
      // both lines parallel y
    end;
    11: begin // 1101
      // one (first) line and second parallel y
      Result := true;
    end;
    12: begin // 0011
      // one point and one line

    end;
    13: begin // 1011
      // first line parallel x
      Result := true;
    end;
    14: begin  // 0111
      // first line parallel y
      Result := true;
    end;
    15: begin  // 1111
      k0 := B0/A0;
      k1 := B1/A1;
      if k0 - k1 <> 0 then
        Result := true
      else
        Result := false;
    end;
  end;

  if Result then
  begin
    if FPointDst.x < FPointSrc.x then
    begin
      p00.X := FPointDst.X;
      p01.X := FPointSrc.X;
    end else
    begin
      p00.X := FPointSrc.X;
      p01.X := FPointDst.X;
    end;

    if FPointDst.y < FPointSrc.y then
    begin
      p00.Y := FPointDst.Y;
      p01.Y := FPointSrc.Y;
    end else
    begin
      p00.Y := FPointSrc.Y;
      p01.Y := FPointDst.Y;
    end;

    if (p00.x > X) or (p01.x < X) or (p00.y > Y) or (p01.y < Y) then
      Result := false;
  end;
end;

function TScenarioConnection.Load(AValues: PWideChar; const AParentLayer: IScenarioLayer): boolean; stdcall;
var
  json: TJsonCinema;
begin
  Result := true;
  json := TJSONObject.ParseJSONValue(AValues);
  if Assigned(json) then
    LoadFromJson(json, AParentLayer);
end;

procedure TScenarioConnection.LoadFromJson(AValues: TJsonCinema; const AParentLayer: IScenarioLayer);
var
  uid: string;
  result: boolean;
begin
  // в FPC ошибка если AValues.TryGetValue сразу в условие разместить, из-за вызова через хелпер
  result := AValues.TryGetValue<Integer>('number', FNumber);
  if not result then
    exit;

  AValues.TryGetValue<Integer>('numberInput', FNumberInput);

  if AValues.TryGetValue<string>('fromUid', uid) then
    FUidSource := StringToGuid(string(uid));

  if AValues.TryGetValue<string>('toUid', uid) then
    FUidDestination := StringToGuid(string(uid));

  if not FUidSource.IsEmpty then
    Source := (AParentLayer as TScenarioLayer).DoGetChildByGuid(FUidSource) as IDataStreamProcessor;

  if not FUidDestination.IsEmpty then
    Destination := (AParentLayer as TScenarioLayer).DoGetChildByGuid(FUidDestination) as IDataStreamProcessor;
end;

function TScenarioConnection.Save: PWideChar; stdcall;
begin
  if not Assigned(Source) or not Assigned(Destination) then
    exit(nil);

  FStringBuffer :=
    StringToWide('"number": ' + IntToStr(FNumber) + ',' + sLineBreak +
    '"numberInput": ' + IntToStr(FNumberInput) + ',' + sLineBreak +
    '"fromUid":"' + (DoGetSource as TDataStreamProcessor).GUID.ToString + '",' + sLineBreak +
    '"toUid":"' + (DoGetDestination as TDataStreamProcessor).GUID.ToString + '"');
  Result := PWideChar(FStringBuffer);
end;

procedure TScenarioConnection.SetDestination(const AValue: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif}); stdcall;
begin
  DoSetDestination(IDataStreamProcessor(AValue));
end;

procedure TScenarioConnection.SetNumber(const AValue: Integer); stdcall;
begin
  FNumber := AValue;
end;

procedure TScenarioConnection.SetNumberInput(const AValue: Integer); stdcall;
begin
  FNumberInput := AValue;
end;

procedure TScenarioConnection.SetQueue(const AValue: Pointer); stdcall;
begin
  FQueue := AValue;
end;

procedure TScenarioConnection.SetSource(const AValue: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif}); stdcall;
begin
  DoSetSource({$ifdef NARI}IDataStreamProcessor{$endif}(AValue));
end;

procedure TScenarioConnection.SetView(const AValue: TObject); stdcall;
begin
  FView := AValue;
end;

{ TScenarioLayer }

procedure TScenarioLayer.AddChild(const AChild: TScenarioNode);
begin
  inherited;
  if AChild.Owner = Self then
  begin
    FChildren.Add(AChild);
    if AChild.InheritsFrom(TScenarioPlugin) then
      FChildrenPlugins.Add(TScenarioPlugin(AChild));
  end;

  if Assigned(FOwner) then
    FOwner.AddChild(AChild);
end;

procedure TScenarioLayer.AfterStart;
var
  i: Integer;
  layer: IScenarioLayer;
begin
  for i := 0 to CountChildren - 1 do
  begin
    if Supports(DoGetChildren(i), IScenarioLayer, layer) then
      (layer as TScenarioLayer).AfterStart;
  end;

  if Assigned(FWorkThread) then
    FWorkThread.Start;
end;

function TScenarioLayer.AskQueues: boolean; stdcall;
var
  i: int32;
  q: TQueueRam;
  data: PByte;
  size: Int64;
  item: TQueueItem;
  plugin: TScenarioPlugin;
begin
  Result := false;
  for i := 0 to FQueues.Count - 1 do
  begin
    q := FQueues.Items[i];
    while q.Read(data, size, item) do
    begin
      Result := true;
      TDataStreamProcessor(q.Owner).QueuesItemsCount := q.Count;
      TDataStreamProcessor(q.Owner).QueuesSize := q.Size;
      plugin := PRealTimeDataStream(item.DataStream).Owner as TScenarioPlugin;
      if item.QueueItemType = TQueueItemType.itData then
        plugin.DoSendData(item.DataStream, data, size)
      else
      if item.QueueItemType = TQueueItemType.itCreateStream then
      begin
        plugin.CreatePlugin;
        plugin.IncCountInputStreams;
      end else
        plugin.DestroyDataStream(item.DataStream);

      ServiceRoutines;
    end;
  end;
  if Result then
    FLastTimeReadQueue := TThread.GetTickCount64;
end;

procedure TScenarioLayer.CalculateStates;
var
  statesCounters: array[TNodeState] of Integer;
  worstHealthState: Byte;
  logStatesCounters: array[TNodeLogState] of Integer;
  logStatesMessages: array[TNodeLogState] of WideString;
  child: IScenarioNode;
  result: TNodeState;
  state: TNodeState;
  countLoaded: int32;
begin
  System.TMonitor.Enter(Self);
  try
    FillChar(statesCounters, SizeOf(Integer)*(Ord(High(TNodeState))+1), 0);
    FillChar(logStatesCounters, SizeOf(Integer)*(Ord(High(TNodeLogState))+1), 0);
    // обобщённое рабочее состояние слоя
    result := TNodeState.nsNone;
    // обобщённое состояние здоровья выполняемой работы слоем
    FHealthState := 100;
    // наихудшее сообщение отправленное в журнал от себя и дочерних объектов
    FLogState := TNodeLogState.lsInfo;

    worstHealthState := 0;
    countLoaded := 0;

    for child in FChildren do
    begin
      inc(statesCounters[child.State]);
      inc(logStatesCounters[child.LogState]);
      if child.State <> TNodeState.nsNone then
        inc(countLoaded);
      FHealthState := round(FHealthState * (child.HealthState / 100));
      logStatesMessages[child.LogState] := child.LogStateMessage;
      if worstHealthState > child.HealthState then
      begin
        worstHealthState := child.HealthState;
        FHealthStateMessage := child.HealthStateMessage;
      end;
    end;

    if logStatesCounters[TNodeLogState.lsError] > 0 then
      FLogState := TNodeLogState.lsError
    else
    if logStatesCounters[TNodeLogState.lsWarn] > 0 then
      FLogState := TNodeLogState.lsWarn;

    // здесь чёткого алгоритма выбора рабочего состояни нет, попробуем по логике приоритета;
    // если есть состояние в котором находятся все дочерние элементы - выбираем его
    if countLoaded > 0 then
    begin
      for state := low(TNodeState) to high(TNodeState) do
      begin
        if statesCounters[state] = countLoaded then
        begin
          result := state;
          break;
        end
      end;

      if (result = TNodeState.nsNone) and (statesCounters[TNodeState.nsNone] <> CountChildren) then
      begin
        if statesCounters[TNodeState.nsNoData] = countLoaded then
          result := TNodeState.nsNoData // если все элементы простаивают - общий статус - простаиваем
        else
        if (statesCounters[TNodeState.nsNoData] > 0) or (statesCounters[TNodeState.nsWorking] > 0) then
          result := TNodeState.nsWorking // если есть работющие элементы - в целом слой выполняет какую-то работу
        else
        if statesCounters[TNodeState.nsStopping] > 0 then
          result := TNodeState.nsStopping
        else
        if statesCounters[TNodeState.nsStarting] > 0 then
          result := TNodeState.nsStarting;
      end;
    end;

    FLogStateMessage := logStatesMessages[FLogState];
    FState := result;
  finally
    System.TMonitor.Exit(Self);
  end;
end;

function TScenarioLayer.HaveRunStopTask: boolean;
var
  layer: IScenarioLayer;
  child: IScenarioNode;
begin
  Result := Assigned(FStopTask) and (FStopTask.Status < TTaskStatus.Completed);
  if not Result then
  begin
    for child in FChildren do
    begin
      if Supports(child, IScenarioLayer, layer) then
      begin
        if (layer as TScenarioLayer).HaveRunStopTask then
          exit(true);
      end;
    end;
  end;
end;

procedure TScenarioLayer.Clear; stdcall;
begin
  DeleteQueues;
  FChildrenPlugins.Clear;
  FDictGuidChildren.Clear;
  FStartPlugins.Clear;
  FQueuePlugins.Clear;
  FStartOutputConnections.Clear;
  FChildren.Clear;
  FChildrenLayers.Clear;
end;

constructor TScenarioLayer.Create(AOwner: TScenarioLayer);
var
  l: TScenarioLayer;
begin
  inherited;
  FWidth := 300;
  FHeight := 250;
  FChildren := TList<IScenarioNode>.Create;
  FChildrenPlugins := TList<TScenarioPlugin>.Create;
  FChildrenLayers := TList<TScenarioLayer>.Create;
  FDictGuidChildren := TDictionary<TGuid, TInterfacedObject>.Create;
  FStartPlugins := TList<TScenarioPlugin>.Create;
  FForceStartPluginsOn := TList<TScenarioPlugin>.Create;
  FForceStartPluginsOff := TList<TScenarioPlugin>.Create;
  FQueuePlugins := TList<TScenarioPlugin>.Create;
  FStartOutputConnections := TList<PRealTimeDataStream>.Create;
  FQueuesDict := TDictionary<TQueueKey, TQueueRam>.Create;
  FQueues := TList<TQueueRam>.Create;
  FLog := TCinemaLog.Create;
  FRequestsRedrawPlugin := TList<TScenarioPlugin>.Create;
  FLastRedrawViewPlugins := TThread.GetTickCount64;
  l := AOwner;
  while Assigned(l) do
  begin
    if l is TCinemaScenario then
    begin
      FScenario := TCinemaScenario(l);
      break;
    end;
    l := l.Owner;
  end;
end;

class constructor TScenarioLayer.Create;
begin
  FNodeImplementators := TDictionary<TGUID, TScenarioNodeClass>.Create;
  FNodeJsonContainerNames := TDictionary<TScenarioNodeClass, WideString>.Create;
end;

function TScenarioLayer.CreateChild(const ANodeType: TGUID): {$ifdef NARI} Pointer {$else} IScenarioNode {$endif}; stdcall;
var
  node: IScenarioNode;
begin
  node := DoCreateChild(ANodeType);
  Result := {$ifdef NARI}Pointer{$endif}(node);
end;

procedure TScenarioLayer.DeleteQueues;
var
  i: Integer;
begin
  for i := 0 to FQueues.Count - 1 do
    FQueues.Items[i].Free;
  FQueues.Clear;
  FQueuesDict.Clear;
end;

destructor TScenarioLayer.Destroy;
begin
  DeleteQueues;
  FreeAndNil(FEventOnQueue);
  FChildren.Free;
  FChildrenPlugins.Free;
  FChildrenLayers.Free;
  FQueuesDict.Free;
  FStartPlugins.Free;
  FForceStartPluginsOn.Free;
  FForceStartPluginsOff.Free;
  FQueuePlugins.Free;
  FDictGuidChildren.Free;
  FStartOutputConnections.Free;
  FRequestsRedrawPlugin.Free;
  inherited Destroy;
end;

class destructor TScenarioLayer.Destroy;
begin
  FNodeImplementators.Free;
  FNodeJsonContainerNames.Free;
end;

procedure TScenarioLayer.DestroyStartProcessers;
var
  i: Integer;
  child: IScenarioNode;
  layer: IScenarioLayer;
begin
  System.TMonitor.Enter(Self);
  try
    State := TNodeState.nsStopping;
    // освобождаем стартовые процессоры
    while FStartPlugins.Count > 0 do
    begin
      DestroyStartProcessor(FStartPlugins.Items[FStartPlugins.Count - 1]);
    end;

    FStartPlugins.Count := 0;
    for i := 0 to CountChildren - 1 do
    begin
      child := Children[i];
      if Supports(child, IScenarioLayer, layer) then
      begin
        if not layer.IsThread then
          (layer as TScenarioLayer).DestroyStartProcessers;
      end else
      //if Supports(child, IScenarioPlugin, plugin) then
        child.State := TNodeState.nsNone;
    end;
    State := TNodeState.nsNone;
  finally
    System.TMonitor.Exit(Self);
  end;
end;

procedure TScenarioLayer.DestroyStartProcessor(APlugin: TScenarioPlugin);
var
  i: Integer;
  plugin: TScenarioPlugin;
  dataStream: PRealTimeDataStream;
begin
  for i := FStartPlugins.Count - 1 downto 0 do
  begin
    plugin := FStartPlugins.Items[i];
    if plugin = APlugin then
    begin
      dataStream := FStartOutputConnections.Items[i];
      FStartOutputConnections.Delete(i);
      FStartPlugins.Delete(i);
      plugin.DestroyDataStream(dataStream);
      break;
    end;
  end;
end;

procedure TScenarioLayer.CreateStartProcessor(APlugin: TScenarioPlugin);
var
  dataStream: PRealTimeDataStream;
  scenario: ICinemaScenario;
begin
  dataStream := APlugin.CreateDataStream(nil, 0);
  FStartOutputConnections.Add(dataStream);

  // добавляем корневые атрибуты к потоку данных - время запуска, имя сценария (слоя), IP-хоста
  APlugin.AddDataStreamAttribute(dataStream, csAttrDateTimeScenarioStart, @FDateTimeStart, SizeOf(FDateTimeStart));

  scenario := GetCinemaScenario;
  if Assigned(scenario) then
    APlugin.AddDataStreamAttribute(dataStream, csAttrScenarioName, @scenario.ScenarioName[0], Length(scenario.ScenarioName)*2);

  if length(LocalHostName) > 0 then
    APlugin.AddDataStreamAttribute(dataStream, csAttrHostScenarioExecution, @LocalHostName[1], Length(LocalHostName)*2);
end;

procedure TScenarioLayer.Execute;
const
  PERIOD_EMPTY_REQ_SLEEP = 3000;
var
  i: Integer;
  plugin: TScenarioPlugin;
begin

  State := TNodeState.nsStarting;
  FStartOutputConnections.Count := 0;
  FLastUpdateNoDataState := TThread.GetTickCount64;
  if FQueues.Count > 0 then
  begin
    if not Assigned(FEventOnQueue) then
      FEventOnQueue := TEvent.Create;
  end else
    FreeAndNil(FEventOnQueue);

  FDateTimeStart := now;
  try
    ///  На старте создаём т.н. процессоры стартовых плагинов. Процессор - внутренний класс-обработчик реализованный плагином,
    ///  его назначение - обработка или получение данных из своих источников (сеть, файловая система, и т.д. и т.п.)
    ///  и отправка результата на выход. Так же стартовыми элементами может быть любой плагин, если это оправданно по сценарию,
    ///  например, Delay, который выполняет задержку (Sleep) выполнения потока и позволяет создать сценарий, выполняющий
    ///  работу с определённой (настраиваемой) периодичностью
    for i := 0 to FStartPlugins.Count - 1 do
    begin
      CreateStartProcessor(FStartPlugins.Items[i]);
    end;

  except
    on e: Exception do
      DoSendNodeMessage(Self, 'Возникла неизвестная ошибка при запуске потока: ' + WideString(e.Message), MESSAGE_ERROR);
  end;

  State := TNodeState.nsWorking;
  repeat

    try

      // в реальном времени добавился как стартовый (будет опрашиваться рабочим потоком напрямую, а не по указке других плагинов)
      if FForceStartPluginsOn.Count > 0 then
        FlushStartPluginsOn;

      if FForceStartPluginsOff.Count > 0 then
        FlushStartPluginsOff;

      for i := 0 to FStartPlugins.Count - 1 do
      begin
        plugin := FStartPlugins.Items[i];
        plugin.ProcessData(FStartOutputConnections.Items[i], nil, 0);
      end;

      if not AskQueues and (FStartPlugins.Count = 0) and (TThread.GetTickCount64 - FLastTimeReadQueue >= PERIOD_EMPTY_REQ_SLEEP) then
      begin
        SetDataWaiting(true);
      end;

      ServiceRoutines;

    except
      on e: Exception do
        DoSendNodeMessage(Self, 'Возникла неизвестная ошибка при выполнении потока: ' + WideString(e.Message), MESSAGE_ERROR);
    end;
  until FWorkThread.Terminated;

  while AskQueues do;

  // при нормальном завершении потока выполнение метода дойдёт сюда
  DestroyStartProcessers;

  FreeAndNil(FEventOnQueue);
end;

procedure TScenarioLayer.FlushStartPluginsOff;
var
  plugin: TScenarioPlugin;
begin
  System.TMonitor.Enter(FForceStartPluginsOff);
  try
    while FForceStartPluginsOff.Count > 0 do
    begin
      plugin := FForceStartPluginsOff.Items[FForceStartPluginsOn.Count - 1];
      FForceStartPluginsOff.Delete(FForceStartPluginsOff.Count - 1);
      DestroyStartProcessor(plugin);
    end;
  finally
    System.TMonitor.Exit(FForceStartPluginsOff);
  end;
end;

procedure TScenarioLayer.FlushStartPluginsOn;
var
  plugin: TScenarioPlugin;
begin
  System.TMonitor.Enter(FForceStartPluginsOn);
  try
    while FForceStartPluginsOn.Count > 0 do
    begin
      plugin := FForceStartPluginsOn.Items[FForceStartPluginsOn.Count - 1];
      FForceStartPluginsOn.Delete(FForceStartPluginsOn.Count - 1);
      {$ifdef fpc}
      if FStartPlugins.IndexOf(plugin) < 0 then
      {$else}
      if FStartPlugins.IndexOfItem(plugin, TDirection.FromEnd) < 0 then
      {$endif}
      begin
        FStartPlugins.Add(plugin);
        CreateStartProcessor(plugin);
      end;
    end;
  finally
    System.TMonitor.Exit(FForceStartPluginsOn);
  end;
end;

procedure TScenarioLayer.GetChangedStates; stdcall;
var
  child: IScenarioNode;
  layer: IScenarioLayer;
begin
  if State <> TNodeState.nsNone then
  begin
    FRequareChangedStates := true;
    for child in FChildren do
    begin
      if Supports(child, IScenarioLayer, layer) then
        (layer as TScenarioLayer).GetChangedStates;
    end;
  end;
end;

function TScenarioLayer.GetChildByGuid(const AGuid: TGuid): {$ifdef NARI} Pointer {$else} IScenarioNode {$endif}; stdcall;
var
  node: IScenarioNode;
begin
  node := DoGetChildByGuid(AGuid);
  Result := {$ifdef NARI}Pointer{$endif}(node);
end;

function TScenarioLayer.GetCinemaScenario: ICinemaScenario;
var
  item: TScenarioLayer;
begin
  item := Self;
  while Assigned(item) do
  begin
    if supports(item, ICinemaScenario, Result) then
      exit;
    item := item.FParent;
  end;
  Result := nil;
end;

function TScenarioLayer.GetIsOpened: boolean; stdcall;
begin
  Result := FIsOpened;
end;

function TScenarioLayer.GetIsThread: Boolean; stdcall;
begin
  Result := FIsThread;
end;

function TScenarioLayer.GetIsThreadTerminated: boolean; stdcall;
begin
  if Assigned(FWorkThread) then
    Result := FWorkThread.Terminated
  else
  if Assigned(Parent) then
    Result := Parent.IsThreadTerminated
  else
    Result := true;
end;

function TScenarioLayer.GetLog: {$ifdef NARI} Pointer {$else} ICinemaLog {$endif}; stdcall;
var
  log: ICinemaLog;
begin
  log := DoGetLog;
  Result := {$ifdef NARI}Pointer{$endif}(log);
end;

function TScenarioLayer.GetNodeType: TGuid;
begin
  Result := IScenarioLayer;
end;

function TScenarioLayer.GetPipeAtPoint(X, Y: Integer): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}; stdcall;
var
  conn: IScenarioConnection;
begin
  conn := DoGetPipeAtPoint(X, Y);
  Result := {$ifdef NARI}Pointer{$endif}(conn);
end;

function TScenarioLayer.GetQueue(AQueueKey: TQueueKey): TQueueRam;
begin
  if not FQueuesDict.TryGetValue(AQueueKey, Result) then
  begin
    Result := TQueueRam.Create;
    Result.OnWrite := OnWriteQueue;
    FQueuesDict.Add(AQueueKey, Result);
    FQueues.Add(Result);
  end;
end;

function TScenarioLayer.GetQueueSize: Integer;
var
  queue: TQueueRam;
begin
  Result := 0;
  for queue in FQueues do
  begin
    Result := Result + queue.Count;
  end;
end;

function TScenarioLayer.GetChildren(AIndex: Integer): {$ifdef NARI} Pointer {$else} IScenarioNode {$endif}; stdcall;
var
  node: IScenarioNode;
begin
  node := DoGetChildren(AIndex);
  Result := {$ifdef NARI}Pointer{$endif}(node);
end;

function TScenarioLayer.GetChildrenAtPoint(X, Y: Integer; AIncludOpenedLayers: boolean): {$ifdef NARI} Pointer {$else} IScenarioNode {$endif}; stdcall;
var
  node: IScenarioNode;
begin
  node := DoGetChildrenAtPoint(X, Y, AIncludOpenedLayers);
  Result := {$ifdef NARI}Pointer{$endif}(node);
end;

function TScenarioLayer.GetCountChildren: Integer; stdcall;
begin
  Result := FChildren.Count;
end;

function TScenarioLayer.GetRectChilds: PRect; stdcall;
var
  i: Integer;
  rectChild: TRect;
begin

  if Assigned(FChildren) and (FChildren.Count > 0) then
    FRectChilds := FChildren.Items[0].Rect^
  else
    FillChar(FRectChilds, SizeOf(TRect), 0);

  if Assigned(FChildren) then
    for i := 1 to FChildren.Count - 1 do
    begin
      rectChild := FChildren.Items[i].Rect^;
      FRectChilds.Left := min(rectChild.Left, FRectChilds.Left);
      FRectChilds.Top := min(rectChild.Top, FRectChilds.Top);
      FRectChilds.Right := max(rectChild.Right, FRectChilds.Right);
      FRectChilds.Bottom := max(rectChild.Bottom, FRectChilds.Bottom);
    end;
  Result := @FRectChilds;
end;

procedure TScenarioLayer.LinkConnections;
var
  i, j: int32;
  processor: IDataStreamProcessor;
  connection: IScenarioConnection;
begin
  for i := 0 to CountChildren - 1 do
  begin
    if Supports(DoGetChildren(i), IDataStreamProcessor, processor) then
    begin
      for j := processor.CountOutputConnections - 1 downto 0 do
      begin
        connection := processor.OutputConnections[j];
        if not Assigned(connection) then
          continue;
        connection.Source;
        connection.Destination;
      end;
      for j := processor.CountInputConnections - 1 downto 0 do
      begin
        connection := processor.InputConnections[j];
        if not Assigned(connection) then
          continue;
        connection.Source;
        connection.Destination;
      end;
    end;
  end;
end;

procedure TScenarioLayer.LoadFormJson(const AValues: TJsonCinema; const AParentLayer: IScenarioLayer);
var
  i: int32;
  item: TJsonCinema;
  values: TJsonCinema;
  layer: TScenarioLayer;
  ilayer: IScenarioLayer;
  pipe: TScenarioConnection;
  plugin: TScenarioPlugin;
  node: IScenarioNode;
  result: boolean;
begin
  inherited;
  FIsLoading := true;
  try
    IsThread := AValues.GetValue<Boolean>('thread');
    IsOpened := AValues.GetValue<Boolean>('IsOpened');
    Parent := AParentLayer;

    // плагины на текущем слое
    result := AValues.TryGetValue<TJsonCinema>('plugins', item);
    if result and Assigned(item) then
    begin
      for i := 0 to TJSONArray(item).Count - 1 do
      begin
        values := TJSONArray(item).Items[i];
        plugin := TScenarioPlugin.Create(Self);
        plugin.LoadFormJson(values, Self);
      end;
    end;

    // вложенный слой
    result := AValues.TryGetValue<TJsonCinema>('layers', item);
    if result and Assigned(item) then
    begin
      for i := 0 to TJSONArray(item).Count - 1 do
      begin
        values := TJSONArray(item).Items[i];
        layer := TScenarioLayer.Create(Self);
        layer.LoadFormJson(values, Self);
      end;
    end;

    // связи/каналы м/у плагинами
    result := AValues.TryGetValue<TJsonCinema>('connections', item);
    if result and Assigned(item) then
    begin
      for i := 0 to TJSONArray(item).Count - 1 do
      begin
        values := TJSONArray(item).Items[i];
        pipe := TScenarioConnection.Create(Self);
        pipe.LoadFromJson(values, Self);
      end;
    end;

    // в конце ещё раз пробегаемся по новым слоям и заставляем слинковать связи по загруженным гуидам
    for i := 0 to CountChildren - 1 do
    begin
      node := Children[i];
      if Supports(node, IScenarioLayer, ilayer) then
        (ilayer as TScenarioLayer).LinkConnections;
    end;

  finally
    FIsLoading := false;
  end;
end;

procedure TScenarioLayer.Lock; stdcall;
begin
  System.TMonitor.Enter(Self);
end;

procedure TScenarioLayer.OnChangeNodeState(ASourceNode: TScenarioNode);
begin
  if ASourceNode <> Self then
    CalculateStates;
  inherited;
end;

procedure TScenarioLayer.OnSetForceAsStartRealtime(APlugin: TScenarioPlugin);
begin
  if FIsThread then
  begin
    if APlugin.AsStart then
    begin
      System.TMonitor.Enter(FForceStartPluginsOn);
      try
        if FForceStartPluginsOn.IndexOf(APlugin) < 0 then
          FForceStartPluginsOn.Add(APlugin);
      finally
        System.TMonitor.Exit(FForceStartPluginsOn);
      end;
    end else
    begin
      System.TMonitor.Enter(FForceStartPluginsOff);
      try
        if FForceStartPluginsOff.IndexOf(APlugin) < 0 then
          FForceStartPluginsOff.Add(APlugin);
      finally
        System.TMonitor.Exit(FForceStartPluginsOff);
      end;
    end;
  end else
  if Assigned(FOwner) then
    FOwner.OnSetForceAsStartRealtime(APlugin);
end;

procedure TScenarioLayer.OnUpdateGuid(const AChild: TScenarioNode; OldGuid, NewGuid: TGuid);
begin
  if not OldGuid.IsEmpty then
    FDictGuidChildren.Remove(OldGuid);
  FDictGuidChildren.Add(NewGuid, AChild);
  if Assigned(FOwner) then
    FOwner.OnUpdateGuid(AChild, OldGuid, NewGuid);
end;

procedure TScenarioLayer.OnWriteQueue(AQueue: TQueueRam);
begin
  if Assigned(FEventOnQueue) then
    SetDataWaiting(false);
end;

procedure TScenarioLayer.RedrawViewPlugins;
var
  i: Integer;
begin
  FLastRedrawViewPlugins := FTimer;
  for i := 0 to FRequestsRedrawPlugin.Count - 1 do
    FRequestsRedrawPlugin.Items[i].RedrawView;

  FRequestsRedrawPlugin.Count := 0;
end;

class procedure TScenarioLayer.RegisterNode(const ANodeType: TGUID; const ANodeImplementation: TScenarioNodeClass; const AJsonContainerName: WideString);
begin
  FNodeImplementators.Add(ANodeType, ANodeImplementation);
  FNodeJsonContainerNames.Add(ANodeImplementation, AJsonContainerName);
end;

procedure TScenarioLayer.RemoveChild(const AChild: TScenarioNode);
begin
  inherited;
  if AChild.Owner = Self then
  begin
    FChildren.Remove(AChild);
    if AChild.InheritsFrom(TScenarioPlugin) then
      FChildrenPlugins.Remove(TScenarioPlugin(AChild));
  end;

  FDictGuidChildren.Remove(AChild.GUID^);
  if Assigned(FOwner) then
    FOwner.RemoveChild(AChild);
end;

procedure TScenarioLayer.RequestRedrawView(APlugin: TScenarioPlugin);
begin
  FRequestsRedrawPlugin.Add(APlugin);
end;

procedure TScenarioLayer.ResetStringBuffer;
var
  child: IScenarioNode;
begin
  inherited;
  for child in FChildren do
    (child as TScenarioNode).ResetStringBuffer;
end;

function TScenarioLayer.Save: PWideChar; stdcall;
var
  child: IScenarioNode;
  nodes: TDictionary<WideString, WideString>;
  pair: TPair<WideString, WideString>;
  nodesJson: WideString;
  connection: WideString;
  connections: WideString;
  jsonContainerName: WideString;
  processor: IDataStreamProcessor;
  pipe: IScenarioConnection;
  i: Integer;
begin
  inherited Save;
  FIsSaving := true;
  try
    FStringBuffer := FStringBuffer + ',' + csCrossLineBreak +
      '"thread": "' + WideString(BoolToStr(FIsThread, true)) + '",' + csCrossLineBreak +
      '"IsOpened": "' + WideString(BoolToStr(FIsOpened, true)) + '"'
      ;

    nodes := TDictionary<WideString, WideString>.Create;
    try
      connections := '';

      for child in FChildren do
      begin
        if not FNodeJsonContainerNames.TryGetValue(TScenarioNodeClass((child as TScenarioNode).ClassType), jsonContainerName) then
          continue; // raise

        nodesJson := '';
        nodes.TryGetValue(jsonContainerName, nodesJson);

        if nodesJson = '' then
          nodesJson := '{' + sLineBreak + (child as TScenarioNode).Save + csCrossLineBreak + '}'
        else
          nodesJson := nodesJson + ',' + csCrossLineBreak + '{' + csCrossLineBreak + (child as TScenarioNode).Save + csCrossLineBreak + '}';

        nodes.AddOrSetValue(jsonContainerName, nodesJson);

        if Supports(child, IDataStreamProcessor, processor) then
        begin

          for i := 0 to processor.CountOutputConnections - 1 do
          begin
            pipe := processor.OutputConnections[i];
            if not Assigned(pipe) then
              continue;

            connection := pipe.Save;
            if connection = '' then
              continue;

            if connections = '' then
              connections := '{' + csCrossLineBreak + connection + csCrossLineBreak + '}'
            else
              connections := connections + ',' + csCrossLineBreak + '{' + csCrossLineBreak + connection + csCrossLineBreak + '}';
          end;

        end;
      end;

      for pair in nodes do
      begin
        FStringBuffer := FStringBuffer + ',' + csCrossLineBreak + '"' + pair.Key + '": [' + pair.Value + ']';
      end;

      if connections <> '' then
        FStringBuffer := FStringBuffer + ',' + csCrossLineBreak + '"connections": [' + connections + ']';

      Result := PWideChar(FStringBuffer);

    finally
      nodes.Free;
    end;
  finally
    FIsSaving := false;
  end;
end;

function TScenarioLayer.DoSendNodeMessage(ASourceNode: TScenarioNode; AMsg: WideString; AMessageType: TMessageType): boolean;
begin
  Result := inherited;
  if (AMessageType <= MESSAGE_ERROR) and FLog.AddMessage(PWideChar(AMsg), ASourceNode.NodeKey, ASourceNode.GUID^, AMessageType) then
    Result := true;
end;

function TScenarioLayer.DoGetChildren(AIndex: Integer): IScenarioNode;
begin
  if Assigned(FChildren) then
    Result := FChildren.Items[AIndex]
  else
    Result := nil;
end;

function TScenarioLayer.DoGetChildByGuid(const AGuid: TGuid): IScenarioNode;
var
  node: TInterfacedObject;
begin
  Result := nil;
  if (FDictGuidChildren.TryGetValue(AGuid, node)) then
    Supports(node, IScenarioNode, Result);

  if not Assigned(Result) and Assigned(FOwner) then
    Result := FOwner.GetChildByGuid(AGuid);
end;

function TScenarioLayer.DoGetChildrenAtPoint(X, Y: Integer; AIncludOpenedLayers: boolean): IScenarioNode;
var
  i: Integer;
  child: IScenarioNode;
  layer: IScenarioLayer;
  r: TRect;
begin

  for i := 0 to CountChildren - 1 do
  begin
    child := Children[i];
    r := child.Rect^;
    if r.Contains(TPoint.Create(X, Y)) then
    begin
      if supports(child, IScenarioLayer, layer) then // вложенный слой
        if layer.IsOpened then
        begin
          // проверяем попадание в рамку открытого слоя; в этом случае выбираем его
          if ((r.Left + OPEN_LAYER_BORDER_SIZE > X) or (r.Right - OPEN_LAYER_BORDER_SIZE <= X)) or
             ((r.Top + OPEN_LAYER_BORDER_SIZE > Y) or (r.Bottom - OPEN_LAYER_BORDER_SIZE <= Y)) then
                exit(child);

          Result := layer.GetChildrenAtPoint(X, Y, AIncludOpenedLayers);
          if not Assigned(Result) and AIncludOpenedLayers then
            exit(child);

          exit;
        end;

      exit(child);
    end else
    if supports(child, IScenarioLayer, layer) then // вложенный слой; даже если в него не попали X и Y авсё равно поищем его объекты, т.к. они могут быть за пределами его границ
    begin
      Result := layer.GetChildrenAtPoint(X, Y, AIncludOpenedLayers);
      if Assigned(Result) then
        exit;
    end;
  end;

  Result := nil;
end;

function TScenarioLayer.DoGetLog: ICinemaLog;
begin
  Result := FLog;
end;

function TScenarioLayer.DoGetPipeAtPoint(X, Y: Integer): IScenarioConnection;
var
  i, j: Integer;
  child: IDataStreamProcessor;
  pipe: IScenarioConnection;
begin

  for i := 0 to CountChildren - 1 do
  begin
    if not supports(DoGetChildren(i), IDataStreamProcessor, child) then
      continue;

    for j := 0 to CountInputConnections - 1 do
    begin
      pipe := child.InputConnections[j];
      if pipe.HitTest(X, Y) then
        exit(pipe);
    end;

    for j := 0 to child.CountOutputConnections - 1 do
    begin
      pipe := child.OutputConnections[j];
      if pipe.HitTest(X, Y) then
        exit(pipe);
    end;

  end;

  Result := nil;
end;

function TScenarioLayer.DoCreateChild(const ANodeType: TGUID): IScenarioNode;
var
  nodeImplementation: TScenarioNodeClass;
begin
  if FNodeImplementators.TryGetValue(ANodeType, nodeImplementation) then
  begin
    Result := nodeImplementation.Create(Self);
    Result.Parent := Pointer(Self as IScenarioLayer);
  end else
    Result := nil;
end;

procedure TScenarioLayer.DoStop;
const
  STEP_SLEEP = 100;
  TIMEOUT_WAIT = 5000;
begin
  if Assigned(FStopTask) and (FStopTask.Status < TTaskStatus.Completed) then
    exit;
  FStopTask := TTask.Create(procedure
  var
    layer: IScenarioLayer;
    child: IScenarioNode;
    countTime: Integer;
  begin
    for child in FChildren do
    begin
      if Supports(child, IScenarioLayer, layer) then
      begin
        (layer as TScenarioLayer).DoStop;
      end;
    end;

    if Assigned(FWorkThread) then
    begin
        if (FStartPlugins.Count = 0) then
        begin
          SetDataWaiting(false);
        end;

        if not FWorkThread.Terminated then
          FWorkThread.Terminate;
        countTime := 0;
        while (FState <> TNodeState.nsNone) and (countTime < TIMEOUT_WAIT) do
        begin
          Sleep(STEP_SLEEP);
          inc(countTime, STEP_SLEEP);
        end;
        if FState = TNodeState.nsNone then
        begin
          FWorkThread.WaitFor;
          FreeAndNil(FWorkThread);
        end;
    end;
  end);
  FStopTask.Start;
end;

procedure TScenarioLayer.SendChangedStates;

  procedure Send(const ALayer: IScenarioLayer);
  var
    i: int32;
    node: IScenarioNode;
    layer: IScenarioLayer;
  begin
    if ALayer.QueryInterface(IScenarioNode, node) = S_OK then
      OnChangeNodeState(node as TScenarioNode);
    for i := 0 to ALayer.CountChildren - 1 do
    begin
      node := ALayer.Children[i];

      if node.State <> TNodeState.nsNone then
      begin
        OnChangeNodeState(node as TScenarioNode);
        if node.QueryInterface(IScenarioLayer, layer) = S_OK then
          Send(layer);
      end;
    end;
  end;

begin
  FRequareChangedStates := false;
  if State <> TNodeState.nsNone then
    Send(Self);
end;

procedure TScenarioLayer.ServiceRoutines;
begin
  FTimer := TThread.GetTickCount64;

  if (FRequestsRedrawPlugin.Count > 0) and (FTimer - FLastRedrawViewPlugins > 1000) then
    RedrawViewPlugins;

  if FRequareChangedStates then
    SendChangedStates;

  if FTimer - FLastUpdateNoDataState > TCinemaConfiguration.TimeoutForStateNoData then
    UpdateNoDataState;

end;

procedure TScenarioLayer.SetIsOpened(const AValue: boolean); stdcall;
begin
  FIsOpened := AValue;
end;

procedure TScenarioLayer.SetIsThread(const AValue: Boolean); stdcall;
begin
  FIsThread := AValue;
end;

procedure TScenarioLayer.SetDataWaiting(AWaiting: boolean);
begin
  if AWaiting then
  begin
    // секция в которую заходит только свой же поток
    { самовольное засыпание потока }
    if not TMonitor.TryEnter(FEventOnQueue) then
      exit;
    try
      if QueueSize > 0 then
        exit;
      FEventOnQueue.ResetEvent;
    finally
      TMonitor.Exit(FEventOnQueue);
    end;
    FEventOnQueue.WaitFor(INFINITE);
  end else
  begin
    { внешнее воздействие на пробуждение }

    System.TMonitor.Enter(FEventOnQueue);
    try
      FEventOnQueue.SetEvent;
    finally
      System.TMonitor.Exit(FEventOnQueue);
    end;
  end;
end;

function TScenarioLayer.TryLock: boolean; stdcall;
begin
  Result := System.TMonitor.TryEnter(Self);
end;

function TScenarioLayer.TryStart: boolean; stdcall;
var
  i, j: Integer;
  child: TScenarioNode;
  plugin: TScenarioPlugin;
  layer: IScenarioLayer;
  qk: TQueueKey;
  wasAdded: boolean;
begin
  if State > nsNone then
    exit(false);
  Result := false;
  FStartPlugins.Count := 0;
  // найдём стартовые плагины - те которые не имеют входных каналов
  for i := 0 to FChildren.Count - 1 do
  begin
    child := FChildren[i] as TScenarioNode;
    if child.InheritsFrom(TScenarioPlugin) then
    begin
      plugin := child as TScenarioPlugin;
      // попутно загружаем плагины у которых включена опция LoadBeforeStartScenario
      if plugin.LoadBeforeStartScenario then
        (plugin as TScenarioPlugin).CreatePlugin;

      if (plugin.CountInputConnections = 0) or plugin.AsStart then
      begin
        FStartPlugins.Add((plugin as TScenarioPlugin));
      end;

      if (plugin.CountInputConnections > 0) then
      begin
        // входные связи есть, тогда проверим что плагин получает данные из другого потока, соотв-но должен забирать из входной очереди,
        // поэтому нужно будет опрашивать эти очереди
        wasAdded := false;
        for j := 0 to plugin.CountInputConnections - 1 do
        begin
          layer := (IDataStreamProcessor(plugin.InputConnections[j].Source) as TDataStreamProcessor).DoGetThreadLayer as TScenarioLayer;
          if (layer.Guid^ <> FGuid) and (layer.IsThread or IsThread) then
          begin
            // одна очередь между двумя потокам (если есть связи в обоих направлениях - тогда две), сколькобы связй небыло между потоками
            // qk.IDSource := layer.GUID;
            // qk.IDDest := FGuid;

            // одна очередь из потока одного потока (если есть связи в обоих направлениях - тогда две), сколькобы связй небыло
            // между элементами потока-источникаи текущим плагином plugin
            qk.IDSource := layer.Guid^;
            qk.IDDest := plugin.Guid^;
            if not wasAdded then
            begin
              FQueuePlugins.Add(plugin);
              wasAdded := true;
            end;
            plugin.InputConnections[j].Queue := GetQueue(qk);
            TQueueRam(plugin.InputConnections[j].Queue).Owner := plugin;
          end;
        end;
      end;
    end else
    if Supports(child, IScenarioLayer, layer) then
      if layer.TryStart then
        Result := true;
  end;

  if (FStartPlugins.Count > 0) or (FQueues.Count > 0) then
  begin
    Result := true;
    if FIsThread then
      FWorkThread := TWorkThread.Create(Self);
  end;
end;

function TScenarioLayer.TryStop: boolean; stdcall;
begin

  DoStop;

  while HaveRunStopTask do
    sleep(1);

  Result := FState = TNodeState.nsNone;
end;

function TScenarioLayer.TryTerminate: boolean; stdcall;
const
  TIMEOUT = 5000;
var
//  threadTerminate: TWorkTerminationThread;
  child: IScenarioNode;
  layer: IScenarioLayer;
  counter: Integer;
begin
  Result := true;
  // пытаемся остановиться мягким способом
  if TryStop then
    exit;

  // по хорошему не вышло, останавливаемся с жёстким прерыванием потоков
  for child in FChildren do
  begin
    if Supports(child, IScenarioLayer, layer) then
      if not layer.TryTerminate then
        Result := false;
  end;

  if (FStartPlugins.Count > 0) then
  begin
    if Assigned(FWorkThread) then
    begin

      {$ifdef MSWINDOWS}
      TerminateThread(FWorkThread.Handle, 0);
      {$endif}

      FreeAndNil(FWorkThread);
    end;
    // возможно сделать отдельным потоком для предотвращения зависания основного если зависнет один из плагинов слоя
    DestroyStartProcessers;
  end;

  // обрабатываем данные которые могли остаться в очередях
  counter := 0;
  while AskQueues and (counter < TIMEOUT) do
  begin
    Sleep(1);
    inc(counter);
  end;

  // если в очередях были данные - рано утверждать что слой завершил работу
  if AskQueues or (FState <> TNodeState.nsNone) then
    Result := false;
end;

procedure TScenarioLayer.Unlock; stdcall;
begin
  System.TMonitor.Exit(Self);
end;

procedure TScenarioLayer.UpdateNoDataState;
var
  child: TScenarioPlugin;
  childLayer: TScenarioLayer;
begin
  FLastUpdateNoDataState := FTimer;
  for child in FChildrenPlugins do
  begin
    if (child.State <> TNodeState.nsNone) then
    begin
      if (child.LastTimeProcessData < FTimer) and (FTimer - child.LastTimeProcessData > TCinemaConfiguration.TimeoutForStateNoData) then
      begin
        if child.State <> TNodeState.nsNoData then
          child.State := TNodeState.nsNoData;
      end else
      if child.State <> TNodeState.nsWorking then
        child.State := TNodeState.nsWorking;
    end;
  end;

  for childLayer in FChildrenLayers do
  begin
    if childLayer.IsThread then
      continue;
    childLayer.Timer := FTimer;
    childLayer.UpdateNoDataState;
  end;

end;

procedure TScenarioLayer.UpdateNodeKey;
var
  child: IScenarioNode;
begin
  inherited;
  for child in FChildren do
  begin
    (child as TScenarioNode).UpdateNodeKey;
  end;
end;

{ TScenarioLayer.TWorkThread }

constructor TScenarioLayer.TWorkThread.Create(AOwner: TScenarioLayer);
begin
  FOwner := AOwner;
  inherited Create(true);
  FreeOnTerminate := false;
end;

destructor TScenarioLayer.TWorkThread.Destroy;
begin

  inherited;
end;

procedure TScenarioLayer.TWorkThread.Execute;
begin
  inherited;
  NameThreadForDebugging(FOwner.Caption, ThreadID);
  FOwner.Execute;
end;

{ TScenarioLayer.TWorkTerminationThread }

constructor TScenarioLayer.TWorkTerminationThread.Create(AOwner: TScenarioLayer);
begin
  FOwner := AOwner;
  inherited Create(false);
  FreeOnTerminate := true;
end;

procedure TScenarioLayer.TWorkTerminationThread.Execute;
begin
  inherited;

end;

{ TPluginTable }

function TPluginTable.AddColumn(AName, ACaption: PWideChar; AType: TCinemaFieldType): Integer;
begin
  FCS.Enter;
  try
    Result := DoAddColumn(AName, ACaption, AType);
  finally
    FCs.Leave;
  end;
end;

function TPluginTable.AddRow: Integer;
begin
  FCS.Enter;
  try
    Result := DoAddRow;
  finally
    FCs.Leave;
  end;
end;

constructor TPluginTable.Create(AOwner: TScenarioPlugin; ACaption, AName: PWideChar);
begin
  FOwner := AOwner;
  FColumns := TList<TColumn>.Create;
  FRows := TList<TRow>.Create;
  FCaption := ACaption;
  FName := AName;
  FCs := SyncObjs.TCriticalSection.Create;
end;

procedure TPluginTable.DeleteColumn(AIndex: int32);
begin
  FColumns.Delete(AIndex);
end;

destructor TPluginTable.Destroy;
begin
  FColumns.Free;
  FRows.Free;
  FCs.Free;
  inherited;
end;

function TPluginTable.DoAddColumn(const AName, ACaption: WideString; AType: TCinemaFieldType): Integer;
var
  column: TColumn;
begin
  Result := FColumns.Count;
  column.Name := AName;
  column.Caption := ACaption;
  column.ColType := AType;
  FColumns.Add(column);
  UpdateRowCountCells;
end;

function TPluginTable.DoAddRow: int32;
var
  row: TRow;
begin
  Result := FRows.Count;
  SetLength(row, FColumns.Count);
  FRows.Add(row);
end;

procedure TPluginTable.DoSetCell(AColumn, ARow: Integer; const AValue: WideString);
begin
  FRows.Items[ARow][AColumn] := AValue;
end;

procedure TPluginTable.DoUpdateColumn(AIndex: int32; const AName, ACaption: WideString; AType: TCinemaFieldType);
var
  column: TColumn;
begin
  column.Name := AName;
  column.Caption := ACaption;
  column.ColType := AType;
  FColumns.Items[AIndex] := column;
end;

function TPluginTable.GetCaption: PWideChar;
begin
  FCS.Enter;
  try
    Result := PWideChar(FCaption);
  finally
    FCs.Leave;
  end;
end;

function TPluginTable.GetCell(AColumn, ARow: Integer): PWideChar;
begin
  FCS.Enter;
  try
    Result := PWideChar(FRows.Items[ARow][AColumn]);
  finally
    FCs.Leave;
  end;
end;

function TPluginTable.GetColumnCaption(AIndex: Integer): PWideChar;
begin
  FCS.Enter;
  try
    Result := PWideChar(FColumns.Items[AIndex].Caption);
  finally
    FCs.Leave;
  end;
end;

function TPluginTable.GetCountColumn: Integer;
begin
  FCS.Enter;
  try
    Result := FColumns.Count;
  finally
    FCs.Leave;
  end;
end;

function TPluginTable.GetCountRows: Integer;
begin
  FCS.Enter;
  try
    Result := FRows.Count;
  finally
    FCs.Leave;
  end;
end;

function TPluginTable.GetJson: PWideChar;
var
  i, j: Integer;
begin
  FCS.Enter;
  try
    FTextResult := '{"MetaInfo":{"Name":"' + FName + '","Caption":"' + FCaption + '","Columns":[';
    for i := 0 to FColumns.Count - 1 do
    begin
      if i > 0 then
        FTextResult := FTextResult + ',';
      FTextResult := FTextResult + '{"Name":"' + FColumns.Items[i].Name + '","Caption":"' + FColumns.Items[i].Caption + '","Type":' + WideString(IntToStr(Integer(FColumns.Items[i].ColType))) + '}';
    end;
    FTextResult := FTextResult + ']},"Rows":[';
    for i := 0 to FRows.Count - 1 do
    begin
      if i > 0 then
        FTextResult := FTextResult + ',{'
      else
        FTextResult := FTextResult + '{';
      for j := 0 to FColumns.Count - 1 do
      begin
        if j > 0 then
          FTextResult := FTextResult + ',';
        FTextResult := FTextResult + '"' + FColumns.Items[j].Name + '":"' + FRows.Items[i][j] + '"';
      end;
      FTextResult := FTextResult + '}';
    end;
    FTextResult := FTextResult + ']}';
    Result := PWideChar(FTextResult);
  finally
    FCs.Leave;
  end;
end;

function TPluginTable.GetName: PWideChar;
begin
  FCS.Enter;
  try
    Result := PWideChar(FName);
  finally
    FCs.Leave;
  end;
end;

function TPluginTable.GetText: PWideChar;
var
  i, j: Integer;
begin
  FCS.Enter;
  try
    FTextResult := FName + csCrossLineBreak + FCaption + csCrossLineBreak;
    for i := 0 to FColumns.Count - 1 do
    begin
      FTextResult := FTextResult + FColumns.Items[i].Name + '#' + FColumns.Items[i].Caption + '#' + WideString(IntToStr(Integer(FColumns.Items[i].ColType))) + csTab;
    end;
    FTextResult := FTextResult + csCrossLineBreak;
    for i := 0 to FRows.Count - 1 do
    begin
      for j := 0 to FColumns.Count - 1 do
        FTextResult := FTextResult + FRows.Items[i][j] + csTab;
      FTextResult := FTextResult + csCrossLineBreak;
    end;
    Result := PWideChar(FTextResult);
  finally
    FCs.Leave;
  end;
end;

procedure TPluginTable.SetCell(AColumn, ARow: Integer; AValue: PWideChar);
begin
  FCS.Enter;
  try
    DoSetCell(AColumn, ARow, AValue);
  finally
    FCs.Leave;
  end;
end;

procedure TPluginTable.SetNeedReload(AValue: Boolean);
begin
  FCS.Enter;
  try
    if AValue then
      FOwner.RequestRedrawView;
  finally
    FCs.Leave;
  end;
end;

procedure TPluginTable.SetText(AText: PWideChar);
var
  {$ifdef FPC}
  sl: TArray<WideString>;
  columns: TArray<WideString>;
  column: TArray<WideString>;
  cells: TArray<WideString>;
  txt: WideString;
  {$else}
  sl: TArray<string>;
  columns: TArray<string>;
  column: TArray<string>;
  cells: TArray<string>;
  txt: string;
  {$endif}
  i, j, count, index: Integer;
begin
  FCS.Enter;
  try
    txt := AText;
    sl := txt.Split([csCrossLineBreak]);
    if Length(sl) < 3 then
      exit;
    index := 0;
    while (index < Length(sl)) and (sl[index] = '') do
      inc(index);
    if Length(sl) < 3 + index then
      exit;
    FName := sl[index];
    FCaption := sl[index+1];
    columns := sl[index+2].Split([csTab]);
    for i := 0 to length(columns) - 1 do
    begin
      column := columns[i].Split(['#']);
      if length(column) <> 3 then
        continue;
      if FColumns.Count <= i then
        DoAddColumn(column[0], column[1], TAttributeType(StrToInt(string(column[2]))))
      else
        DoUpdateColumn(i, column[0], column[1], TAttributeType(StrToInt(string(column[2]))));
    end;

    for i := FColumns.Count - 1 downto length(columns) do
    begin
      DeleteColumn(i);
    end;

    count := 0;

    for i := index + 3 to length(sl) - 1 do
    begin
      cells := sl[i].Split([csTab]);
      if length(cells) < FColumns.Count then
        continue;

      inc(count);

      if FRows.Count <= count then
        DoAddRow;

      for j := 0 to FColumns.Count - 1 do
      begin
        DoSetCell(j, i - 3 - index, cells[j]);
      end;
    end;

    FRows.Count := count;

  finally
    FCs.Leave;
  end;
end;

procedure TPluginTable.UpdateRowCountCells;
var
  i: Integer;
  row: TRow;
begin
  for i := 0 to FRows.Count - 1 do
  begin
    row := FRows.Items[i];
    SetLength(row, FColumns.Count);
    FRows.Items[i] := row;
  end;
end;

{ TCinemaScenario.TStateEvent }

constructor TCinemaScenario.TStateEvent.Create(ASourceNode: TScenarioNode);
begin
  FSource := ASourceNode.Caption;
  FSourceId := ASourceNode.GUID^;
  FState := Integer(ASourceNode.State);
  FHealthState := ASourceNode.HealthState;
  FHealthStateMessage := ASourceNode.HealthStateMessage;
end;

function TCinemaScenario.TStateEvent.GetHealthState: Integer;
begin
  Result := FHealthState;
end;

function TCinemaScenario.TStateEvent.GetHealthStateMessage: PWideChar;
begin
  Result := PWideChar(FHealthStateMessage);
end;

function TCinemaScenario.TStateEvent.GetSource: PWideChar;
begin
  Result := PWideChar(FSource);
end;

function TCinemaScenario.TStateEvent.GetSourceId: System.PGuid;
begin
  Result := @FSourceId;
end;

function TCinemaScenario.TStateEvent.GetState: Integer;
begin
  Result := FState;
end;

{ TCinemaScenario.TMessageEvent }

constructor TCinemaScenario.TMessageEvent.Create(ASource, AMsg: WideString; ASourceGuid: TGuid; AMessageType: TMessageType);
begin
  FSource := ASource;
  FMsg := AMsg;
  FSourceId := ASourceGuid;
  FEventType := AMessageType;
end;

function TCinemaScenario.TMessageEvent.GetEventType: Integer;
begin
  Result := FEventType;
end;

function TCinemaScenario.TMessageEvent.GetMsg: PWideChar;
begin
  Result := PWideChar(FMsg);
end;

function TCinemaScenario.TMessageEvent.GetSource: PWideChar;
begin
  Result := PWideChar(FSource);
end;

function TCinemaScenario.TMessageEvent.GetSourceId: System.PGuid;
begin
  Result := @FSourceId;
end;

{ TValueHolder }

constructor TValueHolder.Create(AData: PByte; ASize: Integer);
begin
  FData := SysGetMem(ASize);
  move(AData^, FData^, ASize);
  FSize := ASize;
end;

destructor TValueHolder.Destroy;
begin
  SysFreeMem(FData);
  inherited;
end;

function TValueHolder.GetSize: Integer;
begin
  Result := FSize;
end;

function TValueHolder.GetValue: PByte;
begin
  Result := FData;
end;

initialization


  InitLocalHost;

  TScenarioLayer.RegisterNode(IScenarioNode, TScenarioNode, 'nodes');
  TScenarioLayer.RegisterNode(IScenarioPlugin, TScenarioPlugin, 'plugins');
  TScenarioLayer.RegisterNode(IScenarioLayer, TScenarioLayer, 'layers');

  IsMultiThread := true;

end.
