unit Cinema.Scenario.Editor;

{$ifdef FPC}
  {$mode Delphi}
{$endif}

interface

uses
    SysUtils
  , Generics.Collections
  , Types
  , Controls
  , Forms
  , StdCtrls
  , Classes
  , Variants
  , Graphics
  , ExtCtrls
  , ActnList
  , ComCtrls
  , Buttons
  , Menus

  , Cinema.Scenario.Interfaces
  , Cinema.Scenario.View
  , Cinema.PluginsDictionary
  ;

type

  { TFrameScenarioEditor }

  TFrameScenarioEditor = class(TFrame)
    acnResize: TAction;
    MenuItem2: TMenuItem;
    sbScenHolst: TScrollBox;
    PopupMenu: TPopupMenu;
    ImageList: TImageList;
    ActionListScenEdit: TActionList;
    acnShowPluginConfig: TAction;
    acnSave: TAction;
    acnNewLayer: TAction;
    acnRename: TAction;
    acnDelete: TAction;
    actnCopy: TAction;
    actnPaste: TAction;
    acnShowPluginInformation: TAction;
    acnLoadBeforeStart: TAction;
    MenuItem1: TMenuItem;
    MenuItem3: TMenuItem;
    MenuItem4: TMenuItem;
    MenuItem5: TMenuItem;
    MenuItem6: TMenuItem;
    MenuItem7: TMenuItem;
    MenuItem8: TMenuItem;
    miLoadBeforeStart: TMenuItem;
    tbScenarioView: TToolBar;
    ToolButton1: TToolButton;
    ToolButton10: TToolButton;
    tbNewLayer: TToolButton;
    ToolButton11: TToolButton;
    ToolButton12: TToolButton;
    ToolButton13: TToolButton;
    ToolButton14: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    ToolButton6: TToolButton;
    ToolButton7: TToolButton;
    ToolButton8: TToolButton;
    ToolButton9: TToolButton;
    procedure acnResizeExecute(Sender: TObject);
    procedure acnShowPluginConfigExecute(Sender: TObject);
    procedure acnSaveExecute(Sender: TObject);
    procedure acnRenameExecute(Sender: TObject);
    procedure actnCopyExecute(Sender: TObject);
    procedure acnNewLayerExecute(Sender: TObject);
    procedure acnDeleteExecute(Sender: TObject);
    procedure actnPasteExecute(Sender: TObject);
    procedure pbHolstMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure pbHolstMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure pbHolstMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure pbHolstPaint(Sender: TObject; Canvas: TCanvas);
    procedure acnShowPluginInformationExecute(Sender: TObject);
    procedure ActionListScenEditUpdate(Action: TBasicAction; var Handled: Boolean);
    procedure acnLoadBeforeStartExecute(Sender: TObject);
    procedure sbScenHolstPaint(Sender: TObject);
  private
    FViewScenario: TViewScenario;
    FPluginsDictionary: TPluginsDictionary;
    FClipBoard: string;
    FOnSelectNode: TOnViewNodeNotify;
    FOnSelectPlugin: TOnSelectPluginNotify;
    FNodesInClipboard: TList<TGuid>;
    FEventListener: TEventListener;
    FOnChangeNode: TNotifyEvent;
    FOnSave: TNotifyEvent;
    procedure EnableDisableTools;
    procedure CalculateScenarioArea;
    procedure Render;
    procedure DoOnSelectPlugin(APlugin: TViewPlugin);
    procedure DoOnSelectNode(ANode: TViewNode<IScenarioNode>);
    procedure OnChangeScenarioNode(Sender: TObject);
    procedure OnLoadLayer(Sender: TObject);
    procedure SetViewScenario(const Value: TViewScenario);
    procedure SetEventListener(const Value: TEventListener);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Redraw;
    property ViewScenario: TViewScenario read FViewScenario write SetViewScenario;
    property PluginsDictionary: TPluginsDictionary read FPluginsDictionary write FPluginsDictionary;
    property OnSelectPlugin: TOnSelectPluginNotify read FOnSelectPlugin write FOnSelectPlugin;
    property OnSelectNode: TOnViewNodeNotify read FOnSelectNode write FOnSelectNode;
    property OnChangeNode: TNotifyEvent read FOnChangeNode write FOnChangeNode;
    property OnSave: TNotifyEvent read FOnSave write FOnSave  ;
    property EventListener: TEventListener read FEventListener write SetEventListener;
  end;

implementation

{$R *.lfm}

uses
  Math,
  Dialogs,
  System.IOUtils,
  {$ifndef fpc}
  {$ifdef posix}
  Posix.Unistd,
  {$endif}
  {$endif}
  Cinema.Utils,
  Cinema.Types,
  Cinema.Configuration,
  Cinema.Dialog.CreateConnection,
  Cinema.DialogString,
  Cinema.Configurator.View,
  Cinema.Interfaces.FuncExport,
  Cinema.Interfaces,
  Cinema.Configurator.View.HelpDialog,
  Cinema.Scenario.Editor.Dialog.Sizes
  ;

{ TFrameScenarioEditor }

procedure TFrameScenarioEditor.acnDeleteExecute(Sender: TObject);
begin
  if Assigned(FViewScenario) and Assigned(FViewScenario.SelectedNode) then
  begin
    FViewScenario.Delete(TViewNode<IScenarioNode>(FViewScenario.SelectedNode.View));
    Redraw;
  end;
end;

procedure TFrameScenarioEditor.acnLoadBeforeStartExecute(Sender: TObject);
begin
  FViewScenario.SelectPlugin.Processor.LoadBeforeStartScenario := not FViewScenario.SelectPlugin.Processor.LoadBeforeStartScenario;
  acnLoadBeforeStart.Checked := FViewScenario.SelectPlugin.Processor.LoadBeforeStartScenario;
  FViewScenario.IsModified := true;
end;

procedure TFrameScenarioEditor.sbScenHolstPaint(Sender: TObject);
begin
  if Assigned(FViewScenario) then
    FViewScenario.Draw(Point(sbScenHolst.HorzScrollBar.Position, sbScenHolst.VertScrollBar.Position));
end;

procedure TFrameScenarioEditor.acnNewLayerExecute(Sender: TObject);
begin
  acnNewLayer.Checked := not acnNewLayer.Checked;
  tbNewLayer.Down := acnNewLayer.Checked;
end;

procedure TFrameScenarioEditor.acnRenameExecute(Sender: TObject);
begin
  if Assigned(FViewScenario) and Assigned(FViewScenario.SelectedNode) then
  begin
    frmStringValue.edStringValue.Text := FViewScenario.SelectedNode.Caption;
    if frmStringValue.ShowModal = mrOk then
    begin
      FViewScenario.SelectedNode.Caption := PWideChar(StringToWide(frmStringValue.edStringValue.Text));
      FViewScenario.IsModified := true;
      Redraw;
    end;
  end;
end;

procedure TFrameScenarioEditor.acnSaveExecute(Sender: TObject);
begin
  if Assigned(FViewScenario) then
  begin
    if FViewScenario.Scenario.ScenarioFullPath = '' then
      FViewScenario.Scenario.SaveScenario(PWideChar(TCinemaConfiguration.RtlScenariosPath + FViewScenario.Scenario.Caption + '.json'))
    else
      FViewScenario.Scenario.SaveScenario(FViewScenario.Scenario.ScenarioFullPath);
    FViewScenario.IsModified := false;
    EnableDisableTools;
    if Assigned(FOnSave) then
      FOnSave(Self);
  end;
end;

procedure TFrameScenarioEditor.acnShowPluginConfigExecute(Sender: TObject);
var
  path: UnicodeString;
  strings: TStringStream;
  outConfig: string;
  {$ifdef LINUX}
  //pid: THandle;
  //command: AnsiString;
  {$endif}
begin
  // возмём из сценария конфигурацию плагина и положим рядом с его конфигуратором
  path := TCinemaConfiguration.DtlAppPath + 'Plugins\' + FViewScenario.SelectPlugin.Processor.Name;

  // если нет конфигуратора, значит есть схема для встроенного конфигуратора
  if not FileExists(path {$ifdef MSWINDOWS} + '.exe'{$endif}) then
  begin
    outConfig := FViewScenario.SelectPlugin.Processor.Configuration;
    // выставим центр окна под курсором
    {$ifdef fpc}
    frmConfigurator.Left := Max(0, Mouse.CursorPos.X - frmConfigurator.Width div 2);
    frmConfigurator.Top := Max(0, Mouse.CursorPos.Y - frmConfigurator.Height div 2);
    {$else}
    frmConfigurator.Left := Max(0, Screen.MousePos.Truncate.X - frmConfigurator.Width div 2);
    frmConfigurator.Top := Max(0, Screen.MousePos.Truncate.Y - frmConfigurator.Height div 2);
    {$endif}
    frmConfigurator.Caption := 'Конфигурация плагина "' + WideToString(FViewScenario.SelectPlugin.Processor.Name) + '" v.' + WideToString(FViewScenario.SelectPlugin.Processor.PluginDescription.Version);
    if frmConfigurator.Configurate(FViewScenario.SelectPlugin.Processor.PluginDescription, outConfig) then
    begin
      FViewScenario.SelectPlugin.Config := outConfig;
    end;

    exit;
  end;

  strings  := TStringStream.Create;
  try
    strings.WriteString(FViewScenario.SelectPlugin.Processor.Configuration);
    strings.SaveToFile(WideToString(path));

    // запускаем конфигуратор плагина и ожидем когда пользователь его закроет
    {$ifdef LINUX}{$endif}FileExecuteWait(WideToString(path) {$ifdef MSWINDOWS} + string('.exe'){$endif});

    {$ifdef LINUX}
    //Application.MainForm.Enabled := false;
//    try
//      command := 'pidof -x ' + AnsiString(path) + ' > /dev/null';
//      while (0 = _system(PAnsiChar(command))) do
//      //while (true) do
//      begin
//        if Assigned(Application.MainForm) then
//        begin
//          Application.ProcessMessages;
//          if Application.Terminated then
//            break;
//        end;
//      end;
//    finally
//      //LayoutMain.Enabled := true;
//    end;
    {$endif}

    // если после себя конфигуратор оставил свой файл конфигурации - забираем его
    if FileExists(path) then
    begin
      strings.Clear;
      strings.LoadFromFile(WideToString(path));
      FViewScenario.SelectPlugin.Config := PWideChar(strings.DataString);
      DeleteFile(path);
    end;

  finally
    strings.Free;
  end;
end;

procedure TFrameScenarioEditor.acnResizeExecute(Sender: TObject);
begin
  if not Assigned(FViewScenario) or not Assigned(FViewScenario.SelectedProcessor) then
    exit;

  frmSetSizes.seWidth.Value := FViewScenario.SelectedProcessor.Width;
  frmSetSizes.seHeight.Value := FViewScenario.SelectedProcessor.Height;
  if frmSetSizes.ShowModal = mrOk then
  begin
    FViewScenario.SelectedProcessor.Resize(frmSetSizes.seWidth.Value, frmSetSizes.seHeight.Value);
    FViewScenario.IsModified := true;
    Redraw;
  end;
end;

procedure TFrameScenarioEditor.actnCopyExecute(Sender: TObject);

  procedure AddGuid(const ANode: IScenarioNode);
  var
    layer: IScenarioLayer;
    i: int32;
  begin
    FNodesInClipboard.Add(ANode.GUID^);
    if (ANode.QueryInterface(IScenarioLayer, layer) = S_OK) then
    begin
      for i := 0 to layer.CountChildren - 1 do
        AddGuid(IScenarioNode(layer.Children[i]));
    end;
  end;

begin
  if not Assigned(FViewScenario) or not Assigned(FViewScenario.SelectedNode) then
    exit;

  FClipBoard := FViewScenario.SelectedNode.Save;
  FNodesInClipboard.Count := 0;
  AddGuid(FViewScenario.SelectedNode);

  actnPaste.Enabled := true;
end;

procedure TFrameScenarioEditor.actnPasteExecute(Sender: TObject);
var
  r: TRectF;
  i: int32;
  mPos: TPoint;
  newGuid: TGuid;
begin
  if not Assigned(FViewScenario) or FClipBoard.IsEmpty then
    exit;

  for i := 0 to FNodesInClipboard.Count - 1 do
  begin
    newGuid := TGuid.NewGuid;
    FClipBoard := FClipBoard.Replace(FNodesInClipboard.Items[i].ToString, newGuid.ToString);
    FNodesInClipboard.Items[i] := newGuid;
  end;

  {$ifdef fpc}
  mPos := Mouse.CursorPos;
  r := sbScenHolst.ClientToScreen(Rect(0, 0, sbScenHolst.Width, sbScenHolst.Height));
  {$else}
  mPos := Screen.MousePos;
  r := pbHolst.AbsoluteRect;
  {$endif}

  FViewScenario.InsertFromData(round(mPos.X - r.Left), round(mPos.Y - r.Top), FClipBoard);
  CalculateScenarioArea;
  Redraw;
end;

procedure TFrameScenarioEditor.pbHolstMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(FViewScenario) then
    FViewScenario.MouseMove(X, Y);
end;

procedure TFrameScenarioEditor.pbHolstMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(FViewScenario) then
    FViewScenario.MouseUp;
end;

procedure TFrameScenarioEditor.acnShowPluginInformationExecute(Sender: TObject);
begin
  frmConfiguratorHelpDialog.ShowHelpDialog(FViewScenario.SelectPlugin.Processor.PluginDescription);
end;

procedure TFrameScenarioEditor.ActionListScenEditUpdate(Action: TBasicAction; var Handled: Boolean);
begin
  Handled := true;
end;

procedure TFrameScenarioEditor.CalculateScenarioArea;
var
  r: TRect;
begin
  if not Assigned(FViewScenario) then
    exit;

  r := FViewScenario.Scenario.RectChilds^;
  r.Left := 0;
  r.Top := 0;
  if r.Height < sbScenHolst.Height then
    r.Bottom := r.Top + round(sbScenHolst.Height);
  if r.Width < sbScenHolst.Width then
    r.Right := r.Left + round(sbScenHolst.Width);

  sbScenHolst.Width := r.Width;
  sbScenHolst.Height := r.Height;
end;

constructor TFrameScenarioEditor.Create(AOwner: TComponent);
begin
  inherited;
  FNodesInClipboard := TList<TGuid>.Create;
end;

destructor TFrameScenarioEditor.Destroy;
begin
  FNodesInClipboard.Free;
  inherited;
end;

procedure TFrameScenarioEditor.EnableDisableTools;
begin
  if not Assigned(FViewScenario) then
    exit;
  acnSave.Enabled := FViewScenario.IsModified;
  acnShowPluginConfig.Enabled := Assigned(FViewScenario) and Assigned(FViewScenario.SelectPlugin) and FViewScenario.SelectPlugin.IsSelected
    and (FileExists(TCinemaConfiguration.DtlAppPathPlugins + FViewScenario.SelectPlugin.Processor.Name + '.exe') or
    (Assigned(FViewScenario.SelectPlugin.Processor.PluginDescription) and Assigned(FViewScenario.SelectPlugin.Processor.PluginDescription.SchemeConfig)));
  acnRename.Enabled := Assigned(FViewScenario) and Assigned(FViewScenario.SelectedNode);
  acnDelete.Enabled := acnRename.Enabled and Assigned(FViewScenario.Scenario) and (FViewScenario.Scenario.State = TNodeState.nsNone);
  actnCopy.Enabled := Assigned(FViewScenario) and Assigned(FViewScenario.SelectedNode);
  actnPaste.Enabled := not FClipBoard.IsEmpty;
  acnShowPluginInformation.Enabled := Assigned(FViewScenario) and Assigned(FViewScenario.SelectPlugin) and Assigned(FViewScenario.SelectPlugin.Processor.PluginDescription);
  acnLoadBeforeStart.Enabled := acnShowPluginInformation.Enabled;
  acnLoadBeforeStart.Checked := Assigned(FViewScenario.SelectPlugin) and FViewScenario.SelectPlugin.Processor.LoadBeforeStartScenario;
  acnResize.Enabled := actnCopy.Enabled;
end;

procedure TFrameScenarioEditor.OnChangeScenarioNode(Sender: TObject);
begin
  Redraw;
  EnableDisableTools;
  if Assigned(FOnChangeNode) then
    FOnChangeNode(Sender);
end;

procedure TFrameScenarioEditor.OnLoadLayer(Sender: TObject);
begin
  EnableDisableTools;
  CalculateScenarioArea;
end;

procedure TFrameScenarioEditor.DoOnSelectNode(ANode: TViewNode<IScenarioNode>);
begin
  EnableDisableTools;
  if Assigned(FOnSelectNode) then
    FOnSelectNode(ANode);
end;

procedure TFrameScenarioEditor.DoOnSelectPlugin(APlugin: TViewPlugin);
begin
  Redraw;
  EnableDisableTools;
  if Assigned(FOnSelectPlugin) then
    FOnSelectPlugin(APlugin);
end;

procedure TFrameScenarioEditor.pbHolstMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  toNode: IDataStreamProcessor;
  fromPlugin, toPlugin: IScenarioPlugin;
  node: IScenarioNode;
  connection: IScenarioConnection;
  i: Integer;
  namedInputConnection: PNamedInputConnection;
begin
  if Assigned(FViewScenario) then
  begin
    // если выбрана кнопка создания слоя - бросаем слой
    if acnNewLayer.Checked then
    begin
      acnNewLayer.Checked := false;
      {$ifdef fpc}
      tbNewLayer.Down := false;
      {$else}
      tbNewLayer.StaysPressed := false;
      {$endif}
      if frmStringValue.ShowModal = mrOk then
      begin
        FViewScenario.AddLayer(X, Y, frmStringValue.edStringValue.Text, true, true);
        Redraw;
      end;
    end else
    // выбран плагин на палитре - бросаем его на сценарий
    if Assigned(FPluginsDictionary.Selected) then
    begin

      frmStringValue.edStringValue.Text := WideToString(FPluginsDictionary.Selected.Name);
      if frmStringValue.ShowModal = mrOk then
      begin
        FViewScenario.AddPlugin(X, Y, frmStringValue.edStringValue.Text, WideToString(FPluginsDictionary.Selected.Name));
        Redraw;
      end;
      // сбросим выделенный плагин
      FPluginsDictionary.Selected := nil;

    end else
    if (ssCtrl in Shift) and Assigned(FViewScenario.SelectedProcessor) then
    begin
      node := IScenarioNode(FViewScenario.CurrentLayer.GetChildrenAtPoint(X, Y, false));
      if Assigned(node) and (node.QueryInterface(IDataStreamProcessor, toNode) = S_OK) then
      begin
        if (FViewScenario.SelectedProcessor.QueryInterface(IScenarioPlugin, fromPlugin) = S_OK) and (node.QueryInterface(IScenarioPlugin, toPlugin) = S_OK) then
        begin
          // связь из плагина в плагин
          connection := IScenarioConnection(fromPlugin.HasOutputConnection(toPlugin));
          if Assigned(connection) then
          begin
            // если такая связь уже есть - значит удалим её без лишних диалогов;
            // в светлом будующем, если утвердится реализация плагинов которые могут иметь разнотипные входы,
            // необходимо будет это учесть
            FViewScenario.DeleteConnection(TViewConnection(connection.View));
            Redraw;
            exit;
          end;
        end;

        // для удобства предустанавливаем в диалоге вероятный номер выхода который захочет создать пользователь
        frmCreateConnection.edIntegerValue.Text := IntToStr(FViewScenario.SelectedProcessor.CountOutputConnections);
        if Assigned(toPlugin) and Assigned(toPlugin.PluginDescription) then
        begin
          if (toPlugin.PluginDescription.CountNamedInputConnections > 0) and Assigned(toPlugin.PluginDescription.NamedInputConnection) then
          begin
            frmCreateConnection.CountInputs := toPlugin.PluginDescription.CountNamedInputConnections;
            namedInputConnection := toPlugin.PluginDescription.NamedInputConnection;
            for i := 0 to toPlugin.PluginDescription.CountNamedInputConnections - 1 do
            begin
              frmCreateConnection.cbInputs.Items[i] := namedInputConnection.Name;
              inc(namedInputConnection);
            end;
          end else
            frmCreateConnection.CountInputs := 1;

          if frmCreateConnection.ShowModal = mrOk then
          begin
            FViewScenario.CreateConnection(TViewDataStreamProcessor<IDataStreamProcessor>(FViewScenario.SelectedProcessor.View),
              TViewDataStreamProcessor<IDataStreamProcessor>(toNode.View), StrToInt(frmCreateConnection.edIntegerValue.Text), frmCreateConnection.cbInputs.ItemIndex);
            Redraw;
          end;
        end else
        {$WARN SYMBOL_DEPRECATED OFF}
          MessageDlg('Плагин назначения связи не найден. Возможно неактуальная версия библиотеки или плагин не зарегистрирован в файле "Plugins.json"', TMsgDlgType.mtError, [TMsgDlgBtn.mbOK], 0);
        {$WARN SYMBOL_DEPRECATED ON}
      end;
    end else
    begin
      FViewScenario.MouseDown(X, Y, Shift);
    end;
  end;
end;

procedure TFrameScenarioEditor.pbHolstPaint(Sender: TObject; Canvas: TCanvas);
begin
  if Assigned(FViewScenario) then
    FViewScenario.Draw(Point(sbScenHolst.HorzScrollBar.Position, sbScenHolst.VertScrollBar.Position));
end;

procedure TFrameScenarioEditor.Redraw;
begin
  //TThread.Queue(nil, Render);
  Render;
end;

procedure TFrameScenarioEditor.Render;
begin
  sbScenHolst.Invalidate;
end;

procedure TFrameScenarioEditor.SetEventListener(const Value: TEventListener);
begin
  FEventListener := Value;
  if Assigned(FViewScenario) then
    FViewScenario.EventListener := FEventListener;
end;

procedure TFrameScenarioEditor.SetViewScenario(const Value: TViewScenario);
begin
  if FViewScenario = Value then
    exit;
  FViewScenario := Value;
  if Assigned(FViewScenario) then
  begin
    FViewScenario.OnSelectPlugin := DoOnSelectPlugin;
    FViewScenario.OnSelectNode := DoOnSelectNode;
    FViewScenario.OnChangeNode := OnChangeScenarioNode;
    FViewScenario.OnShowLayer := OnLoadLayer;
    FViewScenario.EventListener := FEventListener;
  end;
  CalculateScenarioArea;
  EnableDisableTools;
end;

end.
